#ifndef CLASS_L4dqCheck_HH
#define CLASS_L4dqCheck_HH

// header file for the L4dqCheck class library
// =========================================================
// Author: Martin Braeuer 
//
// NOTE: 
//
// Modification history:
// MBr  Nov 99  : Initial developmant as part of ALIGN5
// MBr  29.11.99: 1. porting to ARTE
// MBr  30.11.99: bad-chip-side mechanism to this Class
//                    => aviation1-0
// =========================================================


#include "Struct_Gede.h"   // use Gede_t structs
#include "Struct_Hitb.h"   // use Hitb_t structs
#include "Struct_Track.h"  // use Track_t structs

#include "Class_GEDE.h"    // geometry information
#include "Class_HOLMES.h"  // holmes track finding

#define LAY_N1 1                    // layer ID of first  n-side 
#define LAY_P1 2                    // layer ID of first  p-side 
#define LAY_P2 3                    // layer ID of second p-side 
#define LAY_N2 4                    // layer ID of second n-side

// DQ stuff:

// descriptors:
#define H_WORKING   0 // flag : "Helix OK"
#define H_LOW      -1 // flag : "too few  cluster/chip"
#define H_HIGH      1 // flag : "too many cluster/chip"

// size of occupancy checks:
#define MAXUNSIN 128
#define OCCCHANS 128
#define MAXHELIX 10

// min number of accumulated hits:
#define MIN_ACCU 1000

// helixcuts (default)
// cluster < events/EVCUT => dead!
#define EVCUT     50
    
// cluster < MINCUT       => dead!
#define MINCUT      30    

// < (MAXCUT*meanhit) = max no of hits/chip
#define MAXCUT_N    1.8  

// < (MAXCUT*meanhit) = max no of hits/chip
#define MAXCUT_P    2.8  

// >HELIXCUT number of chips need to work 
#define HELIXCUT    2    

// a plane mus have > meanhit/minmeanhit hits!
#define MINMEANHIT  2.0    


// the class definition for align5 -------------------------------------------

class L4dqCheck_class {
private:
  // functions
       // Print an error message to the user:
       void fError_out(const char *);
       void dump();
   
  // data
       GEDE_class   *GEDEp;           // Access pointer to GEDE
       HOLMES_class *HOLMESp;           // Access pointer to GEDE

       // The stuff for DQ-checks:----------------------------------------------------
       int   ocche[MAXUNSIN][MAXHELIX];
       int   helixstat[MAXUNSIN][MAXHELIX+1];
       int   o_helixstat[MAXUNSIN][MAXHELIX+1];
       float globalmeanhit; // mean number of hits per plane
       int   ntrack;
       int   have_acc;      // number of accumulated hits (last) 
       int   have_events;   // number of accumulated events (last) 

       // the cuts ff:
       int   cEVCUT    ;  // cluster < events/EVCUT => dead!
       int   cMINCUT   ;  // cluster < MINCUT       => dead!
       float cMAXCUT_N ;  // < (MAXCUT*meanhit) = max no of hits/chip
       float cMAXCUT_P ;  // < (MAXCUT*meanhit) = max no of hits/chip
       int   cHELIXCUT ;  // >HELIXCUT number of chips need to work 
       float cMINMEANHIT; // a plane maust >minmeanhit hits
 
       //------------------------------------------------------------------------------

       int    prepare;      // 1 if prepare step was performed, 0 else
       int    accumulate;   // counts the number of accumulation steps so far
       int    evaluate;     // 1 if evaluation step was performed, 0 else

       // the function to call for Occ storing:
       int (*Pstore)(int, int, int);    // pointer to store function ==0 -> int. store


public:
  // constructors
       L4dqCheck_class(); 
  // constructor with parameters:
       /* ==================================================================
        * ==          : Prepare structures for L4dqCheck class            ==
        * == Parameter: Pointer to HOLMES structure                       ==
        * ==            A function taking usid, chip and value to store   ==
        * ==            occupancies in rhp objects                        ==
        * ==                                                              ==
        * ================================================================== 
        */
       L4dqCheck_class(HOLMES_class *HOLMESp,
		       int (*)(int, int, int) ); 
					 
  // Functions
       /* ==================================================================
        * == fClear : Clear/unbook all allocated structures               ==
        * ==                                                              ==
        * ================================================================== 
        */
        void fClear();


       /* ==================================================================
        * == fPrepare : Reset structures for L4dqCheck class              ==
        * == Parameter: ----                                              ==
        * == needs    : Class exisiting                                   ==
        * ==                                                              ==
        * ================================================================== 
        */
       int fPrepare();

       /* ==================================================================
        * == fUnsin :  Calculate the 7bit ID of sides:                    ==
        * == needs  : --                                                  ==
        * == pars   : sl=superlayer, sec=sector, lay=layer (1..count)     ==
        * == return : The USID (0..127)                                   ==
        * ==                                                              ==
        * ================================================================== 
        */
       inline int fUnsin(int sl,int sec,int lay) { 
	 int i = (lay-1)+(4*(sec-1))+(16*(sl-1));
	 return(i); 
       }

       /* ==================================================================
        * == fSideUnsin :  Calculate the 7bit ID of side                  ==
        * == needs  : --                                                  ==
        * == pars   : side                                                ==
        * == return : The USID (0..127)                                   ==
        * ==                                                              ==
        * ================================================================== 
        */
       inline int fSideUnsin(int side) {
	 if (side < 20000)
	   return((side%10) + 4*((side%1000)/300) + 16*(  int(side/10)%10)); 
	 else 
	   return((side%10) + 4*((side%1000)/300) + 16*(1+int(side/10000))); 
       }

       /* ==================================================================
        * == fHelixid :  Sort a cog of a cluster into a HELIX ID          ==
        * == needs    : --                                                ==
        * == pars     : side=side=sideID to distinguish p/n, cog of clust ==
        * == return   : chip no (0..9)                                    ==
        * ==                                                              ==
        * ================================================================== 
        */
       inline int fHelixid(int side,float cog) {
	 int ichip; 
	 if (((side%10) == 1) || ((side%10) == 2)) {
	   if (cog<966)
	     ichip = int((58+cog)/128);
	   else
	     ichip = 0;
	   return(ichip);
	 } else
	   return(int(cog/128));
       }

       /* ==================================================================
        * == fSetPar : Set Cuts for helix states                          ==
        * == needs  : Class prepared                                      ==
        * == pars   : evcut     cluster < events/EVCUT => dead!           ==
        * ==          mincut    cluster < MINCUT       => dead!           ==
        * ==          maxn      (MAXN*meanhit) = max no of hits/chip  N   ==
        * ==          maxp      (MAXP*meanhit) = max no of hits/chip  P   ==
        * ==          minchips  >MINCHIPS number of chips need to work    ==
	* ==          minmeanhit > <hits all sides>/minmeanhit must       ==
	* ==                       be on a plane                          ==
        * == return : 1 if already inited                                 ==
        * ==                                                              ==
        * ================================================================== 
        */
       int fSetPar(int evcut, int mincut, float maxn, float maxp, 
		   int minchips,float minmeanhit);

       /* ==================================================================
        * == fGetPar : Get Cuts for helix states                          ==
        * == needs  :  Class prepared                                     ==
        * == pars   : evcut     cluster < events/EVCUT => dead!           ==
        * ==          mincut    cluster < MINCUT       => dead!           ==
        * ==          maxn      (MAXN*meanhit) = max no of hits/chip  N   ==
        * ==          maxp      (MAXP*meanhit) = max no of hits/chip  P   ==
        * ==          minchips  >MINCHIPS number of chips need to work    ==
	* ==          minmeanhit > <hits all sides>/minmeanhit must       ==
	* ==                       be on a plane                          ==
        * == return : 1 if set correctly                                  ==
        * ==                                                              ==
        * ================================================================== 
        */
       int fGetPar(int *evcut, int *mincut, float *maxn, float *maxp,
		   int *minchips,float *minmeanhit);
       

       /* ==================================================================
        * == fSetHelixState : Set states for the chips                    ==
        * == needs  : Class prepared                                      ==
        * == pars   : int UNSIN [0..MAXUNSIN], chip [0..MAXHELIX+1]       ==
        * == return : 1 if already inited                                 ==
        * == Note   : the 11th column of a side flags the whole side!     ==
        * ==                                                              ==
        * ================================================================== 
        */
       inline int fSetHelixState(int unsin, int chip, int value) {
	 if (prepare != 1) return(0); // prepare before
	 if ((unsin > -1) && (unsin < MAXUNSIN) && 
	     (chip  > -1) && (chip  < MAXHELIX+1)) {
	   helixstat[unsin][chip] = value;
	 }
	 return(1);
       }

       /* ==================================================================
        * == fGetHelixState : Get the helix state table                   ==
        * == needs  : Evaluate or init has to have been run before        ==
        * == pars   : int UNSIN [0..MAXUNSIN], chip [0..MAXHELIX+1]       ==
        * == return : -1: if error, helix state else                      ==
        * == Note   : the 11th column of a side flags the whole side!     ==
        * ==                                                              ==
        * ================================================================== 
        */
       inline int fGetHelixState(int unsin, int chip) {
	 if (evaluate != 1) return(-1); // prepare before
	 if ((unsin > -1) && (unsin < MAXUNSIN) && 
	     (chip  > -1) && (chip  < MAXHELIX)) {
	   return(helixstat[unsin][chip]); // okay here comes the result!
	 } 
	 return(-1);
       }

       /* ==================================================================
        * == fSetOccupancy : Set the occupancy to the internal structures ==
        * == needs  : Class prepared                                      ==
        * == pars   : int UNSIN [0..MAXUNSIN], chip [0..MAXHELIX],        ==
        * ==          value to SET int. chanal to                         ==
        * == return : 1 if already inited                                 ==
        * ==                                                              ==
        * ================================================================== 
        */
       inline int fSetOccupancy(int unsin, int chip, int value) {
	 if (prepare != 1) return(0); // prepare before
	 if ((unsin > -1) && (unsin < MAXUNSIN) && 
	     (chip  > -1) && (chip  < MAXHELIX)) {
	   ocche[unsin][chip] = value;
	   accumulate+=value;
	 }
	 return(1);
       }

       /* ==================================================================
        * == fGetOccupancy : Get the Occupancy in HELIX bins              ==
        * == needs  : Evaluated                                           ==
        * == pars   : int UNSIN [0..MAXUNSIN], chip [0..MAXHELIX]         ==
        * ==                                                              ==
        * == return : -1: if error, occupancy else                        ==
        * ==                                                              ==
        * ================================================================== 
        */
       inline int fGetOccupancy(int unsin, int chip) {
	 if (evaluate != 1) return(-1); // prepare before
	 if ((unsin > -1) && (unsin < MAXUNSIN) && 
	     (chip  > -1) && (chip  < MAXHELIX)) {
	   return(ocche[unsin][chip]);
	 }
	 return(-1);
       }

       /* =====================================================================
	* == fAccuTrack: Insert a whole HOLMES track into the int. occ. maps ==
	* == needs     : Class prepared                                      ==
	* == pars      : tarid=internal holems id of track                   ==
	* == return    : 1 if OK, 0 if not prepared or invalid side-code     ==
	* == Note      : Use THIS always to store. A NULL in the contructor  ==
	* ==             will lead to an internal store, a function pointer  ==
	* ==             to external storage. (YOU have to copy back then!)  ==
        * ==                                                              ==
	* ===================================================================== 
	*/
       int fAccuTrack(int traid);


       /* ====================================================================
        * == fAccuHit : Insert a raw hit into the int. occ. maps            ==
        * == needs    : Class prepared                                      ==
        * == pars     : side=side  OR USID of hit, cog cog of hit           ==
        * == return   : 1 if prepared                                       ==
        * == Note     : This works with usid and side id !!!                ==
	* == Note     : Use THIS always to store. A NULL in the contructor  ==
	* ==             will lead to an internal store, a function pointer ==
	* ==             to external storage. (YOU have to copy back then!) ==
        * ==                                                                ==
        * ==================================================================== 
        */
       int fAccuHit(int side,float cog);


       /* ==================================================================
        * == fAccuHit : Insert a raw hit into the int. occ. maps          ==
        * == needs    : Class prepared                                    ==
        * == pars     : sl= superlayer, sec=sector, lay=layer, float cog  ==
        * == return   : 1 if prepared                                     ==
        * == Note     : ARTE counting, start with 1 !                     ==
        * ==                                                              ==
        * ================================================================== 
        */
       inline int fAccuHit(int sl, int sec, int lay ,float cog) {
	 int usid = fUnsin(sl,sec,lay);
	 if ((usid > -1) &&(usid < 128)) { 
	   if (Pstore == NULL) {
	     ocche[usid][fHelixid(lay-1,cog)]++;
	     ++accumulate;
	   } else
	     Pstore(usid,fHelixid(lay-1,cog),1);
	   return(1);
	 } else return(0);
       }



       /* ==================================================================
        * == fAcceptTrack: check that track does not touch bad chips      ==
        * == needs    : Class prepared                                    ==
        * == pars     : tarid=internal holems id of track                 ==
        * == return   : 1 if track does not contain a suspecious hit      ==
        * ==                                                              ==
        * ================================================================== 
        */
       int fAcceptTrack(int traid);

       /* ==================================================================
	* == fAcceptHit: check that a hit does not contain bad chips      ==
	* == needs    : preppare has to run before                        ==
	* == pars     : side ID and cog of hit                            ==
	* == return   : 1 hit OK, 0 if not prepared or hit bad            ==
	* ==                                                              ==
	* ================================================================== 
	*/
       int fAcceptHit(int side,float cog);

       /* ==================================================================
        * == fEvaluate: evaluate the occupancy matrices to derive helixes ==
        * == needs    : fAccuHit has to have been run                     ==
        * == pars     : evdone = number of processed events               ==
	* == return   : 1: no change, 0: chips changed, -1: error         ==
        * ==                                                              ==
        * ================================================================== 
        */
       int fEvaluate1(int evdone);

       /* ==================================================================
        * == fResult  : Textoutput of result                              ==
        * == needs    : fEvaluate1 has to run before                      ==
        * == pars     :                                                   ==
        * == return   : 1 if OK, 0 if not evaluated before                ==
        * ==                                                              ==
        * ================================================================== 
        */
       int fResult();

       /* ==================================================================
        * == fGetWorking : returns the number of working helixes          ==
        * == needs    : fEvaluate1 has to run before                      ==
        * == pars     :                                                   ==
        * == return   : -1: if error, number otherwise                    ==
        * ==                                                              ==
        * ================================================================== 
        */
       int fGetWorking();

       /* ==================================================================
	* == fGetNonWorking : returns the number of not working helixes   ==
	* == needs    : fEvaluate1 has to run before                      ==
	* == pars     :                                                   ==
	* == return   : -1: if error, number otherwise                    ==
	* ==                                                              ==
	* ================================================================== 
	*/
       int fGetNotWorking();

  // Destructor:
       ~L4dqCheck_class();

  // data public members
    // NONE there!
}; // ============== END of Class_L4dqCheck def. ! ==============================

#endif
