/***********************************************************************/
/*  isaelectron                                                        */
/*                                                                     */
/*  intention: electron selection using RICH                           */
/*  input: An ARTE pointer to the track candidate                      */
/*  output: 1 if RICH accepts the candidate only with soft criteria    */
/*          2 if RICH accepts the candidate with hard criteria         */
/*          0 if RICH denies the candidate                             */
/*          -1 if RICH is not able to judge the candidate              */
/*  comments: RICH, in principle, is not designed for electron         */
/*            identification. However, a new likelihood that takes     */
/*            into account the number of photons (looks for the shower */
/*            hypothesis) has been included.                           */
/*            To know the efficiencies that the implied cuts mean,     */
/*            you must look at http://www-hera-b.desy.de/subgroup/     */
/*            /physics/herab/SingleLepton/peralta/SL_welcome.html      */
/*                                                                     */
/*  author: Daniel Peralta, peralta@ecm.ub.es peralta@mail.desy.de     */
/*                                                                     */
/*  first version: 12 October 2000                                     */
/*  updates: - 14 December 2000. Upgrade of cuts using efficiency and  */
/*                  fakes studies on mini files from 1st reprocessing. */
/*                  Cuts optimized thinking on e/pi separation         */
/*           - 21 April 2000. Upgrade of cuts using efficiency and     */
/*                  fakes studies on mini files from 2nd reprocessing. */
/*                  2 levels of identification implemented             */
/*                                                                     */
/*                                                                     */
/***********************************************************************/

#include "arte/arte.hh"
#include <iostream.h>

int anl_isaelectron(ArtePointer<RTRA> candidate) {

  int BINNING = 5; // 5 momentum bins
  int LEVELS = 2;  // 2 levels of identification
  
  float MOMENTUM_BIN[BINNING+1];  // 6 boundaries;
  MOMENTUM_BIN[0] = 0.;
  MOMENTUM_BIN[1] = 3.;
  MOMENTUM_BIN[2] = 5.;
  MOMENTUM_BIN[3] = 10.;
  MOMENTUM_BIN[4] = 18.;
  MOMENTUM_BIN[5] = 1000.;

  float RITER_CUTS[BINNING][LEVELS]; // Proposed cuts in electron likelihood
  float RISE_CUTS[BINNING][LEVELS];  // Proposed cuts in electron likelihood

  int accept = -1;

  ArtePointer<RTRI> rtri(candidate.index());

  float momentum;
  if (candidate->pf == 0.)
    momentum = 0.;
  else
    momentum = fabs(1./candidate->pf);

  // Soft identification. Ghost removal.
    
  RITER_CUTS[0][0] = 0.05;
  RITER_CUTS[1][0] = 0.05;
  RITER_CUTS[2][0] = 0.05;
  RITER_CUTS[3][0] = 0.05;
  RITER_CUTS[4][0] = 0.05;
  
  RISE_CUTS[0][0] = 0.05;
  RISE_CUTS[1][0] = 0.05;
  RISE_CUTS[2][0] = 0.05;
  RISE_CUTS[3][0] = 0.05;
  RISE_CUTS[4][0] = 0.05;

  // Hard identification. e/pi separation.
    
  RITER_CUTS[0][1] = 0.20;
  RITER_CUTS[1][1] = 0.25;
  RITER_CUTS[2][1] = 0.25;
  RITER_CUTS[3][1] = 0.50;
  RITER_CUTS[4][1] = 0.50;

  RISE_CUTS[0][1] = 1.00;
  RISE_CUTS[1][1] = 1.00;
  RISE_CUTS[2][1] = 1.00;
  RISE_CUTS[3][1] = 1.00;
  RISE_CUTS[4][1] = 1.00;

  if(rtri->fls > 0) { // in RICH acceptance
    accept = 0;
    for(int i = 0; i < BINNING; i++) {
      if(momentum >= MOMENTUM_BIN[i] && momentum < MOMENTUM_BIN[i+1]) {
	if(rtri->lse > RITER_CUTS[i][0] || rtri->lte > RISE_CUTS[i][0])
	  accept = 1;
	if((rtri->lse+rtri->lsme)/(1.+rtri->lsme) > RITER_CUTS[i][1])
	  accept = 2;
      }
    }
  }

  return(accept);
}



/***********************************************************************/
/*  isakaon                                                            */
/*                                                                     */
/*  intention: kaon selection using RICH                               */
/*  input: An ARTE pointer to the track candidate                      */
/*  output: 1 if RICH accepts the candidate only with soft criteria    */
/*          2 if RICH accepts the candidate with hard criteria         */
/*          0 if RICH denies the candidate                             */
/*          -1 if RICH is not able to judge the candidate              */
/*  comments: RICH is designed for kaon/pion separation                */
/*            To know the efficiencies that the implied cuts mean,     */
/*            you must look at http://www-hera-b.desy.de/subgroup/     */
/*            /physics/herab/SingleLepton/peralta/SL_welcome.html      */
/*                                                                     */
/*  author: Daniel Peralta, peralta@ecm.ub.es peralta@mail.desy.de     */
/*                                                                     */
/*  first version: 12 October 2000                                     */
/*  updates: - 14 December 2000. Upgrade of cuts using efficiency and  */
/*                  fakes studies on mini files from 1st reprocessing. */
/*                  Cuts optimized thinking on k/pi separation         */
/*           - 21 April 2000. Upgrade of cuts using efficiency and     */
/*                  fakes studies on mini files from 2nd reprocessing. */
/*                  2 levels of identification implemented             */
/*                                                                     */
/*                                                                     */
/***********************************************************************/

int anl_isakaon(ArtePointer<RTRA> candidate) {

  int BINNING = 5; // 5 momentum bins
  int LEVELS = 2;  // 2 levels of identification
  
  float MOMENTUM_BIN[BINNING+1];  // 6 boundaries;
  MOMENTUM_BIN[0] = 0.;
  MOMENTUM_BIN[1] = 3.;
  MOMENTUM_BIN[2] = 5.;
  MOMENTUM_BIN[3] = 10.;
  MOMENTUM_BIN[4] = 18.;
  MOMENTUM_BIN[5] = 1000.;

  float RITER_CUTS[BINNING][LEVELS]; // Proposed cuts in kaon likelihood
  float RISE_CUTS[BINNING][LEVELS];  // Proposed cuts in kaon likelihood

  int accept = -1;

  ArtePointer<RTRI> rtri(candidate.index());

  float momentum;
  if (candidate->pf == 0.)
    momentum = 0.;
  else
    momentum = fabs(1./candidate->pf);

  // Soft identification. Ghost removal and soft k/pi separation.
    
  RITER_CUTS[0][0] = 0.30;
  RITER_CUTS[1][0] = 0.30;
  RITER_CUTS[2][0] = 0.30;
  RITER_CUTS[3][0] = 0.05;
  RITER_CUTS[4][0] = 0.05;
  
  RISE_CUTS[0][0] = 1.00;
  RISE_CUTS[1][0] = 1.00;
  RISE_CUTS[2][0] = 1.00;
  RISE_CUTS[3][0] = 0.05;
  RISE_CUTS[4][0] = 0.05;

  // Hard identification. Hard k/pi separation.
    
  RITER_CUTS[0][1] = 0.35;
  RITER_CUTS[1][1] = 0.35;
  RITER_CUTS[2][1] = 0.35;
  RITER_CUTS[3][1] = 0.90;
  RITER_CUTS[4][1] = 0.90;

  RISE_CUTS[0][1] = 1.00;
  RISE_CUTS[1][1] = 1.00;
  RISE_CUTS[2][1] = 1.00;
  RISE_CUTS[3][1] = 0.90;
  RISE_CUTS[4][1] = 0.90;

  if(rtri->fls > 0) { // in RICH acceptance
    accept = 0;
    for(int i = 0; i < BINNING; i++) {
      if(momentum >= MOMENTUM_BIN[i] && momentum < MOMENTUM_BIN[i+1]) {
	if(rtri->lsk > RITER_CUTS[i][0] || rtri->ltk > RISE_CUTS[i][0])
	  accept = 1;
	if(rtri->lsk > RITER_CUTS[i][1] || rtri->ltk > RISE_CUTS[i][1])
	  accept = 2;
      }
    }
  }

  return(accept);
}



/***********************************************************************/
/*  isamuon                                                            */
/*                                                                     */
/*  intention: muon selection using RICH                               */
/*  input: An ARTE pointer to the track candidate                      */
/*  output: 1 if RICH accepts the candidate only with soft criteria    */
/*          2 if RICH accepts the candidate with hard criteria         */
/*          0 if RICH denies the candidate                             */
/*          -1 if RICH is not able to judge the candidate              */
/*  comments: RICH is no designed for muon identification              */
/*            To know the efficiencies that the implied cuts mean,     */
/*            you must look at http://www-hera-b.desy.de/subgroup/     */
/*            /physics/herab/SingleLepton/peralta/SL_welcome.html      */
/*                                                                     */
/*  author: Daniel Peralta, peralta@ecm.ub.es peralta@mail.desy.de     */
/*                                                                     */
/*  first version: 12 October 2000                                     */
/*  updates: - 14 December 2000. Upgrade of cuts using efficiency and  */
/*                  fakes studies on mini files from 1st reprocessing. */
/*                  Cuts optimized thinking on mu/k separation         */
/*           - 21 April 2000. Upgrade of cuts using efficiency and     */
/*                  fakes studies on mini files from 2nd reprocessing. */
/*                  2 levels of identification implemented             */
/*                                                                     */
/*                                                                     */
/***********************************************************************/

int anl_isamuon(ArtePointer<RTRA> candidate) {

  int BINNING = 5; // 5 momentum bins
  int LEVELS = 2;  // 2 levels of identification
  
  float MOMENTUM_BIN[BINNING+1];  // 6 boundaries;
  MOMENTUM_BIN[0] = 0.;
  MOMENTUM_BIN[1] = 3.;
  MOMENTUM_BIN[2] = 5.;
  MOMENTUM_BIN[3] = 10.;
  MOMENTUM_BIN[4] = 18.;
  MOMENTUM_BIN[5] = 1000.;

  float RITER_CUTS[BINNING][LEVELS]; // Proposed cuts in muon likelihood
  float RISE_CUTS[BINNING][LEVELS];  // Proposed cuts in muon likelihood

  int accept = -1;

  ArtePointer<RTRI> rtri(candidate.index());

  float momentum;
  if (candidate->pf == 0.)
    momentum = 0.;
  else
    momentum = fabs(1./candidate->pf);

  // Soft identification. Ghost removal.
    
  RITER_CUTS[0][0] = 0.15;
  RITER_CUTS[1][0] = 0.05;
  RITER_CUTS[2][0] = 0.05;
  RITER_CUTS[3][0] = 0.05;
  RITER_CUTS[4][0] = 0.05;
  
  RISE_CUTS[0][0] = 1.00;
  RISE_CUTS[1][0] = 0.05;
  RISE_CUTS[2][0] = 0.05;
  RISE_CUTS[3][0] = 0.05;
  RISE_CUTS[4][0] = 0.05;

  // Hard identification. Soft mu/k separation.
    
  RITER_CUTS[0][1] = 1.00;
  RITER_CUTS[1][1] = 0.30;
  RITER_CUTS[2][1] = 0.30;
  RITER_CUTS[3][1] = 0.30;
  RITER_CUTS[4][1] = 0.30;

  RISE_CUTS[0][1] = 1.00;
  RISE_CUTS[1][1] = 0.30;
  RISE_CUTS[2][1] = 0.30;
  RISE_CUTS[3][1] = 0.30;
  RISE_CUTS[4][1] = 0.30;

  if(rtri->fls > 0) { // in RICH acceptance
    accept = 0;
    for(int i = 0; i < BINNING; i++) {
      if(momentum >= MOMENTUM_BIN[i] && momentum < MOMENTUM_BIN[i+1]) {
	if(rtri->lsmu > RITER_CUTS[i][0] || rtri->ltmu > RISE_CUTS[i][0])
	  accept = 1;
	if(rtri->lsmu > RITER_CUTS[i][1] || rtri->ltmu > RISE_CUTS[i][1])
	  accept = 2;
      }
    }
  }

  return(accept);
}



/***********************************************************************/
/*  isapion                                                            */
/*                                                                     */
/*  intention: pion selection using RICH                               */
/*  input: An ARTE pointer to the track candidate                      */
/*  output: 1 if RICH accepts the candidate only with soft criteria    */
/*          2 if RICH accepts the candidate with hard criteria         */
/*          0 if RICH denies the candidate                             */
/*          -1 if RICH is not able to judge the candidate              */
/*  comments: RICH is not designed for pion identification             */
/*            To know the efficiencies that the implied cuts mean,     */
/*            you must look at http://www-hera-b.desy.de/subgroup/     */
/*            /physics/herab/SingleLepton/peralta/SL_welcome.html      */
/*                                                                     */
/*  author: Daniel Peralta, peralta@ecm.ub.es peralta@mail.desy.de     */
/*                                                                     */
/*  first version: 12 October 2000                                     */
/*  updates: - 14 December 2000. Upgrade of cuts using efficiency and  */
/*                  fakes studies on mini files from 1st reprocessing. */
/*                  Cuts optimized thinking on pi/k+p separation       */
/*           - 21 April 2000. Upgrade of cuts using efficiency and     */
/*                  fakes studies on mini files from 2nd reprocessing. */
/*                  2 levels of identification implemented             */
/*                                                                     */
/*                                                                     */
/***********************************************************************/

int anl_isapion(ArtePointer<RTRA> candidate) {

  int BINNING = 5; // 5 momentum bins
  int LEVELS = 2;  // 2 levels of identification
  
  float MOMENTUM_BIN[BINNING+1];  // 6 boundaries;
  MOMENTUM_BIN[0] = 0.;
  MOMENTUM_BIN[1] = 3.;
  MOMENTUM_BIN[2] = 5.;
  MOMENTUM_BIN[3] = 10.;
  MOMENTUM_BIN[4] = 18.;
  MOMENTUM_BIN[5] = 1000.;

  float RITER_CUTS[BINNING][LEVELS]; // Proposed cuts in pion likelihood
  float RISE_CUTS[BINNING][LEVELS];  // Proposed cuts in pion likelihood

  int accept = -1;

  ArtePointer<RTRI> rtri(candidate.index());

  float momentum;
  if (candidate->pf == 0.)
    momentum = 0.;
  else
    momentum = fabs(1./candidate->pf);

  // Soft identification. Ghost removal.
    
  RITER_CUTS[0][0] = 0.15;
  RITER_CUTS[1][0] = 0.05;
  RITER_CUTS[2][0] = 0.05;
  RITER_CUTS[3][0] = 0.05;
  RITER_CUTS[4][0] = 0.05;
  
  RISE_CUTS[0][0] = 1.00;
  RISE_CUTS[1][0] = 0.05;
  RISE_CUTS[2][0] = 0.05;
  RISE_CUTS[3][0] = 0.05;
  RISE_CUTS[4][0] = 0.05;

  // Hard identification. Soft pi/k+p separation.
    
  RITER_CUTS[0][1] = 1.00;
  RITER_CUTS[1][1] = 0.30;
  RITER_CUTS[2][1] = 0.30;
  RITER_CUTS[3][1] = 0.30;
  RITER_CUTS[4][1] = 0.30;

  RISE_CUTS[0][1] = 1.00;
  RISE_CUTS[1][1] = 0.30;
  RISE_CUTS[2][1] = 0.30;
  RISE_CUTS[3][1] = 0.30;
  RISE_CUTS[4][1] = 0.30;

  if(rtri->fls > 0) { // in RICH acceptance
    accept = 0;
    for(int i = 0; i < BINNING; i++) {
      if(momentum >= MOMENTUM_BIN[i] && momentum < MOMENTUM_BIN[i+1]) {
	if(rtri->lspi > RITER_CUTS[i][0] || rtri->ltpi > RISE_CUTS[i][0])
	  accept = 1;
	if(rtri->lspi > RITER_CUTS[i][1] || rtri->ltpi > RISE_CUTS[i][1])
	  accept = 2;
      }
    }
  }

  return(accept);
}




/***********************************************************************/
/*  isaproton                                                          */
/*                                                                     */
/*  intention: proton selection using RICH                             */
/*  input: An ARTE pointer to the track candidate                      */
/*  output: 1 if RICH accepts the candidate only with soft criteria    */
/*          2 if RICH accepts the candidate with hard criteria         */
/*          0 if RICH denies the candidate                             */
/*          -1 if RICH is not able to judge the candidate              */
/*  comments: RICH is designed for proton/pion separation              */
/*            To know the efficiencies that the implied cuts mean,     */
/*            you must look at http://www-hera-b.desy.de/subgroup/     */
/*            /physics/herab/SingleLepton/peralta/SL_welcome.html      */
/*                                                                     */
/*  author: Daniel Peralta, peralta@ecm.ub.es peralta@mail.desy.de     */
/*                                                                     */
/*  first version: 12 October 2000                                     */
/*  updates: - 14 December 2000. Upgrade of cuts using efficiency and  */
/*                  fakes studies on mini files from 1st reprocessing. */
/*                  Cuts optimized thinking on p/pi separation         */
/*           - 21 April 2000. Upgrade of cuts using efficiency and     */
/*                  fakes studies on mini files from 2nd reprocessing. */
/*                  2 levels of identification implemented             */
/*                                                                     */
/*                                                                     */
/***********************************************************************/

int anl_isaproton(ArtePointer<RTRA> candidate) {

  int BINNING = 5; // 5 momentum bins
  int LEVELS = 2;  // 2 levels of identification
  
  float MOMENTUM_BIN[BINNING+1];  // 6 boundaries;
  MOMENTUM_BIN[0] = 0.;
  MOMENTUM_BIN[1] = 3.;
  MOMENTUM_BIN[2] = 5.;
  MOMENTUM_BIN[3] = 10.;
  MOMENTUM_BIN[4] = 18.;
  MOMENTUM_BIN[5] = 1000.;

  float RITER_CUTS[BINNING][LEVELS]; // Proposed cuts in proton likelihood
  float RISE_CUTS[BINNING][LEVELS];  // Proposed cuts in proton likelihood

  int accept = -1;

  ArtePointer<RTRI> rtri(candidate.index());

  float momentum;
  if (candidate->pf == 0.)
    momentum = 0.;
  else
    momentum = fabs(1./candidate->pf);

  // Soft identification. Ghost removal and soft p/pi separation.
    
  RITER_CUTS[0][0] = 0.30;
  RITER_CUTS[1][0] = 0.30;
  RITER_CUTS[2][0] = 0.30;
  RITER_CUTS[3][0] = 0.45;
  RITER_CUTS[4][0] = 0.05;
  
  RISE_CUTS[0][0] = 1.00;
  RISE_CUTS[1][0] = 1.00;
  RISE_CUTS[2][0] = 1.00;
  RISE_CUTS[3][0] = 1.00;
  RISE_CUTS[4][0] = 0.05;

  // Hard identification. Hard p/pi separation.
    
  RITER_CUTS[0][1] = 0.35;
  RITER_CUTS[1][1] = 0.35;
  RITER_CUTS[2][1] = 0.35;
  RITER_CUTS[3][1] = 0.50;
  RITER_CUTS[4][1] = 0.90;

  RISE_CUTS[0][1] = 1.00;
  RISE_CUTS[1][1] = 1.00;
  RISE_CUTS[2][1] = 1.00;
  RISE_CUTS[3][1] = 1.00;
  RISE_CUTS[4][1] = 0.90;

  if(rtri->fls > 0) { // in RICH acceptance
    accept = 0;
    for(int i = 0; i < BINNING; i++) {
      if(momentum >= MOMENTUM_BIN[i] && momentum < MOMENTUM_BIN[i+1]) {
	if(rtri->lsp > RITER_CUTS[i][0] || rtri->ltp > RISE_CUTS[i][0])
	  accept = 1;
	if(rtri->lsp > RITER_CUTS[i][1] || rtri->ltp > RISE_CUTS[i][1])
	  accept = 2;
      }
    }
  }

  return(accept);
}













