//                              -*- Mode: C++ -*-
//: EDClusHypotModifier.C
//
//
// Author: Mauro Villa   
// ECAL group 
// HERA-B experiment, Desy
// e-mail: villa@bo.infn.it
//
// Created 2/5/97
// Last Update ........
//
#undef HIST

#ifdef HIST
#  include "ana/Ntuple.hh"
#endif
#include "care/EMClusterTree.hh"
#include "care/EDClusHypotModifier.hh"
#include "care/EMHitCell.hh"
#include "arte/ArteMessage.hh"
#include "care/ECStrategyFactory.hh"
#include "care/caredata.hh"

#ifdef HIST
#  include "EAMCTruthAnalyzer.hh"
#endif

/* 
CLASS
  EDClusHypotModifier
  Class for finding the number of hypothetical particles gave origin to
  a cluster.

  The algorithm choseen make use of a constant CutOff level to define an
  hypotetical cluster.


*/


//////////////////// EDClusHypotModifier Class Definitions ////////////////

EDClusHypotModifier::EDClusHypotModifier(){
  //--
  // Get a new Clus-To-Clus measurer
  ECStrategyFactory strfac;
  ctcstrat = strfac.CTCStrategy();
}

// copy constructor
EDClusHypotModifier::EDClusHypotModifier(const EDClusHypotModifier & clhyp){
  numberOfHypotheses = clhyp.numberOfHypotheses;
  newHypot = clhyp.numberOfHypotheses;
  //--
  // Get a new Clus-To-Clus measurer (not really copied)
  ECStrategyFactory strfac;
  ctcstrat = strfac.CTCStrategy();
}


// copy constructor
const EDClusHypotModifier & 
EDClusHypotModifier::operator=(const EDClusHypotModifier & clhyp){
  // delete ctcstrat;
  numberOfHypotheses = clhyp.numberOfHypotheses;
  newHypot = clhyp.numberOfHypotheses;
  //--
  // Get a new Clus-To-Clus measurer (not really copied)
  ECStrategyFactory strfac;
  ctcstrat = strfac.CTCStrategy();

  return *this;
}

// No functions
EDClusHypotModifier::~EDClusHypotModifier(){
  // remember to delete the Clus-To-Clus strategy
  // delete ctcstrat;
}


//--
//  Base function to be called with the TopCluster List
//
void EDClusHypotModifier::analyzeTopList(EMClusterTree & topList){
  doModify(&topList);
}



//--
// Cluster Modifier
void EDClusHypotModifier::doModify(EMCluster * clus){
  //
  //
  Hyp_t storeNumberHypot = numberOfHypotheses;
  numberOfHypotheses = 0;
  newHypot = false;
  if( clus->begin()->isALeaf() ){
    // Analyze a Base Cluster
    // perform operation on all elements of the cluster
    for(EMClusterMember* child = clus->begin(); 
	child!=clus->end();  child = child->next())
      child->modifyMe(*this);
#   ifdef HIST
    h1d[merg1]->accumulate(clus->begin()->maxEnergy());
#   endif
  } else {
    // set the reference cluster 
    // (the cluster to which the distances are evaluated)
    EMCluster* startclus = (EMCluster*) clus->begin();
    ctcstrat->refCluster(startclus);
    // analyze a structured cluster
    // Analyze the first cluster
    clus->begin()->modifyMe(*this);
#ifdef HIST
    // Create a MCtruthAnalyzer
    EAMCtruthAnalyzer  firstClusMC;
    firstClusMC.doAnalysis(clus); // analyze the cluster
    // get the informations of the first cluster
    firstClusMC.sortDeposits();
    ECMCDeposit firstDep = firstClusMC.maxDeposit();
#endif
    //
    // perform operation on all elements of the cluster except the first
    // 
    EMClusterMember* child = clus->begin();
    if( child!=clus->end())
      child = child->next();
    for(; child!=clus->end() && !child->isALeaf();
	child = child->next()){
      EMCluster * otherClus = (EMCluster*) child;
      // Now look for the distance between the first and the other cluster
#ifndef HIST
      ctcstrat->evaluateDis(otherClus);
#    else 
    ChiSquare_t dist  = ctcstrat->evaluateDis(otherClus);

     EAMCtruthAnalyzer otherClusMC;
      // analyze the cluster
      otherClusMC.doAnalysis(otherClus);
      otherClusMC.sortDeposits();
      ECMCDeposit otherDep = otherClusMC.maxDeposit();
      if(((EMCluster *)clus->begin())->getHypotheses()==1 &&
	 otherClus->getHypotheses()==0){
	if(firstDep.track.index()==otherDep.track.index()){
	  h1d[merg2]->accumulate(dist);
	} else {
	  h1d[merg3]->accumulate(dist);
	}
      }
#endif
      // Now look for the distance between the first and the other cluster
      if( ctcstrat->areDifferent()){
	otherClus->modifyMe(*this);
      }
    }
  }
  // test if a base cluster have a flag set 
  if( newHypot && numberOfHypotheses==0)
    numberOfHypotheses++;
  // set the number of hypotheses found
  clus->setHypotheses(numberOfHypotheses);
  // update the number of hypotheses to return
  numberOfHypotheses += storeNumberHypot;
}


//--
// EMHitCell Analyzer
void EDClusHypotModifier::doModify(EMHitCell * hitT){
  if( hitT->maxEnergy() > ArtePointer<ECHS>((hitT->cellIdx()).index())->thrs )
    newHypot = true;
}

//--
// EMClusterMember Analyzer
void EDClusHypotModifier::doModify(EMClusterMember *){
}

//--
// TrackOnEcal Analyzer
void EDClusHypotModifier::doModify(EMTrackOnEcal * ){
}

//--
// ClusterTree Analyzer
// Loop on all clusters on the tree
void EDClusHypotModifier::doModify(EMClusterTree * tree){
  numberOfHypotheses = 0;
  for(EMClusterMember* child = tree->begin(); child!=tree->end(); 
      child = child->next()){
    if( !child->isALeaf()){
      child->modifyMe(*this);
    }
  }
  tree->setHypotheses(numberOfHypotheses);
}
