// *****************************************************************************
//
// source:    DAQECALMC.C
//
// type:      source code
//
// created:   14 Mar 2001
//
// author:    Mauro Villa, Lars Soezueer
//            email: soezueer@mail.desy.de
//
// purpose:   ECALgetFED: provide an FED from ECAL
//            ECALgetFED_IDs: provide the numbers of FEDs in ECAL
//            called by daq_ecal_if.C
//
// changes:   
//
// *****************************************************************************
extern "C" {
#include "esp.h"
}
#include "event.h"
#include "component.h"
#include "arte/arte.hh"
#include "care/ECALtypes.hh"
#include "care/ECMCcalib.hh"
#include "care/ECCellID.hh"
#include "care/dgcal.hh" 

////// prototypes

// Return the number of FEDs
unsigned int ECALgetFED_IDs(unsigned int slbid, unsigned short *buf, 
			      unsigned int maxIDs);

// Assemble an unsparsified FED record in buf in export format. 
// Return 0 on success.
int ECALgetFED(unsigned short fedid, FED_HDR *buf, unsigned int maxSize);

// write out an unsigned short in export format
void ECputNextShort(unsigned char *p, unsigned short ch);


/////// definition of functions


// SLBs ids go from 0 to getN_SLBs()-1. Return the number of FEDs
unsigned int ECALgetFED_IDs(unsigned int slbid, 
			      unsigned short *buf, 
			      unsigned int maxIDs){

  unsigned int nfeds = 0; // return variable
  if(slbid==0 && buf!=0){  // accept only SLB ID equal 0
    unsigned int nfed = ArteTable<ECBO>::nrow();
    nfed = (nfed>maxIDs) ? maxIDs : nfed;
    for(ArteTable<ECBO>::iterator ecbo=ArteTable<ECBO>::begin(); 
	ecbo!=ArteTable<ECBO>::end() && nfeds<nfed; ++ecbo, ++nfeds){
      *buf = (unsigned short) ecbo->fdid;
      buf++;
    }
  }
  return nfeds;
}


// assemble an unsparsified FED record in buf in export format. 
// Return 0 on success.

int ECALgetFED(unsigned short fedid, FED_HDR *buf, unsigned int maxSize){
  int rc = 0; // assume good input data

  if( fedid>=16 && fedid<=241 && 
      maxSize>64+sizeof(FED_HDR)+sizeof(FED_TAIL) && buf!=0){
    // fill the header
    buf->length = 64+sizeof(FED_HDR)+sizeof(FED_TAIL);
    buf->recid = 0x10E0; // good data from ECAL BOARDS
    buf->fltevt = 0;
    buf->bxid = 0;
    buf->compid = COMP_ID_ECAL; // ECAL component
    buf->fedid = fedid;
    // fill the data part
    unsigned short *p = (unsigned short*)(buf+1);
    int ix = 0, iy = 0, ndim = 0;
    ECMCcalib  mcCalib;          // Calibrator object
    ArtePointer<ECBO> ecbo(fedid-15);
    for (int j=0; j<32; j++, p++){
      int cellidx = ecbo->chan[j]; 
      if(cellidx != -1){          
	CellIdx_t cell = cellidx;
	ECCellID  cellid(cell->ceid);
	cellid.getModuleIndex(ix, iy, ndim);
	float ener = dgcal_enecalc(cellid.module(), ix, iy);
      	ener *= mcCalib[ndim];  // MC Energy correction 
	unsigned short adcCh = dgcalech(cellidx, ener); // digitalized energy
	ECputNextShort((unsigned char*)p, adcCh);
      } else {
        ECputNextShort((unsigned char*)p, 0);
      }
    }
    FED_TAIL *tail = (FED_TAIL*)p;
    tail->fltevt = 0; 
    // swap data if needed
    ESP_put_fed_hdr(buf, buf);
    // FED data are always unswapped
    ESP_put_fed_tail(tail, tail);
  } else {
    rc = 1; // bad input data
  }

  return rc;
}

void ECputNextShort(unsigned char *p, unsigned short value){
  *p = (value>>8) & 255;
  *(p+1) = (value & 255);
}
