
// *****************************************************************************
//
// source:    faARTE.C
//
// type:      source code
//
// created:   13 Jun 2000
//
// author:    Dr. Andreas Gellrich
//            HERA-B Collaboration
//            Humboldt-University Berlin
//            Physics Department
//            Invalidenstr. 110
//            10115 Berlin
//            Germany
//            email: Andreas.Gellrich@desy.de
//
// purpose:   various ARTE utilities
//
// changes:         
// 13 Jun 2000 (AG) creation
// 14 Jun 2000 (AG) solve potential SEGV
// 25 Aug 2000 (AG) improved event buffer handling
//  9 Nov 2000 (AG) added "C" functions for ARTE_nnnn()
// 15 Nov 2000 (AG) polished printout
// 28 Nov 2000 (AG) improved printout of ARTE_error()
//  7 Dec 2000 (AG) improved printout
// 15 Dec 2000 (AG) improved printout
// 19 Dec 2000 (AG) added FAMARK
// 01 Mar 2001 (AG) changed printout "error=" to "rc=" to avoid confusions
// 24 Oct 2001 (JH) do not print colors to stdout when online running
//
// *****************************************************************************

// standard includes
#include <iostream.h>
#include <strstream.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <math.h>
#include <unistd.h>

// ONLINE includes
#include "etb.h"
#include "runcfg.h"
#include "component.h"
#include "grr.h"
#include "screen.h"

// Arte includes
#include "arte/arte.hh"
#include "arte/ArteFlag.h"
#include "farm/fa.hh"
#include "farm/farm.hh"

// defines
#define PRCMHITMASK "HITmask" // PRELIMINARY !!!

// globals
int   ARTE_MARK_CODE = 0;
char *ARTE_MARK_TEXT = 0;

// externals
extern int FADEB;  // debug flag
extern int FAMARK; // mark printout mode
extern int FAGRR;  // grr flag
extern int FAON;    // Online 

extern char *LOGFILE; // logfile name

extern char *evtBuf; // event buffer
extern int  SEG_LEN; // event buffer length (set in faLine.C)

// prototypes
#ifdef __linux
extern "C" long gethostid();
extern "C" int  gethostname( char *hostname, unsigned int length );
#endif

extern "C" void mmbook_( int & , const int & );
extern "C" void mmfree_( int & );

extern "C" void atstopcc_( char *text );

//
// get record id from table DEVT
//
int
ARTE_get_recid_DEVT( int *recid )
{
  if( ArteTable<DEVT>::size() == 0 ){
    return ETB_corrupted_length;
  }
  else {
    ArtePointer<DEVT> dnnn( 1 );
    ESP_get_int( (int *) &dnnn->bits, recid, 1 );
    return ETB_ok;
  }
}

//
// get raw event from DEVT and store it in a save place
//
EVENT_HDR *
ARTE_get_event()
{
  // check table
  if( ArteTable<DEVT>::size() == 0 ){
    if( FADEB ){
      cout << "ARTE_get_event(): NO DEVT table found." << endl;
    }
    return 0;
  }  ArtePointer<DEVT> devt( 2 ); // DEVT table pointer
  int length = ArteTable<DEVT>::size() * sizeof( int ); // length of table

  // buffer not yet allocated
  if( evtBuf == 0 ){

    if( FADEB ){
      cout << "ARTE_get_event(): malloc() length=" << length << endl;
    }

    // new buffer
    evtBuf = (char *) malloc( length ); // create buffer
    SEG_LEN = length;                   // store new length
  }
  // buffer too small
  else if( SEG_LEN < length ){

    if( FADEB ){
      cout << "ARTE_get_event(): malloc() length "
	   << "old=" << SEG_LEN << " new=" << length << endl;
    }

    // free old buffer
    free( evtBuf );
    evtBuf = 0;

    // resize buffer
    evtBuf  = (char *) malloc( length ); // new resized buffer
    SEG_LEN = length;                    // set SEG_LEN to new value
  }

  // final check of buffer
  if( evtBuf == 0 ) return 0; // buffer could not be created

  // copy data
  memcpy( evtBuf, &devt->bits, length ); // copy

  return (EVENT_HDR *) evtBuf;
}

//
// error logging
//
extern "C" int
ArteError( int severity, const char *message, int error )
{
  return ARTE_error( severity, message, error );
}

int
ARTE_error( int severity, const char *message, int error )
{
  // host name
  static char *hostName = 0;
  if( !hostName ) hostName = new char[10];
  memset( hostName, '\0', 10 );

  gethostname( hostName, 9 );

  // current date/time
  time_t tp = time( NULL ); // now
  struct tm *t = localtime( &tp );

  static char *date = 0;
  if( !date ) date = new char[30];
  memset( date, '\0', 30 );

  strftime( date, 30, "%d/%m/%Y %X", t );

  // process id
  int pid = getpid();

  static char *id = 0;
  if( !id ) id = new char[8];
  memset( id, '\0', 8 );

  sprintf( id, "%d", pid );

  // Arte version
  static char *version = 0;
  if( !version ) version = new char[7];
  memset( version, '\0', 7 );

  sprintf( version, "%.4f", arteflag_.artevers );

  // ARTE evt/run/exp
  static char *numEvtRunExp = 0;
  if( !numEvtRunExp ) numEvtRunExp = new char[36];
  memset( numEvtRunExp, '\0', 36 );

  sprintf( numEvtRunExp,
	   "Num:%d EvtRunExp:%d/%d/%d",
	   arteflag_.nevents,
	   arteflag_.numevt,
	   arteflag_.numrun,
	   arteflag_.numexp );

  // build message
  static char *newmsg = 0;
  if( !newmsg ) newmsg = new char[512];
  memset( newmsg, '\0', 512 );

  strcpy( newmsg, "[" );
  strcat( newmsg, hostName );
  strcat( newmsg, ":" );
  strcat( newmsg, id );
  strcat( newmsg, "] " );
  strcat( newmsg, "Arte:" );
  strcat( newmsg, version );
  strcat( newmsg, " " );
  strcat( newmsg, numEvtRunExp );
  strcat( newmsg, " " );
  strcat( newmsg, message );

  // send to online error logger
  int rc = GRR_OK;
  if( FAGRR ){
    rc = grr_error( severity, newmsg, error );
  }

  // severity
  static char *txt = 0;
  if( !txt ) txt = new char[10]; 
  memset( txt, '\0', 10 ); 

  if( severity == GRR_SDEBUG ){
    strcpy( txt, "[debug]" );
  }
  else if( severity == GRR_SINFO ){
    strcpy( txt, "[inform]" );
  }
  else if( severity == GRR_SWARNING ){
    strcpy( txt, "[Warning]" );
  }
  else if( severity == GRR_SERROR ){
    strcpy( txt, "[Error]" );
  }
  else if( severity == GRR_SEVERE ){
    strcpy( txt, "[SEVERE]" );
  }
  else if( severity == GRR_SFATAL ){
    strcpy( txt, "[FATAL]" );
  }
  else {
    strcpy( txt, "[Unkown]" );
  }

  if (FAON) {
    printf("%s %s%s rc=%d\n", date, txt, newmsg, error);
  } else {
    // colors  
    switch ( severity ) 
      {
      case GRR_SDEBUG: 
	black_fore;
	break;
      case GRR_SINFO: 
	magenta_fore;
	break; 
      case GRR_SWARNING: 
	green_fore;
	break;
      case GRR_SERROR: 
	blue_fore;
	break;
      case GRR_SFATAL:
	bold;
	red_fore;
	break;
      default:
	black_fore;
	break;
      }
    
    // print to stdout
    //underline;
    //bold;
    cout             << "\033[4m\033[1m"
		     << date     << " "
		     << txt      << "\033[0m "
		     << newmsg   << " "
		     << "rc="    << error
		     << endl;
    plain;
  }
  return rc;
}

//
// position marker
//
extern "C" int
ArteMark( const char *text, int code )
{
   return ARTE_mark( text, code );
}

int ARTE_mark( const char *text, int code )
{
  // init
  if( !ARTE_MARK_TEXT ) ARTE_MARK_TEXT = new char[100];
  memset( ARTE_MARK_TEXT, '\0', 100 );
  strcpy( ARTE_MARK_TEXT, "atmain" );

  // text
  int len = strlen( text );
  if( len < 100 ){
    strcpy( ARTE_MARK_TEXT, text );
  }

  // code
  ARTE_MARK_CODE = code;

  // print 
  if( FAMARK ) {
    if (FAON) cout << "MARK: " << ARTE_MARK_TEXT << endl;
    else      cout << "\033[4m\033[1m" << "MARK: " << ARTE_MARK_TEXT << "\033[0m" << endl;
  }
  return 0;
}

//
// clean ARTE exit
//
extern "C" int
ArteExit( int severity, const char *message, int error )
{
  return ARTE_exit( severity, message, error );
}

int
ARTE_exit( int severity, const char *message, int error )
{
  ARTE_error( severity, message, error ); // write to grr
  ARTE_log  ( LOGFILE, message );         // write to logfile
  atstopcc_( "ARTE_exit()" );             // stop ARTE
  return 0;                               // return (never reached)
}

//
// write to logfile
//
extern "C" int
ArteLog( const char *logfile, const char *message )
{
  return ARTE_log( logfile, message );
}

int
ARTE_log( const char *logfile, const char *message )
{
  // check file name 
  if( !logfile ) return -1;

  // show up
  cout << "LOG:  Write to logfile=" << logfile << endl;

  // open file
  FILE *file = fopen( logfile, "a+" );
  if( file == NULL ) return -2;

  // current date/time
  time_t tp = time( NULL ); // now
  struct tm *t = localtime( &tp );

  static char *date = 0;
  if( !date ) date = new char[20];
  memset( date, '\0', 20 );

  strftime( date, 20, "%d/%m/%Y %X", t );

  // host name
  static char *hostName = 0;
  if( !hostName ) hostName = new char[30];
  memset( hostName, '\0', 30 );

  gethostname( hostName, 29 );

  // process id
  int pid = getpid();

  // print to file
  fprintf( file, "%s [%s:%d] %d %d/%d/%d %s\n",
	   date,
	   hostName,
	   pid, 
	   arteflag_.nevents,
	   arteflag_.numevt,
	   arteflag_.numrun,
	   arteflag_.numexp,
	   message );

  // close file
  fclose( file );

  // return
  return 0;
}

//
// ARTE_get_recid() ************************************************************
//

//
// get record id from table
//
int
ARTE_get_recid_DSVD( int *recid )
{
  if( ArteTable<DSVD>::size() == 0 ) return ETB_corrupted_length;
  ArtePointer<DSVD> dnnn( 1 );
  ESP_get_int( (int *) &dnnn->bits, recid, 1 );
  return ETB_ok;
}

//
// get record id from table
//
int
ARTE_get_recid_DITR( int *recid )
{
  if( ArteTable<DITR>::size() == 0 ) return ETB_corrupted_length;
  ArtePointer<DITR> dnnn( 1 );
  ESP_get_int( (int *) &dnnn->bits, recid, 1 );
  return ETB_ok;
}

//
// get record id from table
//
int
ARTE_get_recid_DPTC( int *recid )
{
  if( ArteTable<DPTC>::size() == 0 ) return ETB_corrupted_length;
  ArtePointer<DPTC> dnnn( 1 );
  ESP_get_int( (int *) &dnnn->bits, recid, 1 );
  return ETB_ok;
}

//
// get record id from table
//
int
ARTE_get_recid_DOTR( int *recid )
{
  if( ArteTable<DOTR>::size() == 0 ) return ETB_corrupted_length;
  ArtePointer<DOTR> dnnn( 1 );
  ESP_get_int( (int *) &dnnn->bits, recid, 1 );
  return ETB_ok;
}

//
// get record id from table
//
int
ARTE_get_recid_DRIC( int *recid )
{
  if( ArteTable<DRIC>::size() == 0 ) return ETB_corrupted_length;
  ArtePointer<DRIC> dnnn( 1 );
  ESP_get_int( (int *) &dnnn->bits, recid, 1 );
  return ETB_ok;
}

//
// get record id from table
//
int
ARTE_get_recid_DTRD( int *recid )
{
  if( ArteTable<DTRD>::size() == 0 ) return ETB_corrupted_length;
  ArtePointer<DTRD> dnnn( 1 );
  ESP_get_int( (int *) &dnnn->bits, recid, 1 );
  return ETB_ok;
}

//
// get record id from table
//
int
ARTE_get_recid_DCAL( int *recid )
{
  if( ArteTable<DCAL>::size() == 0 ) return ETB_corrupted_length;
  ArtePointer<DCAL> dnnn( 1 );
  ESP_get_int( (int *) &dnnn->bits, recid, 1 );
  return ETB_ok;
}

//
// get record id from table
//
int
ARTE_get_recid_DMUT( int *recid )
{
  if( ArteTable<DMUT>::size() == 0 ) return ETB_corrupted_length;
  ArtePointer<DMUT> dnnn( 1 );
  ESP_get_int( (int *) &dnnn->bits, recid, 1 );
  return ETB_ok;
}

//
// get record id from table
//
int
ARTE_get_recid_DMUX( int *recid )
{
  if( ArteTable<DMUX>::size() == 0 ) return ETB_corrupted_length;
  ArtePointer<DMUX> dnnn( 1 );
  ESP_get_int( (int *) &dnnn->bits, recid, 1 );
  return ETB_ok;
}

//
// get record id from table
//
int
ARTE_get_recid_DMUP( int *recid )
{
  if( ArteTable<DMUP>::size() == 0 ) return ETB_corrupted_length;
  ArtePointer<DMUP> dnnn( 1 );
  ESP_get_int( (int *) &dnnn->bits, recid, 1 );
  return ETB_ok;
}

//
// ARTE_get_feds() *************************************************************
//

//
// get list of FED blocks (see ETB)
//
int
ARTE_get_feds_DSVD( int max, ETB_fed *feds, int *n )
{
  // inits
  int rc = ETB_ok;
  *n = 0;
  
  // check fed array
  if( ( max < 1 ) || ( feds == 0 ) ){
    return ETB_out_of_memory;
  }
  
  //
  // use DEVT
  //
  EVENT_HDR *event = ARTE_get_event();
  if( event ){
    if( ( rc = ETB_get_feds( event, COMP_ID_SVD, max, feds, n ) ) != ETB_ok ){
      cout << "ARTE_get_feds_DSVD(): ERROR: fed array too small? "
	   << "max=" << max << " "
	   << "n="   << *n  << " "
	   << "rc="  << rc << endl;
    }
    if( *n > max ) *n = max;
    return rc;
  }
  //
  // use DSVD
  //
  else if( ArteTable<DSVD>::size() ){

    // Arte pointer
    ArtePointer<DSVD> dnnn( 1 );
    
    // data words in table (first word is recid)
    int total = ( ArteTable<DSVD>::size() - 1 ) * sizeof( int );
    
    // recid
    int recid = 0;
    ESP_get_int( (int *) dnnn->bits, &recid, 1 );
    
    // SLB_SUBHDR
    unsigned int  bc   = 0;
    unsigned char *ptr = (unsigned char *) dnnn->bits + sizeof( int );
    int j              = 0;
    
    // loop
    while( bc < ( total - sizeof( SLB_SUBHDR ) ) ){
      
      // check array size
      if( j >= max ){
	return ETB_out_of_memory;
      }
      
      // update FED count
      *n = j + 1;
      
      // ETB_fed header
      SLB_SUBHDR *hdr = &feds[j].hdr; // header of FED block
      ESP_get_slb_subhdr( (SLB_SUBHDR *)( ptr + bc ), hdr ); // copy header
      feds[j].slb  = 0; // pointer to slb record (not existent here)
      feds[j].data = ptr + bc + sizeof( SLB_SUBHDR ); // pointer to data
      
      // increase cunters
      j++; // fed counter
      bc += hdr->length; // byte counter
    }
    return ETB_ok;
  }

  // return
  return ETB_corrupted_event;
}  

//
// get list of FED blocks (see ETB)
//
int
ARTE_get_feds_DITR( int max, ETB_fed *feds, int *n )
{
  // inits
  int rc = ETB_ok;
  *n = 0;
  
  // check fed array
  if( ( max < 1 ) || ( feds == 0 ) ){
    return ETB_out_of_memory;
  }
  
  //
  // use DEVT
  //
  EVENT_HDR *event = ARTE_get_event();
  if( event ){
    if( ( rc = ETB_get_feds( event, COMP_ID_ITR, max, feds, n ) ) != ETB_ok ){
      cout << "ARTE_get_feds_DITR(): ERROR: fed array too small? "
	   << "max=" << max << " "
	   << "n="   << *n  << " "
	   << "rc="  << rc << endl;
    }
    if( *n > max ) *n = max;
    return rc;
  }
  //
  // use DITR
  //
  else if( ArteTable<DITR>::size() ){

    // Arte pointer
    ArtePointer<DITR> dnnn( 1 );
    
    // data words in table (first word is recid)
    int total = ( ArteTable<DITR>::size() - 1 ) * sizeof( int );
    
    // recid
    int recid = 0;
    ESP_get_int( (int *) dnnn->bits, &recid, 1 );
    
    // SLB_SUBHDR
    unsigned int  bc   = 0;
    unsigned char *ptr = (unsigned char *) dnnn->bits + sizeof( int );
    int j              = 0;
    
    // loop
    while( bc < ( total - sizeof( SLB_SUBHDR ) ) ){
      
      // check array size
      if( j >= max ){
	return ETB_out_of_memory;
      }
      
      // update FED count
      *n = j + 1;
      
      // ETB_fed header
      SLB_SUBHDR *hdr = &feds[j].hdr; // header of FED block
      ESP_get_slb_subhdr( (SLB_SUBHDR *)( ptr + bc ), hdr ); // copy header
      feds[j].slb  = 0; // pointer to slb record (not existent here)
      feds[j].data = ptr + bc + sizeof( SLB_SUBHDR ); // pointer to data
      
      // increase cunters
      j++; // fed counter
      bc += hdr->length; // byte counter
    }
    return ETB_ok;
  }

  // return
  return ETB_corrupted_event;
}  

//
// get list of FED blocks (see ETB)
//
int
ARTE_get_feds_DPTC( int max, ETB_fed *feds, int *n )
{
  // inits
  int rc = ETB_ok;
  *n = 0;
  
  // check fed array
  if( ( max < 1 ) || ( feds == 0 ) ){
    return ETB_out_of_memory;
  }
  
  //
  // use DEVT
  //
  EVENT_HDR *event = ARTE_get_event();
  if( event ){
    if( ( rc = ETB_get_feds( event, COMP_ID_HIPT, max, feds, n ) ) != ETB_ok ){
      cout << "ARTE_get_feds_DPTC(): ERROR: fed array too small? "
	   << "max=" << max << " "
	   << "n="   << *n  << " "
	   << "rc="  << rc << endl;
    }
    if( *n > max ) *n = max;
    return rc;
  }
  //
  // use DPTC
  //
  else if( ArteTable<DPTC>::size() ){

    // Arte pointer
    ArtePointer<DPTC> dnnn( 1 );
    
    // data words in table (first word is recid)
    int total = ( ArteTable<DPTC>::size() - 1 ) * sizeof( int );
    
    // recid
    int recid = 0;
    ESP_get_int( (int *) dnnn->bits, &recid, 1 );
    
    // SLB_SUBHDR
    unsigned int  bc   = 0;
    unsigned char *ptr = (unsigned char *) dnnn->bits + sizeof( int );
    int j              = 0;
    
    // loop
    while( bc < ( total - sizeof( SLB_SUBHDR ) ) ){
      
      // check array size
      if( j >= max ){
	return ETB_out_of_memory;
      }
      
      // update FED count
      *n = j + 1;
      
      // ETB_fed header
      SLB_SUBHDR *hdr = &feds[j].hdr; // header of FED block
      ESP_get_slb_subhdr( (SLB_SUBHDR *)( ptr + bc ), hdr ); // copy header
      feds[j].slb  = 0; // pointer to slb record (not existent here)
      feds[j].data = ptr + bc + sizeof( SLB_SUBHDR ); // pointer to data
      
      // increase cunters
      j++; // fed counter
      bc += hdr->length; // byte counter
    }
    return ETB_ok;
  }

  // return
  return ETB_corrupted_event;
}  

//
// get list of FED blocks (see ETB)
//
int
ARTE_get_feds_DOTR( int max, ETB_fed *feds, int *n )
{
  // inits
  int rc = ETB_ok;
  *n = 0;
  
  // check fed array
  if( ( max < 1 ) || ( feds == 0 ) ){
    return ETB_out_of_memory;
  }
  
  //
  // use DEVT
  //
  EVENT_HDR *event = ARTE_get_event();
  if( event ){
    if( ( rc = ETB_get_feds( event, COMP_ID_OTR, max, feds, n ) ) != ETB_ok ){
      cout << "ARTE_get_feds_DOTR(): ERROR: fed array too small? "
	   << "max=" << max << " "
	   << "n="   << *n  << " "
	   << "rc="  << rc << endl;
    }
    if( *n > max ) *n = max;
    return rc;
  }
  //
  // use DOTR
  //
  else if( ArteTable<DOTR>::size() ){

    // Arte pointer
    ArtePointer<DOTR> dnnn( 1 );
    
    // data words in table (first word is recid)
    int total = ( ArteTable<DOTR>::size() - 1 ) * sizeof( int );
    
    // recid
    int recid = 0;
    ESP_get_int( (int *) dnnn->bits, &recid, 1 );
    
    // SLB_SUBHDR
    unsigned int  bc   = 0;
    unsigned char *ptr = (unsigned char *) dnnn->bits + sizeof( int );
    int j              = 0;
    
    // loop
    while( bc < ( total - sizeof( SLB_SUBHDR ) ) ){
      
      // check array size
      if( j >= max ){
	return ETB_out_of_memory;
      }
      
      // update FED count
      *n = j + 1;
      
      // ETB_fed header
      SLB_SUBHDR *hdr = &feds[j].hdr; // header of FED block
      ESP_get_slb_subhdr( (SLB_SUBHDR *)( ptr + bc ), hdr ); // copy header
      feds[j].slb  = 0; // pointer to slb record (not existent here)
      feds[j].data = ptr + bc + sizeof( SLB_SUBHDR ); // pointer to data
      
      // increase cunters
      j++; // fed counter
      bc += hdr->length; // byte counter
    }
    return ETB_ok;
  }

  // return
  return ETB_corrupted_event;
}  

//
// get list of FED blocks (see ETB)
//
int
ARTE_get_feds_DRIC( int max, ETB_fed *feds, int *n )
{
  // inits
  int rc = ETB_ok;
  *n = 0;
  
  // check fed array
  if( ( max < 1 ) || ( feds == 0 ) ){
    return ETB_out_of_memory;
  }
  
  //
  // use DEVT
  //
  EVENT_HDR *event = ARTE_get_event();
  if( event ){
    if( ( rc = ETB_get_feds( event, COMP_ID_RICH, max, feds, n ) ) != ETB_ok ){
      cout << "ARTE_get_feds_DRIC(): ERROR: fed array too small? "
	   << "max=" << max << " "
	   << "n="   << *n  << " "
	   << "rc="  << rc << endl;
    }
    if( *n > max ) *n = max;
    return rc;
  }
  //
  // use DRIC
  //
  else if( ArteTable<DRIC>::size() ){

    // Arte pointer
    ArtePointer<DRIC> dnnn( 1 );
    
    // data words in table (first word is recid)
    int total = ( ArteTable<DRIC>::size() - 1 ) * sizeof( int );
    
    // recid
    int recid = 0;
    ESP_get_int( (int *) dnnn->bits, &recid, 1 );
    
    // SLB_SUBHDR
    unsigned int  bc   = 0;
    unsigned char *ptr = (unsigned char *) dnnn->bits + sizeof( int );
    int j              = 0;
    
    // loop
    while( bc < ( total - sizeof( SLB_SUBHDR ) ) ){
      
      // check array size
      if( j >= max ){
	return ETB_out_of_memory;
      }
      
      // update FED count
      *n = j + 1;
      
      // ETB_fed header
      SLB_SUBHDR *hdr = &feds[j].hdr; // header of FED block
      ESP_get_slb_subhdr( (SLB_SUBHDR *)( ptr + bc ), hdr ); // copy header
      feds[j].slb  = 0; // pointer to slb record (not existent here)
      feds[j].data = ptr + bc + sizeof( SLB_SUBHDR ); // pointer to data
      
      // increase cunters
      j++; // fed counter
      bc += hdr->length; // byte counter
    }
    return ETB_ok;
  }

  // return
  return ETB_corrupted_event;
}  

//
// get list of FED blocks (see ETB)
//
int
ARTE_get_feds_DTRD( int max, ETB_fed *feds, int *n )
{
  // inits
  int rc = ETB_ok;
  *n = 0;
  
  // check fed array
  if( ( max < 1 ) || ( feds == 0 ) ){
    return ETB_out_of_memory;
  }
  
  //
  // use DEVT
  //
  EVENT_HDR *event = ARTE_get_event();
  if( event ){
    if( ( rc = ETB_get_feds( event, COMP_ID_TRD, max, feds, n ) ) != ETB_ok ){
      cout << "ARTE_get_feds_DTRD(): ERROR: fed array too small? "
	   << "max=" << max << " "
	   << "n="   << *n  << " "
	   << "rc="  << rc << endl;
    }
    if( *n > max ) *n = max;
    return rc;
  }
  //
  // use DTRD
  //
  else if( ArteTable<DTRD>::size() ){

    // Arte pointer
    ArtePointer<DTRD> dnnn( 1 );
    
    // data words in table (first word is recid)
    int total = ( ArteTable<DTRD>::size() - 1 ) * sizeof( int );
    
    // recid
    int recid = 0;
    ESP_get_int( (int *) dnnn->bits, &recid, 1 );
    
    // SLB_SUBHDR
    unsigned int  bc   = 0;
    unsigned char *ptr = (unsigned char *) dnnn->bits + sizeof( int );
    int j              = 0;
    
    // loop
    while( bc < ( total - sizeof( SLB_SUBHDR ) ) ){
      
      // check array size
      if( j >= max ){
	return ETB_out_of_memory;
      }
      
      // update FED count
      *n = j + 1;
      
      // ETB_fed header
      SLB_SUBHDR *hdr = &feds[j].hdr; // header of FED block
      ESP_get_slb_subhdr( (SLB_SUBHDR *)( ptr + bc ), hdr ); // copy header
      feds[j].slb  = 0; // pointer to slb record (not existent here)
      feds[j].data = ptr + bc + sizeof( SLB_SUBHDR ); // pointer to data
      
      // increase cunters
      j++; // fed counter
      bc += hdr->length; // byte counter
    }
    return ETB_ok;
  }

  // return
  return ETB_corrupted_event;
}  

//
// get list of FED blocks (see ETB)
//
int
ARTE_get_feds_DCAL( int max, ETB_fed *feds, int *n )
{
  // inits
  int rc = ETB_ok;
  *n = 0;
  
  // check fed array
  if( ( max < 1 ) || ( feds == 0 ) ){
    return ETB_out_of_memory;
  }
  
  //
  // use DEVT
  //
  EVENT_HDR *event = ARTE_get_event();
  if( event ){
    if( ( rc = ETB_get_feds( event, COMP_ID_ECAL, max, feds, n ) ) != ETB_ok ){
      cout << "ARTE_get_feds_DCAL(): ERROR: fed array too small? "
	   << "max=" << max << " "
	   << "n="   << *n  << " "
	   << "rc="  << rc << endl;
    }
    if( *n > max ) *n = max;
    return rc;
  }
  //
  // use DCAL
  //
  else if( ArteTable<DCAL>::size() ){

    // Arte pointer
    ArtePointer<DCAL> dnnn( 1 );
    
    // data words in table (first word is recid)
    int total = ( ArteTable<DCAL>::size() - 1 ) * sizeof( int );
    
    // recid
    int recid = 0;
    ESP_get_int( (int *) dnnn->bits, &recid, 1 );
    
    // SLB_SUBHDR
    unsigned int  bc   = 0;
    unsigned char *ptr = (unsigned char *) dnnn->bits + sizeof( int );
    int j              = 0;
    
    // loop
    while( bc < ( total - sizeof( SLB_SUBHDR ) ) ){
      
      // check array size
      if( j >= max ){
	return ETB_out_of_memory;
      }
      
      // update FED count
      *n = j + 1;
      
      // ETB_fed header
      SLB_SUBHDR *hdr = &feds[j].hdr; // header of FED block
      ESP_get_slb_subhdr( (SLB_SUBHDR *)( ptr + bc ), hdr ); // copy header
      feds[j].slb  = 0; // pointer to slb record (not existent here)
      feds[j].data = ptr + bc + sizeof( SLB_SUBHDR ); // pointer to data
      
      // increase cunters
      j++; // fed counter
      bc += hdr->length; // byte counter
    }
    return ETB_ok;
  }

  // return
  return ETB_corrupted_event;
}  

//
// get list of FED blocks (see ETB)
//
int
ARTE_get_feds_DMUT( int max, ETB_fed *feds, int *n )
{
  // inits
  int rc = ETB_ok;
  *n = 0;
  
  // check fed array
  if( ( max < 1 ) || ( feds == 0 ) ){
    return ETB_out_of_memory;
  }
  
  //
  // use DEVT
  //
  EVENT_HDR *event = ARTE_get_event();
  if( event ){
    if( ( rc = ETB_get_feds( event, COMP_ID_MUON, max, feds, n ) ) != ETB_ok ){
      cout << "ARTE_get_feds_DMUT(): ERROR: fed array too small? "
	   << "max=" << max << " "
	   << "n="   << *n  << " "
	   << "rc="  << rc << endl;
    }
    if( *n > max ) *n = max;
    return rc;
  }
  //
  // use DMUT
  //
  else if( ArteTable<DMUT>::size() ){

    // Arte pointer
    ArtePointer<DMUT> dnnn( 1 );
    
    // data words in table (first word is recid)
    int total = ( ArteTable<DMUT>::size() - 1 ) * sizeof( int );
    
    // recid
    int recid = 0;
    ESP_get_int( (int *) dnnn->bits, &recid, 1 );
    
    // SLB_SUBHDR
    unsigned int  bc   = 0;
    unsigned char *ptr = (unsigned char *) dnnn->bits + sizeof( int );
    int j              = 0;
    
    // loop
    while( bc < ( total - sizeof( SLB_SUBHDR ) ) ){
      
      // check array size
      if( j >= max ){
	return ETB_out_of_memory;
      }
      
      // update FED count
      *n = j + 1;
      
      // ETB_fed header
      SLB_SUBHDR *hdr = &feds[j].hdr; // header of FED block
      ESP_get_slb_subhdr( (SLB_SUBHDR *)( ptr + bc ), hdr ); // copy header
      feds[j].slb  = 0; // pointer to slb record (not existent here)
      feds[j].data = ptr + bc + sizeof( SLB_SUBHDR ); // pointer to data
      
      // increase cunters
      j++; // fed counter
      bc += hdr->length; // byte counter
    }
    return ETB_ok;
  }

  // return
  return ETB_corrupted_event;
}  

//
// get list of FED blocks (see ETB)
//
int
ARTE_get_feds_DMUX( int max, ETB_fed *feds, int *n )
{
  // inits
  int rc = ETB_ok;
  *n = 0;
  
  // check fed array
  if( ( max < 1 ) || ( feds == 0 ) ){
    return ETB_out_of_memory;
  }
  
  //
  // use DEVT
  //
  EVENT_HDR *event = ARTE_get_event();
  if( event ){
    if( ( rc = ETB_get_feds( event, COMP_ID_MPIX, max, feds, n ) ) != ETB_ok ){
      cout << "ARTE_get_feds_DMUX(): ERROR: fed array too small? "
	   << "max=" << max << " "
	   << "n="   << *n  << " "
	   << "rc="  << rc << endl;
    }
    if( *n > max ) *n = max;
    return rc;
  }
  //
  // use DMUX
  //
  else if( ArteTable<DMUX>::size() ){

    // Arte pointer
    ArtePointer<DMUX> dnnn( 1 );
    
    // data words in table (first word is recid)
    int total = ( ArteTable<DMUX>::size() - 1 ) * sizeof( int );
    
    // recid
    int recid = 0;
    ESP_get_int( (int *) dnnn->bits, &recid, 1 );
    
    // SLB_SUBHDR
    unsigned int  bc   = 0;
    unsigned char *ptr = (unsigned char *) dnnn->bits + sizeof( int );
    int j              = 0;
    
    // loop
    while( bc < ( total - sizeof( SLB_SUBHDR ) ) ){
      
      // check array size
      if( j >= max ){
	return ETB_out_of_memory;
      }
      
      // update FED count
      *n = j + 1;
      
      // ETB_fed header
      SLB_SUBHDR *hdr = &feds[j].hdr; // header of FED block
      ESP_get_slb_subhdr( (SLB_SUBHDR *)( ptr + bc ), hdr ); // copy header
      feds[j].slb  = 0; // pointer to slb record (not existent here)
      feds[j].data = ptr + bc + sizeof( SLB_SUBHDR ); // pointer to data
      
      // increase cunters
      j++; // fed counter
      bc += hdr->length; // byte counter
    }
    return ETB_ok;
  }

  // return
  return ETB_corrupted_event;
}  

//
// get list of FED blocks (see ETB)
//
int
ARTE_get_feds_DMUP( int max, ETB_fed *feds, int *n )
{
  // inits
  int rc = ETB_ok;
  *n = 0;
  
  // check fed array
  if( ( max < 1 ) || ( feds == 0 ) ){
    return ETB_out_of_memory;
  }
  
  //
  // use DEVT
  //
  EVENT_HDR *event = ARTE_get_event();
  if( event ){
    if( ( rc = ETB_get_feds( event, COMP_ID_MPAD, max, feds, n ) ) != ETB_ok ){
      cout << "ARTE_get_feds_DMUP(): ERROR: fed array too small? "
	   << "max=" << max << " "
	   << "n="   << *n  << " "
	   << "rc="  << rc << endl;
    }
    if( *n > max ) *n = max;
    return rc;
  }
  //
  // use DMUP
  //
  else if( ArteTable<DMUP>::size() ){

    // Arte pointer
    ArtePointer<DMUP> dnnn( 1 );
    
    // data words in table (first word is recid)
    int total = ( ArteTable<DMUP>::size() - 1 ) * sizeof( int );
    
    // recid
    int recid = 0;
    ESP_get_int( (int *) dnnn->bits, &recid, 1 );
    
    // SLB_SUBHDR
    unsigned int  bc   = 0;
    unsigned char *ptr = (unsigned char *) dnnn->bits + sizeof( int );
    int j              = 0;
    
    // loop
    while( bc < ( total - sizeof( SLB_SUBHDR ) ) ){
      
      // check array size
      if( j >= max ){
	return ETB_out_of_memory;
      }
      
      // update FED count
      *n = j + 1;
      
      // ETB_fed header
      SLB_SUBHDR *hdr = &feds[j].hdr; // header of FED block
      ESP_get_slb_subhdr( (SLB_SUBHDR *)( ptr + bc ), hdr ); // copy header
      feds[j].slb  = 0; // pointer to slb record (not existent here)
      feds[j].data = ptr + bc + sizeof( SLB_SUBHDR ); // pointer to data
      
      // increase cunters
      j++; // fed counter
      bc += hdr->length; // byte counter
    }
    return ETB_ok;
  }

  // return
  return ETB_corrupted_event;
}  
