/***********************************************************************
*-- Author :    Siegmund Nowak
*
*  This is the CERN user routine called at the end of each tracking step
*  Here actions are collected which are specific for HERA-B but general
*  for all HERA-B users. The individual user routine is called USSTEP 
*  The actions called from GUSTEP are:
*  - drawing of tracks
*  - filling of temporary MIMP/MCAL banks for sensitive detectors
*  - storing of new generated particles, restoring of STAK particles 
*  - calling of USSTEP
*  - limiting of spiraling particles in the magnet
*  - setting length and flag parameters for ending tracks
*  - limiting of endless loops on boundaries for flat tracks 
***********************************************************************/
#include <string.h>
#include <math.h>
#include <vector.h>
#include <stdio.h>
#include <ctype.h>
#include "gean/gckine.h"
#include "gean/gcking.h"
#include "gean/gcflag.h"
#include "gean/gctrak.h"
#include "gean/gcvolu.h"
#include "gean/gctmed.h"
#include "gean/gcsets.h"
#include "gean/gcnum.h"
#include "gean/gcpush.h"
#include "gean/ZGMOTH.h"
#include "gean/ZGESL.h"
#include "gean/ZGDETV.h"
#include "gean/zgkine.hh"
#include "gean/ZGTRDE.h"
#include "gean/gcchan.h"
#include "arte/arte.hh"
#include "arte/ArteFlag.h"
#include "arte/ArteUtil.h"
#include "arte/ArteMessage.hh"

extern "C" {
   void gustep_();               
   void gsxyz_();                
   void gdcxyz_();
   void trade_();
   void gsking_(int *);          
   void ghstak_();
   void muturb_();               
   void usstep_();
}
   void crhibk(int);            
   int  gspiral();
   static int SameVolume=0, SameTrack, 
           OldCopyNr = 0, OldItra =0, OldIstak=0, OldNstep = -1;
   static char OldVolName[] = "    ";

void gustep_() {
   extern vector<hgmtra> *hg_mtra; 
   char cname[4], s2name[4], sdname[4];

// get the system name s2name and the component name cname
   memcpy(cname,&gcvolu_.names[gcvolu_.nlevel-1],4);
   for(int i =0; i<4; i++) cname[i] = toupper(cname[i]);  

   gcvolu_.nlevel > 1 ? 
         memcpy(s2name,&gcvolu_.names[1],4): memcpy(s2name,"    ",4);
   for(int i =0; i<4; i++) s2name[i] = toupper(s2name[i]);

//  Check whether it is in the same volume,    
   if(gctrak_.inwvol == 1) {
      SameVolume = 0;
      if(!strncmp(cname,OldVolName,4) && gcvolu_.number[gcvolu_.nlevel-1] 
                       ==  OldCopyNr)         SameVolume = gcchan_.lsamvl;
      else {
         strncpy(OldVolName,cname,4);
         OldCopyNr = gcvolu_.number[gcvolu_.nlevel-1];
      } 
   }

//  Check whether it is the same track
   if(!gctrak_.nstep) {
     SameTrack = 0;  
     OldItra = gckine_.itra;
     OldIstak = gckine_.istak;
     OldNstep = 0;
   }
   else if(gctrak_.nstep >= OldNstep) { 
     if(OldItra == gckine_.itra && OldIstak == gckine_.istak) SameTrack =1;    
     OldNstep = gctrak_.nstep;
   }

   int lsv = 0;
   if(gctrak_.inwvol == 1 || gctrak_.inwvol == 2 || gctrak_.inwvol == 3)
     lsv = SameVolume & SameTrack;

// get some track values
   float  abs_charge = fabs(gckine_.charge); 
   vector<hgmtra>::iterator mt = hg_mtra->begin() + gckine_.itra -1;

// set userstop
   int userstop = 0; if(gctrak_.istop) userstop =1;

// select whether secondary particles are put only in stack (jspec != 0)
   int jspec = 0;
   if (!strncmp(s2name,"MAGN",4)) {
      if(!strncmp("SCRP",cname,4) || !strncmp("POLS",cname,4) || 
         !strncmp("POLC",cname,4) ) jspec = 1; 
   }
   else if((gctrak_.vect[2]-zgmoth_.ZZERO) >= zgmoth_.ZSTACK) {
      jspec =2; 
      if(!strncmp(s2name,"MUON",4)) {
        gcvolu_.nlevel > 2 ? 
        memcpy(sdname,&gcvolu_.names[2],4): memcpy(sdname,"    ",4);
        for(int i =0; i<4; i++) sdname[i] = toupper(sdname[i]);
        if(!strncmp("MUD1",sdname,4) || !strncmp("MUD1",sdname,4) ||
           !strncmp("MUD3",sdname,4) || !strncmp("MUD4",sdname,4) ||
           !strncmp("MXD1",sdname,4) || !strncmp("MXD2",sdname,4) ||
           !strncmp("MXD3",sdname,4) || !strncmp("MXD4",sdname,4)) jspec = 0;
      }
   }

// draw tracks if wanted
   if(!arteflag_.ibatch && gcflag_.iswit[0] && abs_charge >= 1.e-2) {
       gsxyz_();   gdcxyz_();
   }

// make a calorimeter endpoint mtcp  at the entry of the calorimeter 
   if(!strncmp(s2name,"CALW",4)) {
      if(!gckine_.istak && !mt->ep ) {
         if(!strncmp("CAY1",cname,4) || !strncmp("CAY2",cname,4) || 
            !strncmp("CAY3",cname,4) || !strncmp("CASC",cname,4) ||
	    !strncmp("CAPB",cname,4) || !strncmp("CATU",cname,4) ) crhibk(5);
      }
   }    

// stop the particle if it enters in volume "kill"     
   else if(!strncmp(cname,"KILL",4)) {  
        userstop = 2;     goto L60;
   }

// call special code for trd-simulation
   else if(!strncmp(s2name,"TRSL",4) || gckine_.ipart == 48 ) { 
        if(trdsw_.itrswt == 1) trade_(); 
   }

// switch on fast electron tracing by means of mututb
   else if(!strncmp(s2name,"MUON",4)) {
        extern int muon_turbfl;
        if(muon_turbfl == 1) muturb_();
   }

// stop tracks to avoid spirals in magnetic field
   else if(gctmed_.ifield &&  abs_charge >= .1 
        && gctrak_.vect[6] < 1. && !mt->nhit) userstop = gspiral(); 


// avoid long loops for flat tracks on boundaries (temporary)
   if(fabs(gctrak_.vect[5]) <= .04) {
      static int nust,vflag; 
      if (gctrak_.nstep == 0) nust = 0;
      if(gctrak_.inwvol == 2 && !userstop) vflag = 1;
      if( gctrak_.inwvol == 1 && vflag == 1 &&  SameTrack) { 
         if(SameVolume) {
             nust = nust++;
             if(nust >= 20) {
                gctrak_.vect[0] = gctrak_.vect[0] + .01 *gctrak_.vect[3]; 
                gctrak_.vect[1] = gctrak_.vect[1] + .01 *gctrak_.vect[4]; 
                gctrak_.vect[2] = gctrak_.vect[2] + .01 *gctrak_.vect[5]; 
             }
             if(nust >=30) userstop = 3;
         }
         else nust = 0;
         vflag=0;
      }  
   } 

// mimps will be written out
   if(arteflag_.mimpsw) {
      if(!gcsets_.idtype) {}
      else if(gcsets_.idtype == LGESL_ECA) {
             if(gctrak_.destep >0. ) crhibk(2);
      }
      else {  
         int wirehit = 0;
         if(!lsv && gctrak_.inwvol == 1 && abs_charge > 1.e-2) wirehit =1; 

         if      (gcsets_.idtype == LGESL_SIL && wirehit) crhibk(6);
         else if (gcsets_.idtype == LGESL_OTR && wirehit) crhibk(8);
         else if (gcsets_.idtype == LGESL_MSG && wirehit) crhibk(6);
         else if (gcsets_.idtype == LGESL_TRD && wirehit) crhibk(7);
         else if (gcsets_.idtype == LGESL_RIC && wirehit) crhibk(1);
         else if (gcsets_.idtype == LGESL_HOD3) 
             {if (gctrak_.destep>0)  crhibk(3); }
         else if (wirehit)  crhibk(1);
         if(gcsets_.idtype == LGESL_TRD && trdsw_.itrswt == 1 
                                     && gctrak_.destep > 0. ){
	     if(abs_charge > 1.e-2 || gckine_.ipart == 48) crhibk(4);
         }
      }
   }
      
// store newly generated particles in jkine/jvertx if they are outside
// the volumes idnam and in stak only if they are inside
   if(gcking_.ngkine > 0) {
      int ifl0 = 1;   if(jspec) ifl0 = 0; 
      if(ifl0 == 1 && 
               (gcnum_.ntrack + gcpush_.npkine > 64000 
             || gcnum_.ntrack + gcking_.ngkine > 64000)) {
	 if(gcflag_.ieotri !=0) {  
	    arteMsg << "event skipped: more than 64 k geant tracks";
            arteMsg.warn();
         }
         gcflag_.ieotri = 1;
         arteutil_.eventout = 0;
         return;
      } 
      for (int i=0; i<gcking_.ngkine; i++)  gcking_.iflgk[i] = ifl0;  
      int i0 = 0;     gsking_(&i0);  
   }  

// restore stak particles in jkine/jvertx banks again, if they come
// out of volumes idnam
   if(gckine_.istak != 0 && gctrak_.inwvol == 1) {
      if(!strncmp(cname,"TR",2) || gckine_.ipart == 48) {}
      else if(!jspec) { 
        if(gcnum_.ntrack + gcpush_.nckine > 64000 ) {
	   if(gcflag_.ieotri !=0) {
	      arteMsg << "event skipped: more than 64 k geant tracks";
              arteMsg.warn();
           } 
           gcflag_.ieotri = 1;
           arteutil_.eventout = 0;
           return;
        }
        ghstak_();  userstop = 4; 
      } 
   }

L60:   // call user subroutine
    usstep_();
    if(gctrak_.istop && userstop !=1 ) userstop = 5;
    if(!gctrak_.istop && userstop) gctrak_.istop = 1; 

//  stopping tracks: give length and flag  
    if(gctrak_.istop) {               
       if(!gckine_.istak && arteflag_.mimpsw) {
          mt->len = gctrak_.sleng;

          // leaving the detector  
          if(gctrak_.inwvol == 3 || userstop == 2) {
             mt->flag = 5;   return; 
          }

          // spiraling cut
          if(userstop < 0) {
	     mt->flag = 7;   return;
          }

          // user stop  
          if(userstop == 5){
             mt->flag = 8;   return;
          }

          // stop due to physical processes
          for(int i=0; i<gctrak_.nmec; i++) {

             if(gctrak_.lmec[i] == 5) {         // decay
                mt->flag = 3;   return; 
             }

             if((gctrak_.lmec[i] >= 6 && gctrak_.lmec[i]<= 8) 
                   ||   gctrak_.lmec[i] ==11 || gctrak_.lmec[i]== 12) {
                mt->flag = 4;   return;         //  interaction
              }

             if(gctrak_.lmec[i] == 30) {        // energy cutoff
                 mt->flag = 6;   return; 
             }
          }
       }
    }
    
}



