// Defines abstract class HPT_DAQ_Abstract.
// The programmer should derive his own class from HPT_DAQ_Abstract and define
// at least the virtual function run(short int BX);
//
// E.g.:
//   class HPT_DAQ : public HPT_DAQ_Abstract {
//   public:
//     HPT_DAQ() {
//       ... // initialize your NTUPLEs etc.
//     }
//     virtual void run(short int BX) {
//       ... // fill your NTUPLEs etc.
//     }
//     ~HPT_DAQ() {
//       ... // delete everything what you need etc.
//     }
//   };
// Then he can define some static object of type HPT_DAQ, e.g.,
// somewhere in the *.C file:
//   static HPT_DAQ hpt_daq;
// Thus the initialization and deletion of the hpt_daq is made in the beginning
// and in the end of the program.
// Finally hpt_daq.process_event(BX, fed_numbers, data, n_feds);
// should be called somewhere from the
// loop over events, e.g. from 
//   int dmonuser_event(EVENT_HDR *event, int length);
// Generally this function makes some checks that the data is Ok and then
// calls hpt_daq.process_event(...).
// The parameters in process_event(BX, fed_numbers, data, n_feds) means the
// following:
// fed_numbers - an array of integers with FED numbers;
// data[fed_number_i] points to the data for FED#i;
// n_feds - total number of high-pt FEDs.
//
// HPT_DAQ_Abstract::process_event is fixed in the sense that it fills the
// data using the standard algorithm and then calls virtual function run().
// The programmer normally should not touch HPT_DAQ_Abstract::process_event.
//
// If you'll need to read the high-pt data: geometry, cabling or hit-map
// not from ./hpt_setup.dat, ./hpt_cables.dat, ./hpt_hit_map.dat
// respectively, but from some other external files (e.g. hpt_setup.dat
// from standard HERA-B place where all geometry files are stored)
// then you can provide the appropriate arguments to
//  HPT_DAQ_Abstract(const string& hpt_setup_file   = *((string*)NULL),
//		     const string& hpt_cables_file  = *((string*)NULL),
//		     const string& hpt_hit_map_file = *((string*)NULL));
// e.g.:
// HPT_DAQ() : HPT_DAQ_Abstract("some_other_setup_file",
//                              "hpt_cables.dat",
//                              "hpt_hit_map.dat") { ... }
// If you do not need to open hpt_hit_map file, supply an empty string ""
// as the argument (e.g. HPT_DAQ x("hpt_setup.dat", "hpt_cables.dat", "");)
//

// Author: Vladislav Balagura (balagura@hera-b.desy.de, 03.11.1999)

#ifndef _HPT_DAQ_Abstract_hh
#define _HPT_DAQ_Abstract_hh 1

#include "ptset/high_pt.hh"
#include "ptset/detector.hh"
#include "ptset/FEDs.hh"

class HPT_DAQ_Abstract {
public:
  Cables<HPT_Setup::IChamber>* cables;
  FEDs<HPT_Setup::IChamber>*   feds;

  HPT_DAQ_Abstract(const string& hpt_geo_file      = *((string*)NULL),
		   const string& hpt_cables_file   = *((string*)NULL),
		   const string& hpt_hit_map_file  = *((string*)NULL));
  virtual ~HPT_DAQ_Abstract();
  virtual void run(short int BX) = 0;

  void process_event(short int BX, unsigned int* fed_numbers,
		     unsigned char** data, int n_feds);
};

#endif
