// 
//  Author:  Rainer Mankel                     Created:  29-Sep-98
//           Humboldt University Berlin        Modified: 29-Mar-99
//  Email:   mankel@ifh.de                     HERA-B Collaboration
//
// Modifications:
// - Introduced list of active walls   14-Aug-00 Wouter Hulsbergen 
// - use of all gwals in the vds                 17/05/01 A. Spiridonov
// - Removed trd/hipt walls                      04/07/01 Wouter
// - Added routine getActiveVolumes (in .hh)
// - Increased mult.scat. fudge factor for main  20/07/01 Wouter 
//   tracker  

#include <iostream>
#include <stdio.h>
#include <algo.h>
#include "arte/arte.hh"
#include "arte/ArteMessage.hh"
#include <arte/artefunc.h>
#include "rang/Geom.hh"
#include "rang/Gede.hh"
#include "rang/Gwal.hh"
#include "rang/sortzv.hh"
#include "gean/ZGESL.h"

#define GWALL_SELECT
#undef GWALL_NOWALLS  
#undef GWALL_NOVDS

vector<rangerVolume*> rangerGeom::volumes;
vector<vector<rangerVolume*>::iterator> rangerGeom::vGede;
vector<vector<rangerVolume*>::iterator> rangerGeom::vGwal;

vector<rangerVolume*> rangerGeom::activeVolumes;
int rangerGeom::firstPlane=0;
vector<vector<rangerVolume*>::const_iterator> rangerGeom::planeToUpstrActive;

void rangerGeom::update() {
  reset();
  build();
  gsort();
  arteMsg << "rangerGeom::update: geometry table (re)built" 
	  << " with " << volumes.size() 
	  << " (" << activeVolumes.size() << ") volumes" << endl ;
  arteMsg.warn();
}

void rangerGeom::condUpdate() {
  unsigned int nVol = ArteTable<GEDE>::nrow() + ArteTable<GWAL>::nrow();
  if (nVol != volumes.size()) update();
}

void rangerGeom::reset() {
  for (vector<rangerVolume*>::iterator vIt=volumes.begin();
       vIt != volumes.end(); ++vIt) delete (*vIt);
  volumes.erase(volumes.begin(),volumes.end());
  vGede.erase(vGede.begin(),vGede.end());
  vGwal.erase(vGwal.begin(),vGwal.end());
  activeVolumes.clear() ;
  planeToUpstrActive.clear() ;
}

void rangerGeom::build() {
  volumes.clear() ;
  volumes.reserve(ArteTable<GEDE>::nrow()+ArteTable<GWAL>::nrow());
  
  for ( ArteTable<GEDE>::iterator gede = ArteTable<GEDE>::begin();
	gede != ArteTable<GEDE>::end(); ++gede ) {
    rangerGede* Gede = new rangerGede( gede ); // allocate new Gede
    rangerGeWa* GeWa = Gede;
    rangerVolume* vol = rangerVolume::makeVolume( GeWa );
    volumes.push_back(vol);
  }

  for ( ArteTable<GWAL>::iterator gwal = ArteTable<GWAL>::begin();
	gwal != ArteTable<GWAL>::end(); ++gwal ) {
    rangerGwal* Gwal = new rangerGwal( gwal ); // allocate new Gwal
    rangerGeWa* GeWa = Gwal;
    rangerVolume* vol = rangerVolume::makeVolume( GeWa );
    volumes.push_back(vol);
  }
}

inline bool 
compareVolumesZ(const rangerVolume* lhs, const rangerVolume* rhs)
{
  double refzlhs = 1e-8*lhs->plNumber()+lhs->zMean() ; 
  double refzrhs = 1e-8*rhs->plNumber()+rhs->zMean() ; 
  return refzlhs<refzrhs ;
}

void rangerGeom::gsort() {

  // sort the volumes
  //sort(volumes.begin(),volumes.end(), compareCall( &rangerVolume::zMean ) );
  sort(volumes.begin(),volumes.end(),&compareVolumesZ) ;

  // initialize the vGede and vGwal pointers
  vGede.reserve(ArtePointer<GEDE>::nrow());
  for ( ArteTable<GEDE>::iterator gede = ArteTable<GEDE>::begin();
	gede != ArteTable<GEDE>::end(); ++gede ) {
      vGede.push_back(volumes.end());
  }
  vGwal.reserve(ArtePointer<GWAL>::nrow());
  for ( ArteTable<GWAL>::iterator gwal = ArteTable<GWAL>::begin();
	gwal != ArteTable<GWAL>::end(); ++gwal ) {
      vGwal.push_back(volumes.end());
  }
  for (vector<rangerVolume*>::iterator vi=volumes.begin();
       vi != volumes.end(); ++vi) {
    if ((*vi)->plNumber() > 0) {
      if ((*vi)->plNumber() > (int) ArtePointer<GEDE>::nrow()) {
	cout << "warning: gede::nrow=" << ArtePointer<GEDE>::nrow()
	     << " pln=" << (*vi)->plNumber() << endl;
      }
      vGede[(*vi)->plNumber()-1] = vi;
    }
    else {
      if ((*vi)->plNumber() < 0) {
	if (-(*vi)->plNumber() > (int) ArtePointer<GWAL>::nrow()) {
	  cout << "warning: gwal::nrow=" << ArtePointer<GWAL>::nrow()
	       << " pln=" << (*vi)->plNumber() << endl;
	}
	vGwal[-(*vi)->plNumber()-1] = vi;
      }
      else {
	cout << "plNumber zero!" << endl;
      }
    }
  }

  // create vector of active volumes
  activeVolumes.clear() ;
  // this reserve is necessary, otherwise pointers are not correctly
  // set. it is a bit a waste of space though ...
  activeVolumes.reserve(volumes.size()) ;
  // these are pointers to the first active volume with a lower z value
  planeToUpstrActive.clear() ;
  firstPlane = -ArteTable<GWAL>::size() ;
  int nplanes = ArteTable<GWAL>::size()+ArteTable<GEDE>::size()+1 ;
  planeToUpstrActive.resize(nplanes,activeVolumes.begin()) ;

  for(vector<rangerVolume*>::iterator vi=volumes.begin();
      vi != volumes.end(); ++vi) {
    int plane = (*vi)->plNumber() ;

    // set pointer of this plane to first active volume with lower z
    if(!activeVolumes.empty())
      planeToUpstrActive[plane-firstPlane] = activeVolumes.end() - 1 ;
    else
      planeToUpstrActive[plane-firstPlane] = activeVolumes.begin() ;

    // here the code to de/activate volume
    bool active = true;
#ifdef GWALL_SELECT
    float z = (*vi)->zMean() ;
    rangerGeWa* gewa = const_cast<rangerGeWa*> ((*vi)->getGeWa()) ;
    int cmp = 0 ;
    if( gewa->getGede().index() ) cmp = gewa->getGede()->gesl->cmp ;
    if( gewa->getGwal().index() ) cmp = gewa->getGwal()->gesl->cmp ;
    
    // remove all objects with very low ms:
    // (there are no gedes below 1.e-7)
    active = active && !(plane<0 && gewa->ms() < 1.e-8) ;
    
    // remove the g10 strips and end-pieces in the otr (they don't
    // need correction)
    active = active && !(cmp==LGESL_OTR && plane<0 && gewa->ms()<1.e-7) ;    
    // the remaining walls in the otr are coverplates and end-caps.

    // remove all walls in otr itr and replace by fudge factors
    if( cmp==LGESL_OTR ) {
      const float otrfudge = 4. ;
      if(plane<0 ) active = false ;
      else (*vi)->setMs((*vi)->ms()*otrfudge) ;
    }
    
    // remove all walls in itr and replace by fudge factors
    if( cmp==LGESL_ITR ) { 
      const float itrfudge = 3. ;
      if(plane<0) active = false ;
      else (*vi)->setMs((*vi)->ms()*itrfudge) ;
    }
    
    // remove trd and hi-pt volumes (saves about 10% CPU)
    if( cmp==LGESL_TRD || cmp==LGESL_PTC) active = false ;
    
#ifdef GWALL_NOVDS
    const float vdsfudge = 1.2 ; 
    bool isvds = z<205. ;
    if(active && isvds) {
      if(plane>0) (*vi)->setMs((*vi)->ms()*vdsfudge) ;
      else        active = false ;
    }
#endif

#endif 
    // add volume to active list
    if(active) activeVolumes.push_back(*vi) ;
  }
}
