// 
//  Author:  Rainer Mankel                     Created:  29-Sep-98
//           Humboldt University Berlin        Modified:  
//  Email:   mankel@ifh.de                     HERA-B Collaboration
// 
// Modifications:
// - Corrected smoothing of residuals and chi2
// - Introduced ::filterOhneAbsorb, ::predictGlobal and ::absorbGlobal
// - Faster matrix manipulation in ::absorb
//                                         14-Aug-00 Wouter Hulsbergen 


#include <iostream>
#include <math.h>
#include "arte/arte.hh"
#include "arte/ArteMessage.hh"
#include "rang/State.hh"
#include "rang/HitBase.hh"
#include "rang/ScatThick.hh"
#include "rang/Geom.hh"
#include "rang/zfpar.hh"
#include "rang/utfeld.hh"
#include "rang/RadEl.hh"
#include "CLHEP/Matrix/DiagMatrix.h"

//-- Constructor
rangerHitBase::rangerHitBase() 
  :rangerNode(), m(0), V(0), H(0), r(0), R(0), chiSq(0)
{
  cout << "This is the rangerHitBase default constructor" << endl;
}

rangerHitBase::rangerHitBase(ArtePointer<RHIT> rhit, rangerFitMode fm) 
  : rangerNode(rhit,fm), m(0), V(0), H(0), r(0), R(0), chiSq(0)
{
#ifdef FULL
  cout << "This is the rangerHitBase(rhit) constructor, hitb=" 
       << rhit->hitb << endl;
#endif
}

rangerHitBase::rangerHitBase(ArtePointer<HITB> hitb, rangerFitMode fm) 
  : rangerNode(hitb,fm), m(0), V(0), H(0), r(0), R(0), chiSq(0)
{
#ifdef FULL
    cout << "This is the rangerHitBase(hitb) constructor, hitb=" 
	 << hitb << endl;
#endif
}

rangerHitBase::rangerHitBase(ArtePointer<RCCL> rccl,int charge,rangerFitMode fm) 
  : rangerNode(rccl,charge,fm), m(0), V(0), H(0), r(0), R(0), chiSq(0)
{
#ifdef FULL
  cout << "This is the rangerHitBase(rccl) constructor, rccl=" 
       << rccl << endl;
#endif
}

rangerHitBase::rangerHitBase(ArtePointer<RSEG> rseg,rangerFitMode fm,unsigned char side) 
  : rangerNode(rseg,fm,side), m(0), V(0), H(0), r(0), R(0), chiSq(0)
{
#ifdef FULL
  cout << "This is the rangerHitBase(rseg) constructor, rseg=" 
       << rseg << endl;
#endif
}

rangerHitBase::~rangerHitBase(){
  // delete V ;
  // delete m ;
  delete H ;
  delete r ;
  delete R ;
}

void rangerHitBase::processNoise(rangerParticleType pType) {

  rangerScatThin scat( ms(), state->qPEst(),
		       state->tx(),
		       state->ty());

  state->applyScat( scat );

  if (pType == ELECTRON) {
    rangerRadEl rad( radiationThickness(), predictedState->qP() );
    predictedState->applyRadEl( rad );
  }

  state->applyDeDx( radiationThickness() );

}



void rangerHitBase::predProcessNoise(rangerParticleType pType) {

  rangerScatThick scat( ms(), predictedState->qPEst(),
			predictedState->tx(),
			predictedState->ty(), -dz() );

  predictedState->applyScat( scat );
  
  if (pType == ELECTRON) {
    rangerRadEl rad( radiationThickness(), predictedState->qP() );
    predictedState->applyRadEl( rad );
  }

  predictedState->applyDeDx( radiationThickEff() );

#ifdef FULL
  cout << "rangerHitBase::processNoise: correction applied, cov matrix is now "
       << predictedState->cov() << endl;
#endif
}


int rangerHitBase::filter() { 

#ifdef FULL
  cout << "rangerHitBase::filter(): this is the node at z=" << z << endl;
  list();
#endif
  if (! predictedState) {
    predictedState = rangerState::makeState( state->fitMode() );
  }
  else {
  }

  *(predictedState) = *state;      // copy initialization into prediction
  predictedState->covInit();      // initialize cov matrix
                                  // no transport matrix needed
  int absorbStatus = absorb();
#ifdef FULL
  cout << "rangerHitBase::filter(): absorb returns status" 
       << absorbStatus << endl;
#endif
  return absorbStatus;
}


int rangerHitBase::filter(rangerNode* n,rangerParticleType pType) {

// for the transport matrix, the predicted state & the projection matrix
// allocation is done dynamically: 
//   if they do not exist, they are created,
//   if they do exist, they are reused.
// on destruction of the Node, both are destroyed

#ifdef FULL
  cout << "rangerHitBase::filter(...): this is the node at z=" << z << endl;
  cout << "                         call prediction" << endl;
  if (predictedState) cout << "      predictedState OK";
  else {
    cout << "      predictedState " << predictedState;
  }
  if (transportMatrix) cout << "     transportMatrix OK " << endl;
  else {
    cout << "    transportMatrix " << transportMatrix << endl;
  }
  list();
#endif

  int returnStatus = n->predict(z,predictedState,transportMatrix);

  if (returnStatus < 0) return returnStatus;

#ifdef FULL
  cout << "rangerHitBase::filter(...): the predicted state is" 
       << predictedState->vec() << endl
       << predictedState->cov() << endl
       << "                         the transport matrix is "
       << *transportMatrix << endl;
#endif
  
  if (plane() != n->plane())    predProcessNoise(pType);
  int absorbStatus = absorb();

#ifdef FULL
  cout << "rangerHitBase::filter(...): the filtered state is"
       << state->vec() << endl;
  list();
  if (state->tx() >0 || state->tx() <=0);
  else {
    cout << "rangerHitBase::filter(...): invalid tx parameter" << endl;
    exit(1);
  }
#endif

  return absorbStatus;
}
  


int rangerHitBase::absorb() { 
  if (! H) {
    H = new HepMatrix(dim(),predictedState->dim());
    if (! H) {
      cout << "rangerHitBase::absorb(): matrix allocation failed, H=" << H 
	   << endl;
      exit(1);
    }
  }

  if (!r)  r = new HepVector( dim() );   // residual


#ifdef FULL
  cout << "rangerHitBase::absorb(): calling State::project(...)" << endl ;
#endif

  predictedState->project(*r,(*H),this);
#ifdef FULL
  cout << "HitBase::absorb: back from project, matrix=" << endl;
  cout << *H << endl;
  cout << "The residual is " << *r << endl;
#endif

  int invertStatus = 0;

#ifdef FULL
  cout << "HitBase::absorb: start calculation of K matrix" << endl;
  cout << "The predicted state cov matrix is " << predictedState->cov()
       << endl;
  cout << "The H matrix is " << (*H) << endl;
  cout << "The V matrix is " << (*V) << endl;
#endif
  
  //HepMatrix K( predictedState->cov() * H->T() 
  //	       * ((*V) + predictedState->cov().similarity(*H)).inverse(invertStatus) );
  HepMatrix CH = predictedState->cov() * H->T() ;
  HepSymMatrix HCH ; HCH.assign((*H) * CH) ;
  HepMatrix K( CH * ((*V) + HCH).inverse(invertStatus) );

#ifdef FULL
cout << "K matrix after assignment " << K << endl;
#endif

#ifdef FULL
  if (dim() == 2) {
    cout << "<any key...>" << endl;
    char cc = cin.get();
  }
#endif

  if (invertStatus != 0) {
    arteMsg << "HitBase::absorb: invert status " << invertStatus << endl;
    cout << "HitBase::absorb: invert status " << invertStatus << endl;
    static int id1;
    arteMsg.allp(id1,1);
    
    return -10;
  }

  state->vec() = predictedState->vec() + K * (*r);

  if ( fabs(state->x()) < 700 &&   // check validity of rkmain input
       fabs(state->y()) < 700 &&
       fabs(state->tx()) < 10   &&
       fabs(state->ty()) < 10   &&
       fabs(state->qP()) < 100 ) ;
  else {
#ifdef FULLSECOND
    cout << "HitBase::absorb: invalid state="
	 << state->vec() 
	 << " at z=" << state->z() << endl;
    if (predictedState) {
      cout << "The prediction was state=" << predictedState->vec()
	   << " cov=" << predictedState->cov() << endl;
    }
    else {
      cout << "There is no predicted state!" << endl;
    }
    //char cc = cin.get();
#endif
    arteMsg << "HitBase::absorb: second kind error";
    //cout << "HitBase::absorb: second kind error: res = " << (*r) ;
    static int id2;
    arteMsg.allp(id2,1);
    
    return -10;
  }

#ifdef FULL
  cout << " state->vec after assignment =" << state->vec() << endl;
#endif

  static HepDiagMatrix unit(0) ;
  if(unit.num_row()!=predictedState->dim()) 
    unit = HepDiagMatrix(predictedState->dim(),1) ;

  state->cov().assign( (unit - K * (*H)) * predictedState->cov() );

  // Wouter: This is supposed to be more stable (my mail to Rainer of
  // 20 Apr 99), but I could not find any difference:
  //state->cov().assign(predictedState->cov().similarity(unit - K * (*H)) +
  //		      (*V).similarity(K)) ;
  
  HepMatrix UminHK = HepDiagMatrix(dim(),1) - (*H) * K ;
  (*r) = UminHK * (*r);
  if (!R) R = new HepSymMatrix( dim() );
  (*R).assign( UminHK * (*V) );

  //    HepMatrix unitM(dim(),dim(),1);
  //   (*r) = (unitM - (*H) * K) * (*r);
  //   if (!R) R = new HepSymMatrix( dim() );
  //   (*R).assign( (unitM - (*H) * K) * (*V) );


  invertStatus = 0;
  double ch2 = ((R->inverse(invertStatus)).similarity(*r)) ;

  if (invertStatus >= 0) {
    chiSq = (ch2>0) ? ch2 : 0;
  }
  else {
    cout << "HitBase::absorb: invert error in chiSq calculation " << endl;
    chiSq = 0;
  }

#ifdef FULL
  cout << "Filtering done" << endl << endl << endl << endl;
#endif
  
  return 0;
}

void rangerHitBase::residUpdate() {
  // update the residual
  HepMatrix HNew(dim(),predictedState->dim());
  //Wouter: This matrix is not needed whatsoever. We should rewrite
  //Hit::Project such that we can pass a zero !
  state->project(*r,HNew,this);
}


void rangerHitBase::residUpdSmooth(rangerNode* n,HepMatrix& A) {
  // update the residual with smoothing
  HepMatrix HNew(dim(),predictedState->dim());
  state->project(*r,HNew,this);
  
  (*R) = (*R) - 
    ( (n->getState().cov() - n->getPrediction()->cov()).similarity(A) ).similarity(*H);

  // Wouter: According to a discussion with Rainer we should NOT smooth the chi2
  // The two valid definitions of chi2 are:
  // * for non smoothed residuals: res * 1/R * res (the `Kalman' definition)
  // * for smoothed residuals:     res * 1/V * res (the `global' definition)
  //
  //   if ( (*R)(1,1) <= 0) {
  //     cout << " rangerHitBase::residUpdate: R(1,1)=" << (*R)(1,1) << endl;
  //     chiSq = 0;
  //   }
  //   else {
  //     int invertStatus = 0;
  //     double ch2 =  (r->T() 
  // 		   * R->inverse(invertStatus) * (*r))(1,1);
  //     if (invertStatus >= 0) {
  //       chiSq = (ch2>0) ? ch2 : 0;
  //     }
  //     else {
  //       cout << " rangerHitBase::residUpdate: R invert failed" << endl;
  //       exit(1);
  //     }
  //   }
}

int rangerHitBase::filterOhneAbsorb(const rangerState* baseState, 
				    const double qPExp)
{ 
  // Added by Wouter. This is only the prediction step of the
  // filter. It can be used in the calibration for calculating the
  // residual of hits that are not used in the fit.
  
  // To make this routine also usefull in cases where there is no base
  // node, but only a base state, there are two versions. Only in case
  // there is a base node, the `process noise' is correctly processed.

  int returnStatus;
  if (!predictedState) 
    predictedState = rangerState::makeState( state->fitMode() );
  if(!H) H = new HepMatrix(dim(),predictedState->dim());
  if(!r) r = new HepVector( dim() );
  if(!R) R = new HepSymMatrix( dim() ) ;
  
  if(!predictedState || !H || !r || !R) {
    cout << "RangerHitBase::filterOhneAbsorb(..): memory allocation error."
	 << endl ;
    returnStatus = -1 ;
  } else {
    returnStatus = baseState->predict(z,qPExp,predictedState,transportMatrix) ;
    if(returnStatus < 0) {
      cout << "rangerHitBase::filterOhneAbsorb(..): predict returns returnStatus=" 
	   <<  returnStatus << endl;
    } else { 
      predictedState->project(*r,*H,this);
      // To get the right lrsign
      updInternal() ;
      (*R).assign(*V + predictedState->cov().similarity(*H)) ;
    }
  }
  return returnStatus ;
}

int rangerHitBase::filterOhneAbsorb(rangerNode* n,rangerParticleType pType) 
{ 
  if (!predictedState) 
    predictedState = rangerState::makeState( state->fitMode() );
  if(plane() != n->plane()) predProcessNoise(pType);
  return filterOhneAbsorb(&(n->getState()),n->getQPExp()) ;
}

int rangerHitBase::projectGlobal(const rangerState* globstate)
{ 
  if (!state)
    state = rangerState::makeState( globstate->fitMode() );
  
  globstate->predict(z,globstate->qP(),state,transportMatrix);

#ifdef FULL
  cout << "rangerHitBase::projectGlobal: input state"
       << globstate->z() << " " << globstate->vec() << " "
       << "output state: " 
       << state->z() << " " << state->vec() ;
#endif  

  if(!H) H = new HepMatrix(dim(),state->dim());
  if(!r) r = new HepVector( dim() );   // residual
  if(!R) R = new HepSymMatrix( dim() ) ;
  state->project(*r,(*H),this);
  *R = *V - state->cov().similarity(*H) ;
    
  int ierr;
  chiSq = (*V).inverse(ierr).similarity(*r) ;
  return 0;
}

int rangerHitBase::absorbGlobal(HepVector* hdChi2dQ,
				HepSymMatrix* hd2Chi2dQ2) const
{
  int ierr;
  static HepSymMatrix W ;
  static HepMatrix HF ;
  W  = (*V).inverse(ierr) ;
  HF = (*H) * (*transportMatrix) ;
  (*hdChi2dQ)   += - HF.T() * W * (*r) ;
  (*hd2Chi2dQ2) += W.similarityT(HF) ;
  return 0;
}
