// 
//  Author:  Rainer Mankel                     Created:  29-Sep-98
//           Humboldt University Berlin        Modified:  
//  Email:   mankel@ifh.de                     HERA-B Collaboration
// 
// Modifications:
// - Adapted to `withinBox' scheme      16-Jul-01 Wouter Hulsbergen
// - Reimplemented sLength, since it contained two bugs.

#include <algorithm>
#include <iostream>
#include "rang/VolTube.hh"
#include "rang/FLAGRotFix.hh"
#include "rang/Arrow.hh"

// Before application of the rotation matrix, the tube/s axis is assumed
// to be identical with the z axis
// 
// The following information from GEDE/GWAL is used:
//
//   gwal->xmin   inner radius of tube
//       ->xmax   outer radius of tube
//       ->ymin   x offset of CM
//       ->ymax   y offset of CM
//       ->zmin   lower z bound
//       ->zmax   upper z bound
//       ->rot    rotation matrix
//
// See also gean/crvtrm.F


//-- Constructor
rangerVolTube::rangerVolTube(){};

//-- Constructor
rangerVolTube::rangerVolTube( rangerGeWa* gewa ) : rangerVolume( gewa ) 
{
  // calculate the reference coordinates in ACS and PCS
  float zP = 0.5*(Gewa->zmin() + Gewa->zmax()) ;
  float xP = Gewa->ymin() ;
  float yP = Gewa->ymax() ;
  rangerArrow a(zP,xP,yP,1,1) ;
  rotateArrow(a,-1) ;
  zA = a.z ;

  // put box around in PCS and transform all 8 corners
  xA1=yA1= 1.e9 ;
  xA2=yA2=-1.e9 ;
  float radius = max(Gewa->xmin(),Gewa->xmax()) ;
  for(int i=0; i<8; ++i) {
      xP  = Gewa->ymin() + ((i%4)/2 ? -1 :1)*radius ;
      yP  = Gewa->ymax() + ((i%2)   ? -1: 1)*radius ;
      zP = (i/4) ? Gewa->zmin() : Gewa->zmax() ;
      rangerArrow a(zP,xP,yP,1,1) ;
      rotateArrow(a,-1) ;
      xA1 = min(xA1,a.x) ;
      xA2 = max(xA2,a.x) ;
      yA1 = min(yA1,a.y) ;
      yA2 = max(yA2,a.y) ;	
  }	
}

//-- Destructor
rangerVolTube::~rangerVolTube() {};

bool rangerVolTube::withinAngle(const rangerArrow&, float z) const 
{
  return true ;
}

float rangerVolTube::sLength(float zz,   // give traversed length
			     float x,    // in units of
			     float y,    // zmax - zmin
			     float tx,
			     float ty) const {
  // rotation into PCS
  // see Hera-B 97-090 p 67
  rangerArrow a(zz,x,y,tx,ty);
  rotateArrow(a);

  // translate arrow to cylinder center
  a.x -= Gewa->ymin(); // x-coordinate of cylinder axis
  a.y -= Gewa->ymax(); // y-coordinate of cylinder axis

  // the idea is to calculate z positions at which cylinder walls are
  // hit
  
  // inner wall:
  double a2 = a.tx*a.tx+a.ty*a.ty ;
  double a1 = 2*(a.tx*a.x + a.ty*a.y) ;
  double a0 = a.x*a.x + a.y*a.y - Gewa->xmin()*Gewa->xmin() ;
  double discr = a1*a1 - 4*a2*a0 ;
  double zinn1(0),zinn2(0) ;
  bool hitinn = false ;
  if( discr > 0 && Gewa->xmin()>0) {
    hitinn = true ;
    zinn1 = a.z + 0.5*(-a1 - sqrt(discr))/a2 ;
    zinn2 = a.z + 0.5*(-a1 + sqrt(discr))/a2 ;
  }

  // outer wall
  a0 = a.x*a.x + a.y*a.y - Gewa->xmax()*Gewa->xmax() ;
  discr = a1*a1 - 4*a2*a0 ;
  double zout1(0),zout2(0) ;
  bool hitout = false ;
  if( discr > 0 ) {
    hitout = true ;
    zout1 = a.z + 0.5*(-a1 - sqrt(discr))/a2 ;
    zout2 = a.z + 0.5*(-a1 + sqrt(discr))/a2 ;
  }
  
  // now, did we actually hit anything ?
  double cylzmin = Gewa->zmin() ;
  double cylzmax = Gewa->zmax() ;
  double len(0) ;

  if( !hitinn ) {
    // first the case that only the outer wall is hit 
    // (only inner wall not being possible)
    if( hitout && 
	( (cylzmin < zout1 && zout1 < cylzmax) || 
	  (cylzmin < zout2 && zout2 < cylzmax) ) ) {
      double zmin = max(min(zout1,zout2),cylzmin) ; 
      double zmax = min(max(zout1,zout2),cylzmax) ;
      // a first simple approach:
      if(withinAngle(a,0.5*(zmin+zmax))) len += (zmax-zmin)*sqrt(a2) ;
      //cout << "third solution: " << zmin << " " << zmax << endl;
    }
  } else if(hitout) { // should always be, but just for safety
    // solution 1
    if( ( cylzmin < zinn1 && zinn1 < cylzmax) ||
	( cylzmin < zout1 && zout1 < cylzmax) ) {
      double zmin = max(min(zinn1,zout1),cylzmin) ; 
      double zmax = min(max(zinn1,zout1),cylzmax) ;
      if(withinAngle(a,0.5*(zmin+zmax))) len += (zmax-zmin)*sqrt(a2) ;
      //cout << "first solution: " << zmin << " " << zmax << endl ;
    }
    // solution 2:
    if( ( cylzmin < zinn2 && zinn2 < cylzmax) ||
	( cylzmin < zout2 && zout2 < cylzmax) ) {
      double zmin = max(min(zinn2,zout2),cylzmin) ; 
      double zmax = min(max(zinn2,zout2),cylzmax) ;
      if(withinAngle(a,0.5*(zmin+zmax))) len += (zmax-zmin)*sqrt(a2) ;
      //cout << "second solution: " << zmin << " " << zmax << endl ;
    }
  }
  
  // now normalize
  return len/(cylzmax-cylzmin) ;
}

