// 
//  Author:  Rainer Mankel                     Created:  29-Sep-98
//           Humboldt University Berlin        Modified:  
//  Email:   mankel@ifh.de                     HERA-B Collaboration
//
// Modifications:
// - Faster access to rotation matrix   14-Aug-00 Wouter Hulsbergen 
// - Adapted to `withinBox' scheme      16-Jul-01 Wouter Hulsbergen
// - Removed bug from rotateArrow (tx,ty were not rotated)

#include <iostream.h>
#include "arte/ArteMessage.hh"
#include "rang/Volume.hh"
#include "rang/VolBox.hh"
#include "rang/VolEpi.hh"
#include "rang/VolTrd1.hh"
#include "rang/VolTrd2.hh"
#include "rang/VolTube.hh"
#include "rang/VolTubs.hh"
#include "rang/VolCone.hh"
#include "rang/VolDummy.hh"
#include "rang/Node.hh"
#include "rang/Arrow.hh"
#include "rang/FLAGRotFix.hh"

//-- Constructor
rangerVolume::rangerVolume(){
  Gewa = 0;
  for (int i=0; i<=2; i++) {         // initialize rot matrix
    for (int j=0; j<=2; j++) {
      rotTot[i][j] = (i==j)? 1 :0;
    }
  }
  zA = 0;
  msValue = 0;
  rotated = 0;
  zTilted = 0;
  totRotated = 0;
  xA1=yA1=-1.e4 ;
  xA2=yA2= 1.e4 ;
};

//-- Constructor from gewa
rangerVolume::rangerVolume( rangerGeWa* gewa ) 
  : Gewa(gewa),zA(0),xA1(0),yA1(0),xA2(0),yA2(0)
{
  // make total rotation matrix (only needed for GEDEs)
  rotTot[0][0] = gewa->rot(1) * gewa->cos() - gewa->rot(2) * gewa->sin();
  rotTot[0][1] = gewa->rot(4) * gewa->cos() - gewa->rot(5) * gewa->sin();
  rotTot[0][2] = gewa->rot(7) * gewa->cos() - gewa->rot(8) * gewa->sin();
  rotTot[1][0] = gewa->rot(1) * gewa->sin() + gewa->rot(2) * gewa->cos();
  rotTot[1][1] = gewa->rot(4) * gewa->sin() + gewa->rot(5) * gewa->cos();
  rotTot[1][2] = gewa->rot(7) * gewa->sin() + gewa->rot(8) * gewa->cos();
  rotTot[2][0] = gewa->rot(3);
  rotTot[2][1] = gewa->rot(6);
  rotTot[2][2] = gewa->rot(9);

  rotated = (gewa->rot(1) == 1 && gewa->rot(5) == 1 && gewa->rot(9) == 1)?
    0 : 1;
  zTilted = !(gewa->rot(3) == 0 && gewa->rot(6) == 0 && 
	      gewa->rot(7) == 0 && gewa->rot(8) == 0);
  totRotated = (rotTot[0][0] == 1 && rotTot[1][1] == 1 && rotTot[2][2] == 1)?
    0 : 1;
  
  switch (gewa->cmp()) {
  case 0:
    msValue = 1.0 * gewa->ms();  // wall
    break;
  case 1:
    msValue = 1.1 * gewa->ms();  // SiVD
    break;
  case 2:
    msValue = 1.4 * gewa->ms();  // HDC
    break;
  case 3:
    msValue = 1.6 * gewa->ms();  // MSGC
    break;
  default:
    msValue = 1.1 * gewa->ms();
    break;
  }
  float z = 0.5*(gewa->zmin() + gewa->zmax()) ; // not correct for rotated volumes ;
  if (z>200. && z<650.) {
    if (gewa->cmp() != 1 && gewa->cmp() != 3) {
      msValue *= 1.8;  // extra MS for magnet
    }    
    if (z>400.) {
      msValue *= 1.3; // extra extra MS
    }
  }
}

//-- Destructor
rangerVolume::~rangerVolume(){
  delete Gewa;
}


void rangerVolume::rotateArrow(rangerArrow& a, int direction) const
{
  static float invgwrot[9] ;

  if (rotated) {
    double xP(a.x), yP(a.y), zP(a.z), txP(a.tx), tyP(a.ty);
    const float* gwrot = Gewa->rotmatrix() ;
    
    if(direction<0 ) {
      // use inverted rotation matrix:
      for(int i=0; i<9; ++i) invgwrot[i] = gwrot[ (i/3) + (i%3)*3 ] ;
      gwrot = invgwrot ;
    }
    
    if (zTilted) {
#ifdef ROTFIX
      xP = gwrot[0] * a.x + gwrot[3] * a.y + gwrot[6] * (a.z-1080);
      yP = gwrot[1] * a.x + gwrot[4] * a.y + gwrot[7] * (a.z-1080);
      zP = gwrot[2] * a.x + gwrot[5] * a.y + gwrot[8] * (a.z-1080)
	+ 1080;
#else   
      xP = gwrot[0] * a.x + gwrot[3] * a.y + gwrot[6] * a.z;
      yP = gwrot[1] * a.x + gwrot[4] * a.y + gwrot[7] * a.z;
      zP = gwrot[2] * a.x + gwrot[5] * a.y + gwrot[8] * a.z;
#endif
      
      double tzn = gwrot[2] * a.tx + gwrot[5] * a.ty + gwrot[8];
      if (tzn == 0) {
	cout << "rangerVolume::rotateArrow: tzn=" << tzn << endl;
	tzn = .00001;
      }
      txP = (gwrot[0] * a.tx + gwrot[3] * a.ty + gwrot[6]) / tzn;
      tyP = (gwrot[1] * a.tx + gwrot[4] * a.ty + gwrot[7]) / tzn;
    } else {
      xP  = gwrot[0] * a.x + gwrot[3] * a.y;  // coordinates in PCS 
      yP  = gwrot[1] * a.x + gwrot[4] * a.y;  // see Hera-B 97-090 p 68
      txP = gwrot[0] * a.tx + gwrot[3] * a.ty ;
      tyP = gwrot[1] * a.tx + gwrot[4] * a.ty ;
    }   
    a.z = zP;
    a.x = xP;
    a.y = yP;
    a.tx = txP;
    a.ty = tyP;
  }
}

// returns traversed lengt in units of (zmax-zmin),
float rangerVolume::sLength(float , float , float , float , float) const { 
  cout << "Volume::sLength: Oh no!! this is just"
       << " the default!!" << endl;
  return 0;}


rangerVolume* rangerVolume::makeVolume( rangerGeWa* gewa ) {
  rangerVolume* vb = 0;
  switch ( gewa->shp() ) {
  case 1:
    vb = new rangerVolBox( gewa );
    break;
  case 2:
    vb = new rangerVolTrd1( gewa );
    break;
  case 3:
    vb = new rangerVolTrd2( gewa );
    break;
  case 5:
    vb = new rangerVolTube( gewa );
    break;
  case 6:
    vb = new rangerVolTubs( gewa );
    break;
  case 7:
    vb = new rangerVolCone( gewa );
    break;
  case 10:
    vb = new rangerVolEpi( gewa );
    break;
  case 34:
    vb = new rangerVolBox( gewa );   // Lanyov trapezoid, treat like a box
    break;
  case 42:
    vb = new rangerVolBox( gewa );   // Lanyov trapezoid, treat like a box
    break;
  default:
    arteMsg << "rangerVolume::makeVolume: use rangerVolDummy for shp="
	    << gewa->shp() << " pl=" << gewa->plNumber() << endl;
    static int id;
    arteMsg.warn(id);
    vb = new rangerVolDummy( gewa );
    break;
  };

  return vb;
};

// void rangerVolume::initializeACS(float& z, float& x1, float& x2,
// 				 float& y1, float& y2) const {
//   // This is some nonsense default
//   z = 0.5*(Gewa->zmin() + Gewa->zmax()) ;
//   x1 = y1 = -1.e5 ;
//   x2 = y2 = 1.e5 ; 
// }
