// *****************************************************************************
//
// source:    JPSIvertex
//
// created:   12 Dec 2000
//	
// author:    Silvia Masciocchi
//            MPI Munich
//            E-mail: sma@mppmu.mpg.de
//
//            Many thanks to Max Sang for the very precious help!
//
// purpose:   fit a vertex with the 2 selected leptons and store it in
//            RVER either always or only when detached
// 
// input:     the 2 RTRA entries of the 2 leptons, and
//            JPSIVtxType = MUONS_ALWAYS         mu+mu-: store RVER in any case
//                          MUONS_DETACHED       mu+mu-: store RVER only if detached
//                          ELECTRONS_ALWAYS     e+e-:   store RVER in any case
//                          ELECTRONS_DETACHED   e+e-:   store RVER only if detached
//
// vertexing: fit only without mass constraint
//            reject vertices with chi2>=50
//            for the detached: separation from closest wire must be >= 4 sigma
//            if RVER is filled, an RTRA entry is also created. At the moment this
//               is identified by the rtra->fid=443, and not yet by the relation to
//               the topp table. This awaits a common decision.
//
// returns    0 if no vertex was stored in RVER
//            1 if a vertex was stored in RVER (always)
//            2 if a vertex was stored in RVER (detached)
//
// changes:
// 12 Dec 2000 (SM) Creation
// 13 Dec 2000 (SM) The mass of the daughter particles is set using a topp
//                      link, so that the proper relations are stored in RREL.
//                  The first GroverParticle is always the negative lepton.
//                  When required the RVER entry is stored, with the link to
//                      the jpsi TOPP, rtra->fit 2, rtra->fid=443 if set: 
//                       443  e+e- not detached
//                      1443  e+e- detached
//                      2443  mu+mu- not detached
//                      3443  mu+mu- detached
// 10 Jul 2001 (SM) Update of the part about the detachment: the closest wire is
//                  now chosen with the J/psi mother track, and the separation in
//                  sigma is corrected including the wire thickness
//
// *****************************************************************************

// ARTE includes
#include "arte/arte.hh"
#include "farm/farm.hh"
#include "grover/grover.hh"
#include "reco/JPSIvertex.hh"

//
// function code
//
int
JPSIvertex(ArtePointer<RTRA> track1,ArtePointer<RTRA> track2,JPSIVtxType t)
{

  const float DVlower = 3.5;   // default lower cut on vertex separation in sigma
  float verdet = 0.;           // vertex detachment in sigma
  int stored = 0;
  ArtePointer<TOPP> lepm,lepp;
  if ( t == ELECTRONS_ALWAYS || t == ELECTRONS_DETACHED ){
    lepm = ArtePointer<TOPP>(18);
    lepp = ArtePointer<TOPP>(19);
  }
  else if( t == MUONS_ALWAYS || t == MUONS_DETACHED ){
    lepm = ArtePointer<TOPP>(22);
    lepp = ArtePointer<TOPP>(23);
  }
  
  GroverParticle mu1,mu2;
  if(track1->pf <= 0) {
    mu1=GroverParticle(track1); 
    mu2=GroverParticle(track2);
  }
  else{
    mu1=GroverParticle(track2); 
    mu2=GroverParticle(track1);
  }

  mu1.GlobalToLocal(); mu2.GlobalToLocal();
  mu1.SetMass(lepm);
  mu2.SetMass(lepp);
  
  // unconstrained fit
  GroverVertex psi(&mu1,&mu2);
  psi.Fit(-1.0);
  if(psi.Chi2() >= 50) return 0;
  psi.LocalToGlobal();
  // store this vertex in my list:
  GroverVertexList dileptons;
  dileptons.push_back(psi);
  
  // Reload the wires from RTAR 
  GroverWireList DVWires;
  if(ArteTable<RTAR>::size() != 0)
    GroverFillFromRTAR(&DVWires);
  else
    GroverFillFromGTAR(&DVWires);
  
  // Pick the associated wire.
  // Before we chose the one which is closest in z.
  // Now we use the impact of the J/psi mother track:
  GroverParticle motherj(&psi);
  float mindist2 = 1e10; float dist2 = 1e9;
  GroverWire* closestw2 = 0;
  if ( !DVWires.empty() ){
    for (GroverWireList::iterator itar = DVWires.begin();
	 itar != DVWires.end(); ++itar) {
      // consider only wires close to beam
      if((sqrt(itar->X()*itar->X() + itar->Y()*itar->Y()))<1.0){
	//dist2 = fabs(psi.Z() - itar->Z()) ;
	dist2 = motherj.DistanceTo(itar);
	if(dist2 < mindist2){
	  mindist2 = dist2;
	  closestw2 = itar;
	}
      }
    }
  }

  // Calculate the separation in sigma:
  float dzpsi = 0.; float dz = 0.;
  if (closestw2 != 0){
    dzpsi = (psi.Cov(5)>0.0) ? sqrt(psi.Cov(5)) : 1e9;
    dz = sqrt( dzpsi*dzpsi + ((closestw2->SizeZ()*closestw2->SizeZ())/12));
    verdet = (psi.Z() - closestw2->Z())/dz ;
    //cout << "DV: separation in sigma = " << verdet << endl;
  }  
  else cout << "*** JPSIvertex: No closest wire found !!!!!!" << endl;
  
  if ( t == ELECTRONS_ALWAYS    || 
       t == MUONS_ALWAYS        ||
       (t == ELECTRONS_DETACHED && fabs(verdet) >= DVlower ) || 
       (t == MUONS_DETACHED && fabs(verdet) >= DVlower )){
    
    ArtePointer<TOPP> jpsi(295);
    GroverAddToRVER(&dileptons, 2, jpsi);
    stored = 1;

    ArtePointer<RTRA> detached(ArteTable<RTRA>::size());
    if ( t == ELECTRONS_ALWAYS || t == ELECTRONS_DETACHED )
      detached->fid = 443;
    else if( t == MUONS_ALWAYS || t == MUONS_DETACHED )
      detached->fid = 2443;
    if ( fabs(verdet) >= DVlower ){
      stored = 2;
      detached->fid += 1000;
    }
  }

  return stored;
  
}

