//***********************************************************************
//
//  Routine:      rcSelectRtraBasePatt
//  Author:       W. Hulsbergen (20/01/01)
//

//  Description: Removes otr clones from otr-vds RTRAs (for
//  rtra2001). A quality criterium is defined on the basis of the
//  number of hits and the chi2 from the fit. Tracks are sorted
//  according to their quality and collisions are removed.
//
//  Modification:
//
//***********************************************************************

#include <algorithm>
#include <vector>
#include <arte/arte.hh>
#include <arte/RSEG_cmp.hh>


float vdspattquality(const ArtePointer<RTRA>& rtra)
{
  int nhits = rtra->hsi + rtra->hou + rtra->hin ;
  int ndof  = rtra->ndof>0 ? rtra->ndof : nhits - 5 ;
  if( rtra->chi <=0 ) return 0 ;
  return nhits*ndof/rtra->chi ;
  // So, is this a good quality definition ? Keep in mind the
  // following:
  // * chi/ndof is more or less independent of the number of hits. in
  //   MC it is perfectly equal to 1; in data it is on average more
  //   equal to 2 (today). However, it certainly does not increase with the
  //   number of hits (I checked :).
  // * if tracks have a comparable chi2, we prefer the long
  //   track. note that we are only removing OTR clones, so the number
  //   of hits can only differ in the number of vds hits. The difference
  //   is therefore rather small.
}

bool vdspattcompare(const ArtePointer<RTRA>& lhs, 
		    const ArtePointer<RTRA>& rhs)
{
  // overcome ranger refit error flag
  if(lhs->fit==100 && rhs->fit!=100) return false ;
  if(lhs->fit!=100 && rhs->fit==100) return true ;
  return vdspattquality(lhs) > vdspattquality(rhs) ;
}

int rcSelectRtraBasePatt()
{
  // removes PC clones from RTRA bank
  typedef vector<ArtePointer<RTRA> >  rtracontainer ;
  rtracontainer rtras ;
    
  for(ArteTable<RTRA>::iterator irtra = ArteTable<RTRA>::begin() ;
      irtra !=  ArteTable<RTRA>::end() ; ++irtra)
    if( ((*irtra)->cmp & Rsegc::patt) &&
	((*irtra)->cmp & Rsegc::vxd) &&
	!((*irtra)->cmp & Rsegc::bitsign) ) rtras.push_back(*irtra) ;

  sort(rtras.begin(),rtras.end(),vdspattcompare) ;

  for(rtracontainer::iterator irtra = rtras.begin() ;
      irtra != rtras.end() ; ++irtra)
    if ( !((*irtra)->cmp & Rsegc::bitsign) )
      for(rtracontainer::iterator jrtra = irtra+1 ;
	  jrtra != rtras.end() ; jrtra++)
	if( !((*jrtra)->cmp & Rsegc::bitsign) &&
	    (*jrtra)->cmain == (*irtra)->cmain ) {
	  (*jrtra)->cmp |= Rsegc::bitsign ;
	  (*jrtra)->fit += 1000 ;
	}
  
  return 0 ;
}
