// *****************************************************************************
//
// source:    Selection of D*+/- mesons using the following channels:
//               D*+ :  D0 pi+   (i.e. K- pi+ pi+ - K- pi+)
//               D*- :  ~D0 pi-  (i.e. K+ pi- pi- - K+ pi-)
//             Returns:
//               1 upon success, 0 otherwise  
//
// dependences: RTRA bank filled, D0 reconstruction finished
//
// type:      source code
//
// created:   11 Oct 2000
//	
// author:    Denis Dujmic
//
// changes:
//   Oct 30, 2000 (DD) Missing track analysis added, polishing
//    23 Oct 2001 (DS) fill RTRA rich likelihoods only if not filled before
//
// *****************************************************************************

#include "reco/rcX.h"
#include "reco/rcXkinema.h"
#include "rhp.h"
#include "dq/DQ.h"


rcX *rcD2010a=0;
rcX *rcD2010b=0;


#ifdef USE_ROOT_DD
#define DM_LO 1.85
#define DM_HI 2.25
#else
#define DM_LO 1.85
#define DM_HI 2.25
#endif


int D2010plus__D0_Piplus_RC();
int D2010minus__D0bar_Piminus_RC();
int maximizeRTRA();
void rcResetRichRtra(float x=-1);


int D2010_rlevel=2;



//
//  Main D* reco
//
int 
D2010()
{
  D2010plus__D0_Piplus_RC();
  D2010minus__D0bar_Piminus_RC();
  return 0;
}



/////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////  REAL RECO   /////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////

int
D2010plus__D0_Piplus_RC()
{
  int verb=0;
  
  if (verb) cout<<">>>>>>>>>>>>>>>>>>  D*_RC1 <<<<<<<<<<<<<<<<<<<"<<endl;
  if (!rcD2010a) {
    cout<<"initializing D* a"<<endl;
    rcD2010a=new rcX("rec_D2010a", "D2010+ -> D0 Pi+", verb);
    decay dstar;
    dstar.rlevel        = D2010_rlevel; // 0=full, 1=semi, 2=mtra
    dstar.fit           = 202;          // = 'long' decay
    dstar.book_tables   = 1;            // 1/0
    dstar.root_output   = 1;            // make root file or not
    dstar.verb          = verb;         // printout level
    rcD2010a->setMother(dstar);         // set
    vertexAttrib vdstar;
    vdstar.rlevel            = 2;     // 0=no, 1=chi2, 2=kalman, 3=kalman+mass
    vdstar.topp              = kinema::findTopp(413); // pointer to mother
    vdstar.max_vertex_chi2   = 5;  // limit for vertex acceptance 
    vdstar.verb              = verb;  // printout level
    vdstar.min_mass          = DM_LO; // minimum mass
    vdstar.max_mass          = DM_HI; // maximum mass
    //vdstar.max_drho_from_wire= 0.025; // maximum impact parameter at z of primary vertex
    //vdstar.min_dz_from_wire  = 0.15;  // minimum decay length
    rcD2010a->setVertex(vdstar);      // set

    attribute pi1;
    pi1.topp          = kinema::findTopp(211);
    pi1.mass          = pi1.topp->mass;
    pi1.minP          = 0;
    pi1.maxP          = 200;
    pi1.minHIO        = 15;
    pi1.minHSI        = 6;
    //pi1.use_mc_p      = 1;
    pi1.use_missing_p = 1;
    pi1.minHRI       = 1;
    pi1.minPt        = 0.15;
    //pi1.minLRPI      = 0.1;
    pi1.minLREMUPI   = 0.1;
    pi1.maxEoP        = 0.8;
    pi1.maxCandidate  = 10;
    pi1.useVdsOnce    = 1;
    pi1.useMainOnce   = 1;
    pi1.maxChi       = 500;
    //pi1.lund         = 211; // MC
    //pi1.mother       = 413; // MC
    rcD2010a->addDaughter(pi1);

    attribute d0;
    d0.topp         = kinema::findTopp(421);
    d0.mass         = d0.topp->mass;
    d0.fit          = 3;
    d0.matchTOPP    = 1; // require RTRA-TOPP relation booked for D candidates
    d0.minMass      = 0;
    d0.maxMass      = 5;
    //d0.lund         = 421;  // MC
    //d0.mother       = 413;  // MC
    rcD2010a->addDaughter(d0);

    if (verb) rcD2010a->printDaughters();
  }
  rcD2010a->rc();
  if (verb) cout<<">>>>>>>>>>>>>>>>>>  D*_RC1 <<<<<<<<<<<<<<<<<<<"<<endl;
  return 0;
}

//
//
//
int
D2010minus__D0bar_Piminus_RC()
{
 int verb=0;
  
  if (verb) cout<<">>>>>>>>>>>>>>>>>>  D*_RC2 <<<<<<<<<<<<<<<<<<<"<<endl;
  if (!rcD2010b) {
    cout<<"initializing D* b"<<endl;
    rcD2010b=new rcX("rec_D2010b", "D2010- -> D0 Pi-", verb);
    decay dstar;
    dstar.rlevel        = D2010_rlevel; // 0=full, 1=semi, 2=mtra
    dstar.fit           = 202;          // = 'long' decay
    dstar.book_tables   = 1;            // 1/0
    dstar.root_output   = 1;            // make root file or not
    dstar.verb          = verb;         // printout level
    rcD2010b->setMother(dstar);         // set
    vertexAttrib vdstar;
    vdstar.rlevel            = 2;     // 0=no, 1=chi2, 2=kalman, 3=kalman+mass
    vdstar.topp              = kinema::findTopp(-413); // pointer to mother
    vdstar.max_vertex_chi2   = 5;     // limit for vertex acceptance 
    vdstar.verb              = verb;  // printout level
    vdstar.min_mass          = DM_LO; // minimum mass
    vdstar.max_mass          = DM_HI; // maximum mass
    //vdstar.max_drho_from_wire= 0.025; // maximum impact parameter at z of primary vertex
    //vdstar.min_dz_from_wire  = 0.15;  // minimum decay length
    rcD2010b->setVertex(vdstar);      // set

    attribute pi1;
    pi1.topp          = kinema::findTopp(-211);
    pi1.mass          = pi1.topp->mass;
    pi1.minP          = -200;
    pi1.maxP          = 0;
    pi1.minHIO        = 15;
    pi1.minHSI        = 6;
    pi1.use_missing_p = 1;
    pi1.minHRI       = 1;
    pi1.minPt        = 0.15;
    //pi1.minLRPI      = 0.1;
    pi1.minLREMUPI   = 0.1;
    pi1.maxEoP        = 0.8;
    pi1.maxCandidate  = 10;
    pi1.useVdsOnce    = 1;
    pi1.useMainOnce   = 1;
    pi1.maxChi       = 500;
    pi1.lund         = -211; // MC
    pi1.mother       = -413; // MC
    rcD2010b->addDaughter(pi1);

    attribute d0;
    d0.topp         = kinema::findTopp(-421);
    d0.mass         = d0.topp->mass;
    d0.fit          = 3;
    d0.matchTOPP    = 1; // require RTRA-TOPP relation booked for D candidates
    d0.minMass      = 0;
    d0.maxMass      = 5;
    d0.lund         = -421;  // MC
    d0.mother       = -413;  // MC
    rcD2010b->addDaughter(d0);

    if (verb) rcD2010b->printDaughters();
  }
  rcD2010b->rc();
  if (verb) cout<<">>>>>>>>>>>>>>>>>>  D*_RC2 <<<<<<<<<<<<<<<<<<<"<<endl;
  return 0;
}


//
//
//
int
selectD2010()
{
  ARTE_mark( "rcX_D* 0", 0 );
  static int old_total1=0;
  static int old_total2=0;
  int ret=0;

  //=== check if rich likelihoods in RTRA are filled,
  //  take maximal likelihoods from  RTRI otherwise

  int richpid=1;
  if (ArteTable<RTRA>::size()!=0){
    ArtePointer<RTRA> rtra(1);
    if (rtra->lre<0&&rtra->lrmu<0&&rtra->lrpi<0&&rtra->lrk<0&&rtra->lrp<0){
      int mr=maximizeRTRA();
      richpid=-1;
    }
  }

//   rhp_hf1(DQ_ID( COMP_RECO,LVL_REC, 1040 ), 0., 1.); // new event
  ArteTable<RTRA>::iterator d1=ArteTable<RTRA>::end();
  D2010();
  ArteTable<RTRA>::iterator d2=ArteTable<RTRA>::end();
  if (rcD2010a&&rcD2010a->total>old_total1) { old_total1=rcD2010a->total; ret=1; }
  if (rcD2010b&&rcD2010b->total>old_total2) { old_total2=rcD2010b->total; ret=1; }
  if (ret) {
//     rhp_hf1(DQ_ID( COMP_RECO,LVL_REC, 1040 ), 1., 1.); // D found
    // -- fill mass
    for (ArteTable<RTRA>::iterator id=d1; id!=d2; ++id) {
      ArtePointer<RVER> rver=(*id)->ever;
      float mass=rver->mass;
//       rhp_hf1(DQ_ID( COMP_RECO,LVL_REC, 1041 ), mass, 1.);
    }
  }

  //== reset rich likelihoods in RTRA to default value if no rich likelihood
  //   was filled before calling this routine

  if (richpid<0){
       rcResetRichRtra(-1);
  }


  ARTE_mark( "rcX_D* 1", 0 );
  return ret;
}

//
// 
//
void
stopD2010()
{
  if (rcD2010a) delete rcD2010a;
  if (rcD2010b) delete rcD2010b;
}
