// *****************************************************************************
//
// source:   Geometrical acceptance routine
//
// type:      source code for evaluation of geometrical acceptance
//
// created:   14 Oct 2000
//	
// author:    Denis Dujmic
//
// changes: 
//   16 Oct 2000 (DD) RICH acceptance added
//
// *****************************************************************************
#include "reco/rcXgeo.h"
#include "rich/rich.h"

float  rcXgeo::max_z_pc=1000;
float  rcXgeo::min_z_pc=675;
int    rcXgeo::gesl_cmp=2;

//
//  Checks for a hit inside a pitch, (after W. Hulsbergen)
//
bool rcXgeo::hit_within_pitch(ArtePointer<GEDE> gede, double* track) {
  
  // this is the parameter you would have to tune
  const float npitches = 1; 
  
  if(gede->cmp!=gesl_cmp || gede->hitl.index() < 1) return false;
 
  // -- check if track passes through
  float Cos = gede->rot[0] ;
  float Sin = gede->rot[1] ;
  //float tanThetaU = cos*state[3] - sin*state[4] ;
  //float cosThetaU =  1 / sqrt( 1 + tanThetaU * tanThetaU);    
  float zgede = (gede->zmin + gede->zmax)/2; // z position of gede
  float dz = zgede - track[0] ;
  float x = track[1] + dz*track[3] ; // x position of track at gede
  float y = track[2] + dz*track[4] ; // y position of track at gede
  float rotatedx = Cos*x - Sin*y ;   // track x in coordinates of gede
  float rotatedy = Sin*x + Cos*y ;   // track y in coordinates of gede
  
  // is gede hit ?
  if(rotatedx < gede->sxmn || rotatedx > gede->sxmx ||
     rotatedy < gede->symn || rotatedy > gede->symx) return false;

  int   gede_hit = 0;
  float dx_min   = npitches*gede->pit;
  for(ArteTable<HITB>::iterator hitbIt = gede->hit1 ;
      hitbIt != gede->hitl + 1; hitbIt++) {
    float dx_tmp = hitbIt->x - rotatedx;
    if ( fabs(dx_tmp)<npitches*gede->pit ) {gede_hit++; }
    if ( fabs(dx_tmp)<fabs(dx_min) )       dx_min=dx_tmp;
  }
  if (fabs(dx_min)<npitches*gede->pit) { 
    return true ;
  }

  return false ;
}



//
//  Checks for geometrical acceptance of a track by a gede
//
bool rcXgeo::track_within_gede(ArtePointer<GEDE> gede, double* track) {
  int verb=0;

  bool ret=false;
  
  if (gede->cmp!=gesl_cmp) return false;
 
  // -- check if track passes through gede
  float Cos = gede->rot[0] ;
  float Sin = gede->rot[1] ;
  float zgede = (gede->zmin + gede->zmax)/2; // z position of gede
  float dz = zgede - track[0] ;
  float x = track[1] + dz*track[3] ; // x position of track at gede
  float y = track[2] + dz*track[4] ; // y position of track at gede
  float rotatedx = Cos*x - Sin*y ;   // track x in coordinates of gede
  float rotatedy = Sin*x + Cos*y ;   // track y in coordinates of gede

  if (verb>10) cout<<"track0:"<<track[0]<<" "<<track[1]<<" "<<track[2]
		   <<" track1:"<<track[0]+dz<<" "<<rotatedx<<" "<<rotatedy
		   <<"  gede:"<<gede->zmin<<" "<<gede->zmax<<" "
		   <<gede->sxmn<<" "<<gede->sxmx<<" "
		   <<gede->symn<<" "<<gede->symx;

  if(rotatedx > gede->sxmn && rotatedx < gede->sxmx &&
     rotatedy > gede->symn && rotatedy < gede->symx) ret=true;

  if (ret&&verb>1) cout<<" ok "<<endl;
  else if (verb>1) cout<<endl;

  return ret;
}



//
//  Counts gede-hits on track path
//
int  rcXgeo::num_otr_hits(double* track) {
  int count=0;
  rcXgeo::gesl_cmp=2;
  rcXgeo::min_z_pc=675;
  rcXgeo::max_z_pc=1000;
  for (ArteTable<GEDE>::iterator igede=ArteTable<GEDE>::begin();
       igede!=ArteTable<GEDE>::end(); ++igede) {

    // -- select PC
    if ((*igede)->cmp!=gesl_cmp) continue;
    if ((*igede)->zmin>max_z_pc) continue;
    if ((*igede)->zmin<min_z_pc) continue;

    if (hit_within_pitch((*igede),track)) count++;
  }
  return count;
}

//
//  Counts OTR gede planes on track path
//
int  rcXgeo::num_otr_planes(double* track) {
  int count=0;
  rcXgeo::gesl_cmp=2;
  rcXgeo::min_z_pc=675;
  rcXgeo::max_z_pc=1000;
  for (ArteTable<GEDE>::iterator igede=ArteTable<GEDE>::begin();
       igede!=ArteTable<GEDE>::end(); ++igede) {

    // -- select PC
    if ((*igede)->cmp!=gesl_cmp) continue;
    if ((*igede)->zmin>max_z_pc) continue;
    if ((*igede)->zmin<min_z_pc) continue;

    if (track_within_gede((*igede),track)) count++;
  }

  //cout<<"track:"<<track[0]<<" "<<track[1]<<" "<<track[2];
  //cout<<"  total number of hit planes="<<count<<endl;
  return count;
}

//
//  Counts ITR gede planes on track path
//
int  rcXgeo::num_itr_planes(double* track) {
  int count=0;
  rcXgeo::gesl_cmp=3;
  rcXgeo::min_z_pc=675;
  rcXgeo::max_z_pc=1000;
  for (ArteTable<GEDE>::iterator igede=ArteTable<GEDE>::begin();
       igede!=ArteTable<GEDE>::end(); ++igede) {

    // -- select PC
    if ((*igede)->cmp!=gesl_cmp) continue;
    if ((*igede)->zmin>max_z_pc) continue;
    if ((*igede)->zmin<min_z_pc) continue;

    if (track_within_gede((*igede),track)) count++;
  }

  return count;
}


//
//  Counts VDS gede planes on track path
//
int  rcXgeo::num_vds_planes(double* track) {
  int count=0;
  rcXgeo::gesl_cmp=1;
  rcXgeo::min_z_pc=0;
  rcXgeo::max_z_pc=230;
  for (ArteTable<GEDE>::iterator igede=ArteTable<GEDE>::begin();
       igede!=ArteTable<GEDE>::end(); ++igede) {

    // -- select PC
    if ((*igede)->cmp!=gesl_cmp) continue;
    if ((*igede)->zmin>max_z_pc) continue;
    if ((*igede)->zmin<min_z_pc) continue;

    if (track_within_gede((*igede),track)) count++;
  }

  return count;
}



//
//   Counts the number RICH photons inside acceptance
//    (maximum 35)
//
int rcXgeo::num_rich_photons(double *track) {

  static int    n=35;
  static double costhe=cos(.0526);

  Vec3 x(track[1],track[2],track[0]);
  Vec3 p(track[3],track[4],      1.);

  CoordinateSystem A;
  A.setrotation(p); // x,y arbitrary, but perpendicular to \vec{p}.
  A.x0()=10*x;

  Photon pp;
  int i;
  int eff=0;
  for (i=0;i<n;i++) { 
    double phi=2*M_PI*rnd_normal();
  
    pp.n()     = A.RotateToOld(FromPolar(1,costhe,phi));
    pp.x()     = A.x0();
    
    int r1=RICH_Vessel->geometry_trace(&pp);
    int r2=RICH_Detector->hit_in_acceptance(pp);
    if (!r1&&r2 ) eff++;
  }
  return eff;
}
