//--*-C++-*--
//
//  Description:
//    HERAB specific function for filling of a photon list
//    from ARTE tables.
//    RCopyHITB() fills the photon list from HITB for futher use
//    RSelectROIHITB() selects ROI from the photon list
//
//
//  Author:  Denis Dujmic
//
//  Created: Apr 04 2001
//
//  Modifications:
//     Apr 26, 2001  (DD)  Global alignment added
//     May 18, 2001  (DD)  Printout and checks added
//     Jun 28, 2001  (DD)  Printout changed
//
//
#include "arte/arte.hh"
#include "arte/RSEG_cmp.hh"
#include "rich/rich.h"
#include "rise/RObject.h"
#include "rise/RRingFit.h"
#include "rise/RErrorMsg.h"
#include "rise/RAttribute.h"

#define A16 2.493e-6
#define A4  9.972e-6
#define S16 7e-4
#define S4  1e-3



/**
   Create a hit list in ARTE enviroment.

   Copy RICH hits from HITB table to a list of photons 
   RichObject<RHit>::list and apply global alignment of hits
*/
void RCopyHITB() {

  // .. get alignment parameters
  float dphi=0, dlam=0;
  bool  applied=false;
  static bool first=true;
  for( ArteTable<ACGA>::iterator iacga = ArteTable<ACGA>::begin();
       iacga != ArteTable<ACGA>::end(); ++iacga ) {
    if ((*iacga)->cmp!=Rsegc::rich) continue;
    dphi    = (*iacga)->dtx; // rich -> global
    dlam    = (*iacga)->dty; // rich -> global
    applied = true;
    if (first) {
      char msg[80];
      //sprintf(msg,"Global alignment will be applied to RICH hits dtx=%.5f dty=%.5f",dphi,dlam);
      //RErrorMsg err(msg,0);
      RErrorMsg err("Global alignment applied to RICH hits",0);
      if (RAttribute::verb) err.print();
      first=false;
    }    
  }
  if (first && !applied) {
      RErrorMsg err("Global alignment NOT applied to RICH hits",0);
      if (RAttribute::verb) err.print();
      first=false;
  }


  // .. check if same event
  static int evt_no=0;
  if (ArtePointer<EVHD>(1)->evt==evt_no) return;
  evt_no=ArtePointer<EVHD>(1)->evt;

  // .. clear old list
  RichObject<RHit>::list.clear();

  // .. loop over HITB
  int n0=RichRun::gede->hit1.index();
  int n1=RichRun::gede->hitl.index();
  if (n1-n0<=0) {
    cout<<"RCopyHITB: RICH hitl before hit1"<<endl;
    return;
  }

  RHit rhit;
  for (int i=n0;i<n1;i++) {
    ArtePointer<HITB> hitb(i);
    ArteRelationIterator<DIGB> idigb=hitb->digb.begin();
    ArtePointer<DIGB> digb(idigb);
    int id     = digb->id;
    if (!RICH_Detector->ChannelExist(id)) {
      static int ierr=0;
      if (++ierr<10)
	cout<<"RCopyHITB: requested channel not initialized - check DB/data files!"<<endl;
      continue;
    }
    rhit.x        = hitb->q + dphi;
    rhit.y        = hitb->r + dlam;
    rhit.plane    = (smFromChannelID(id)>6) ? 1 : 0 ;
    rhit.area     = (hitb->chid==4) ? A4 : A16;
    rhit.sigma    = (hitb->chid==4) ? S4 : S16;
    rhit.weight   = 1;
    rhit.assigned = 0;
    rhit.index    = hitb.index();
    rhit.mctrack  = 9999;
    // ... MC truth
    if (ArteTable<MTRA>::size()) {
      ArteRelationIterator<MIMP> imimp=hitb->mimp;
      ArtePointer<MTRA> mtra=(*imimp)->mtra;
      rhit.mctrack = mtra.index();
    }
    RichObject<RHit>::list.push_back(rhit);
  }
  if (!RichObject<RHit>::list.size()) {
    RErrorMsg err("Event without RICH hits",-1);
    throw err;
  }
}



/**
   From RichObject<RHit>::list extract ROI photons and make HERAB 
   optical corrections
*/
void RSelectROIHITB() {
  int verb=0;

  // .. get ring, track, hits
  if (!gRingFit) {
    RErrorMsg err("RSelectROIHITB() can be called from a RRingFit only, unrecoverable error",-1);
    throw err;
  }
  RRing  &ring =(*gRingFit->getRing());
  RTrack &track=(*ring.itrack);
  vector<RCorrHit> &hit_list=*gRingFit->getHitList();
  float  Delta2=ring.Delta*ring.Delta;

  // .. clear list
  if (hit_list.size()) hit_list.clear();

  if (verb) {
    cout<<"ROI center="<<track.X<<" "<<track.Y<<"   radius="<<ring.radius<<" "<<ring.r2min<<" "<<ring.r2max<<endl;
  }
  // .. initialize corrections
  float bx=(track.x-track.z*track.tx)*.01;
  float by=(track.y-track.z*track.ty)*.01;
  RICH_LambdaPhi->Correct()->newRing(track.X, track.Y, ring.radius, bx, by);
  for (vector<RHit>::iterator ihit=RichObject<RHit>::list.begin();
       ihit!=RichObject<RHit>::list.end(); ++ihit) {
    // .. check ROI
    float dx = - RICH_LambdaPhi->Correct()->dPhi   ( (*ihit).x, (*ihit).y, (*ihit).plane);
    float dy = - RICH_LambdaPhi->Correct()->dLambda( (*ihit).x, (*ihit).y, (*ihit).plane);
    float r2 = dx*dx + dy*dy;
    if (r2<ring.r2min || r2>ring.r2max) continue;
    // .. save hit
    RCorrHit rcorrhit;
    rcorrhit.X        = track.X-dx;
    rcorrhit.Y        = track.Y-dy;
    rcorrhit.ihit     = ihit;
    rcorrhit.assigned = (r2<Delta2) ? true : false;
    rcorrhit.b        = 0;
    rcorrhit.w        = 0;
    rcorrhit.flux     = 0;
    hit_list.push_back(rcorrhit);
  }
  ring.nROI=hit_list.size();
  if (!ring.nROI) {
    RErrorMsg err("No hits in ROI",0);
    if (verb) err.print();
    throw err;
  }
  if (verb) cout<<"N ROI photons="<<hit_list.size()<<endl;
}
