//-----------------------------------------------------------------------//
//
//  RingSearchARTE: ARTE interface for stand-alone ring reconstruction
//
//  Author: Reinhard Eckmann/Denis Dujmic
//
//  Created: May-98
//
//  Modifications:
//    May 15, 2000 (DD) nhit filled with Ring::numhit_3sigma
//    May 24, 2000 (DD) OTR based reconstruction: ringsearchRSEG()
//    May 29, 2000 (DD) ringsearchRSEG: use underdogs, SN>4 with seed,
//                      RingSearchAttrib::finer_search_step added
//    May 31, 2000 (DD) makeHitMap: check if radius list empty
//    Jun 02, 2000 (DD) makeHitMap: occupancy cut checked only for 
//                      stand-alone search
//                      swim(const RPNT&,RPNT&,const double) introduced
//    Jun 06, 2000 (DD) ringsearchRTRA fine-tuned
//    Jun 09, 2000 (DD) new kuip commands
//    Jul 03, 2000 (DD) 'refined' search omitted by default
//    Jul 06, 2000 (DD) crich filled for rtra-seed search
//    Jul 21, 2000 (DD) rtra->hri filled with rseg->nhit (if no rtra-hitb)
//    Jul 25, 2000 (DD) slight changes in rtra filling
//    Sep 17, 2000 (DD) small printout change 
//    Oct 26, 2000 (DD) attempt to make order in rlevel
//    Nov 27, 2000 (DD) minSearchRadius check fixed
//    Jan 12, 2001 (DD) minor polishing, rtra->hri filled
//    Feb 16, 2001 (DD) flr filled for sa rings
//    Mar 21, 2001 (DD) crich saved only for s/a rings
//    Apr 16m 2001 (DD) flag convention changed
//
//-----------------------------------------------------------------------//
#include "rich/rich.h"
#include "rise/RingSearchARTE.h"
#include "rise/RingSearchTiming.h"
#include "arte/RSEG_cmp.hh"



#ifdef RISE_ROOT
RingSearchTiming *RICH_SearchTiming=0; 
#endif

/////////////////////////// Ring search ////////////////////////////////////////

RingSearchARTE::RingSearchARTE() : RingSearchDB() {

#ifdef RISE_TIMING
  if (!RICH_SearchTiming) RICH_SearchTiming=new RingSearchTiming;
#endif
  mm[0] = 2.6e-7;
  mm[1] = 1.12e-2;
  mm[2] = 1.95e-2;
  mm[3] = .24372;
  mm[4] = .88035;

  n_mass=5;
}


int RingSearchARTE::copyHITB() {
  int verb=0;

  if (verb) cout<<"RingSearchARTE::copyHITB"<<endl;

  static int evt_no=0;
  if (ArtePointer<EVHD>(1)->evt==evt_no) return 1;
  evt_no=ArtePointer<EVHD>(1)->evt;

  hitbhits.reset();

  int n0=RichRun::gede->hit1.index();
  int n1=RichRun::gede->hitl.index();
  if (n1-n0<=0) return 3;
 
  for (int i=n0;i<n1;i++) {
    ArtePointer<HITB> hitb(i);
    ArteRelationIterator<DIGB> idigb=hitb->digb.begin();
    ArtePointer<DIGB> digb(idigb);
    int id    = digb->id;
    int plane = (smFromChannelID(id)>6) ? 1 : 0 ;
    if (!RICH_Detector->ChannelExist(id)) {
      static int ierr=0;
      if (++ierr<10)
	cout<<"$ RISE: requested channel not initialized - check DB/data files!"<<endl;
      continue;
    }
    int size  = hitb->chid;
    int imtra = 9999;
    // -- mc track
    if (ArteTable<MTRA>::size()) {
      ArteRelationIterator<MIMP> imimp=hitb->mimp;
      ArtePointer<MTRA> mtra=(*imimp)->mtra;
      imtra = mtra.index();
    }
    hitbhits.append(d.dualx(hitb->q),d.dualy(hitb->r), plane, size, imtra, hitb);
    if (verb) cout<<".";
  }
  if (verb) cout<<"copyHITB hits= "<<hitbhits.numhit<<endl;
  return 0;
}



int RingSearchARTE::makeHitMap() {

  int verb=0;

  // -- reset hit bank
  h.reset();
  d.nextevent();

  // -- ocupancy cut
  int n0=RichRun::gede->hit1.index();
  int n1=RichRun::gede->hitl.index();
  if (!RingSearchTrack::exist && n1-n0>RingSearchAttrib::occupancyCut)      return 1; 
  if (RingSearchTrack::exist  && n1-n0>RingSearchAttrib::seedOccupancyCut &&
      RingSearchAttrib::seedOccupancyCut>0)                                 return 2; 
  if (n1-n0<=0)                                                             return 3;

  // -- track based ring search - set selection cuts
  float min_rad=0, max_rad=0;
  if (RingSearchTrack::exist && RingSearchAttrib::id) {
    if (verb>1) cout<<"new ring input:"<<RingSearchTrack::phi<<" "<<RingSearchTrack::lambda<<" "<<RingSearchTrack::bx
		    <<" ("<<-RingSearchTrack::phi*4.5<<") "
		    <<RingSearchTrack::by<<endl;
    float rr=ArtePointer<RITH>(1)->rad0;
    RICH_LambdaPhi->Correct()->newRing(RingSearchTrack::phi, RingSearchTrack::lambda, rr,
				       RingSearchTrack::bx,RingSearchTrack::by);
    min_rad =  60;
    max_rad =  0;
    if (!radius_list.size()) {
      cout<<"$ RISE: RingSearchARTE::makeHitMap: search radius list empty "<<endl;
      return 3;
    }
    for (vector<float>::iterator irad=radius_list.begin(); 
	 irad!=radius_list.end(); ++irad) {
      if (min_rad>(*irad)) min_rad=(*irad);
      if (max_rad<(*irad)) max_rad=(*irad);
    }
    min_rad*=.001;
    max_rad*=.001;
    min_rad-=RingSearchTrack::max_rad_residual;
    max_rad+=RingSearchTrack::max_rad_residual;
    if (verb) cout<<"min/max="<<min_rad<<"/"<<max_rad<<endl;
  }


  // -- append photons
  for (int i=0;i<hitbhits.numhit;i++) {

    Hit &sh=hitbhits.hit[i];

    // -- check if photon has been used
    if (!RingSearchAttrib::reuse &&
	sh.photon_hitb->rseg.begin()!=sh.photon_hitb->rseg.end()) continue;    

    // -- track seed search
    if (RingSearchTrack::exist) {
      float dx=fabs(RICH_LambdaPhi->Correct()->dPhi   (sh.x, sh.y, sh.plane));
      float dy=fabs(RICH_LambdaPhi->Correct()->dLambda(sh.x, sh.y, sh.plane));
      float dr=sqrt(dx*dx+dy*dy);
      if (dr>max_rad || dr<min_rad) continue;
    }
    h.append(d.dualx(sh.x),d.dualy(sh.y), sh.plane, sh.size, sh.mctrack, sh.photon_hitb);
    if (verb) cout<<".";
  }
  if (verb) cout<<" numhit="<<h.numhit<<endl;
  if (ArteTable<MTRA>::size()) h.randomize(); 
  return 0;
}


void RingSearchARTE::makeRadiusList(float r1, float r2, float drad) {
  int verb=0;
  
  if (verb) cout<<"RingSearchARTE::makeRadiusList"<<endl;

  if (radius_list.size()) radius_list.clear();

  if (RingSearchTrack::exist && RingSearchAttrib::id) {
    float oldrad=100;

    for (int i=0; i<5; i++) {
      float rr=R2P2rad2[i]*1e6;
      float R2P2sigma=1e6/sqrt(iR2P2sigma2[i]);
      if (rr<0) continue;
      float r=sqrt(rr);
      if (r<minSearchRadius()) continue;
      if (RingSearchAttrib::finer_search_step) { 
	float dr=sqrt(rr+4*R2P2sigma)-r;
	if (drad<0) {
	  if (oldrad-r>2*dr) radius_list.push_back(r);
	  if (oldrad-r>3*dr) radius_list.push_back(r+dr);
	  if (oldrad-r>dr) { radius_list.push_back(r-dr); oldrad=r; }
	}
	else {
	  if (oldrad-r>3*dr)   radius_list.insert(radius_list.begin(),r+dr);
	  if (oldrad-r>2*dr)   radius_list.insert(radius_list.begin(),r);
	  if (oldrad-r>dr)   { radius_list.insert(radius_list.begin(),r-dr); oldrad=r; }
	}
      }
      else radius_list.push_back(r);
    }

  }
  else {
    if (drad<0) for (float radius=r2;radius>r1;radius+=drad) {// mrad /////
      radius_list.push_back(radius);
    }
    else for (float radius=r1;radius<r2;radius+=drad) {// mrad /////
      radius_list.push_back(radius);
    }
  }
}


void RingSearchARTE::print() {
  cout<<"$ RISE: Acctual search parameters:"<<endl;
  if (RichRun::magnet) cout<<"$         RingSearch with magnet ON"<<endl;
  else                 cout<<"$         RingSearch with magnet OFF"<<endl;
  cout<<"$       minSearchRadius="<<RingSearchAttrib::minSearchRadius;
  cout<<" maxSearchRadius="<<RingSearchAttrib::maxSearchRadius;
  cout<<" searchStep="<<RingSearchAttrib::searchStep;
  cout<<" peakThreshold="<<RingSearchAttrib::peakThreshold<<endl;
  cout<<"$       occupancyCut for SA="<<RingSearchAttrib::occupancyCut;
  cout<<" occupancyCut for seed="<<RingSearchAttrib::seedOccupancyCut;
  cout<<endl;
  if (RingSearchAttrib::type==RingSearchType::nobg) {
    cout<<"$       chi2 fit";
    cout<<" SN="<<RingSearchAttrib::SNcut<<endl;
    if (RingSearchAttrib::fitLevel==3) {
      cout<<"$       likelihood refit";
      cout<<" S="<<RingSearchAttrib::Scut<<endl;
    }
  }
  else {
    cout<<"$       likelihood fit";
    cout<<" S="<<RingSearchAttrib::Scut<<endl;
  }
  cout<<"$       refined search="<<RingSearchAttrib::finer_search_step<<endl;
  cout<<"$       minimum cell distance=radius*"<<RingSearchAttrib::minCellDistance<<endl;
}


void RingSearchARTE::getrings(float r1,float r2,float dr,int t) {
  int verb=0;

  static int multi=0;
  if (!multi) {
    print();
    multi=1;
  }


#ifdef RISE_TIMING
  RICH_SearchTiming->startTiming();
#endif

  switch (RingSearchAttrib::seed_type) {
  case RingSearchType::rtra_seed: if (!SetSeedRTRA()) return; break;
  case RingSearchType::rseg_seed: if (!SetSeedRSEG()) return; break;
  default: cout<<"$ RISE: RingSearchARTE::getrings: Wrong seed type"<<endl;
  }
  makeRadiusList(r1,r2,dr);
  copyHITB();
  if (makeHitMap()) return;

  int threshold=t;
  for (vector<float>::iterator irad=radius_list.begin();
       irad!=radius_list.end(); ++irad) {
    //////////if (RingSearchTrack::exist && (*irad)<40) threshold=3;
    //////////else if (RingSearchTrack::exist) threshold=4;
    if (verb) cout<<"radius="<<(*irad)<<"  threshold="<<threshold<<"  seed="<<RingSearchTrack::exist<<endl;
    d.define((*irad)*1e-3,threshold);
    d.ringsearch(h);
  }
  
  if (RingSearchAttrib::choose_best_ring) Choose();
  if (RingSearchAttrib::bookrseg)         FillRSEG();

#ifdef RISE_TIMING
  RICH_SearchTiming->stopTiming();
  RICH_SearchTiming->fillTimingHistograms();
#endif
}



void RingSearchARTE::getrings(int i) {

  if (R2P2rad2[i]<1e-4) return;

  Ring &testring=getRing(0);

  // -- set parameters
  testring.thetaC()=sqrt(R2P2rad2[i]);
  bool old_use_phi=RingSearchAttrib::use_seed_phi;
  bool old_use_lam=RingSearchAttrib::use_seed_lambda;
  RingSearchAttrib::use_seed_phi=true;
  RingSearchAttrib::use_seed_lambda=true;

  if (!testring.calcring(h)) return;
  else numrings()=1;

  if (RingSearchAttrib::bookrseg) FillRSEG();

  // -- restore parameters
  RingSearchAttrib::use_seed_phi   =old_use_phi;
  RingSearchAttrib::use_seed_lambda=old_use_lam;  
}




void RingSearchARTE::getrings() {
  if (!ArteTable<MTRA>::nrow()) getSearchParameters(ArtePointer<EVHD>(1)->run);
  getrings(minSearchRadius(),maxSearchRadius(),searchStep(),threshold());
}


//////////////////////////  Seed setup //////////////////////////////////

bool RingSearchARTE::SetSeedRTRA() {
  int verb=0;

  if (verb) cout<<"SetSeedRTRA"<<endl;

  if (!RingSearchTrack::exist) return true;

  // -- set track parameters
  RingSearchTrack::ip    = RingSearchTrack::rtra->pf;
  //!!!!!!!!!!!!!!!!! make sure RPNT is the right one !!!!!!!!!!!!!!
  ArtePointer<RPNT> rpnt;
  float max_z=0;
  for (ArteRelationIterator<RPNT> irpnt=RingSearchTrack::rtra->rpnt.begin();
       irpnt!=RingSearchTrack::rtra->rpnt.end(); irpnt++) {
    if ((*irpnt)->zf>max_z && 
	(*irpnt)->zf<RingSearchTrack::max_rpnt_z &&
	(*irpnt)->zf>RingSearchTrack::min_rpnt_z) { rpnt=(*irpnt); max_z=(*irpnt)->zf; }
  }
  if (max_z<RingSearchTrack::min_rpnt_z) {
    if (verb) cout<<"$ RISE: RingSearchARTE::SetSeedRTRA: no valid RPNT found"<<endl;
    if (verb) cout<<"$ RISE: RingSearchARTE::SetSeedRTRA: swimming"<<endl;
    ArtePointer<RTRA> rtra=RingSearchTrack::rtra;
    RPNT  pin, pou;
    pin.zf=rtra->zf;   pin.xf=rtra->xf;   pin.yf=rtra->yf;
    pin.txf=rtra->txf; pin.tyf=rtra->tyf; pin.pf=rtra->pf;
    swim(pin, pou, 1000);
    rpnt=ArteTable<RPNT>::push_back(pou);
  }


  // -- get track parameters inside RICH
  float x  = rpnt->xf;
  float y  = rpnt->yf;
  float z  = rpnt->zf;
  float tx = rpnt->txf;
  float ty = rpnt->tyf; 

  RingSearchTrack::phi    = RICH_LambdaPhi->phi   (tx, ty);
  RingSearchTrack::lambda = RICH_LambdaPhi->lambda(tx, ty);
  RingSearchTrack::bx     = ( x - tx*z )*.01; // m
  RingSearchTrack::by     = ( y - ty*z )*.01; // m

  if (verb) cout<<"rpnt: "<<x<<" "<<y<<" "<<z<<" "<<tx<<" "<<ty<<" "
		<<RingSearchTrack::phi<<" "<<RingSearchTrack::lambda<<" "
		<<RingSearchTrack::bx<<" "<<RingSearchTrack::by<<endl;

  // -- calculate expected positions and errors on R2P2 lines
  if (RingSearchAttrib::id) makeR2P2lines();
  return true;
}

bool RingSearchARTE::SetSeedRSEG() {
  int verb=0;

  if (verb) cout<<"SetSeedRSEG"<<endl;

  if (!RingSearchTrack::exist) return true;

  // -- get track parameters inside RICH
  float x  = RingSearchTrack::rseg->xe;
  float y  = RingSearchTrack::rseg->ye;
  float z  = RingSearchTrack::rseg->ze;
  float tx = RingSearchTrack::rseg->txe;
  float ty = RingSearchTrack::rseg->tye; 

  if (z<550) { 
    cout<<" $ RICH: RingSearchARTE::SetSeedRSEG: RSEG not valid for seed, z="<<z<<endl;
    return false;
  }

  // -- set track parameters
  //RingSearchTrack::ip    = RingSearchTrack::rseg->pf;
  RingSearchTrack::ip    = (x-tx*z)/(-288.);

  RingSearchTrack::phi    = RICH_LambdaPhi->phi   (tx, ty);
  RingSearchTrack::lambda = RICH_LambdaPhi->lambda(tx, ty);
  RingSearchTrack::bx     = ( x - tx*z )*.01; // m
  RingSearchTrack::by     = ( y - ty*z )*.01; // m

  // -- calculate expected positions and errors on R2P2 lines
  if (RingSearchAttrib::id) makeR2P2lines();
  return true;
}



void RingSearchARTE::Choose() {
  int verb=0;

  if (verb) cout<<"RingSearchARTE::Choose() numrings="<<numrings()<<endl;

  static int multi=0;
  if (!multi) {
    switch (RingSearchAttrib::choose_best_ring) {
    case 1:  cout<<"$ RISE: Choose best significance"<<endl; break;
    case 2:  cout<<"$ RISE: Choose best r2p2"<<endl; break;
    case 3:  cout<<"$ RISE: Choose heavier mass"<<endl; break;
    case 4:  cout<<"$ RISE: Choose closest to seed"<<endl; break;
    }
    multi=1;
  }

  if (!RingSearchTrack::exist) return;
  // -- in case of seeded search find best ring
  float best_S=0;
  int   besti=-1;
  float best_r2p2=100;
  int   best_r2p2i=-1;
  float best_res=100;
  for (int i=0; i<numrings(); i++) {
    Ring &ring=getRing(i);
    ring.accept=false;
    if (!ring.IsCloseToSeed()) continue;
    // -- find best significance or S/Sqrt(N)
    if (RingSearchAttrib::type==RingSearchType::lkbg &&
	RingSearchAttrib::choose_best_ring==1) {
      if (best_S<ring.refit_S) {
	best_S=ring.refit_S;
	besti=i;
      }
    }
    else if (RingSearchAttrib::type==RingSearchType::nobg &&
	     RingSearchAttrib::choose_best_ring==1) {
      float sn=ring.refit_sn;
      if (best_S<sn) {
	best_S=sn;
	besti=i;
      }
    }
    else if (RingSearchAttrib::choose_best_ring==2) {
      // --- lowest r2p2
      int b=R2P2likelihood(ring.radius);
      if (b<0) continue;
      float r2p2=GetLr(b);
      if (r2p2<best_r2p2) {best_r2p2=r2p2; besti=i; }
      float sn=ring.refit_sn;
      if (verb) cout<<RingSearchTrack::rtra<<" "<<ring.radius<<" r2p2("<<b<<")"<<r2p2<<"  sn="<<sn
		    <<" tx/ty="<<ring.phi()-RingSearchTrack::phi
		    <<"/"<<ring.lambda()-RingSearchTrack::lambda
		    <<"  p="<<1/RingSearchTrack::ip<<endl;
    }
    else if (RingSearchAttrib::choose_best_ring==3) {
      // --- highest id
      int b=R2P2likelihood(ring.radius);
      if (b<0) continue;
      float r2p2=GetLr(b);
      if (b>best_r2p2i) { best_r2p2i=b; besti=i; }
      if (verb) cout<<RingSearchTrack::rtra<<" "<<ring.radius<<" r2p2("<<b<<")"<<r2p2
		    <<" tx/ty="<<ring.phi()-RingSearchTrack::phi
		    <<"/"<<ring.lambda()-RingSearchTrack::lambda
		    <<"  sn="<<ring.refit_sn
		    <<"  p="<<1/RingSearchTrack::ip<<endl;
    }
    else if (RingSearchAttrib::choose_best_ring==4) {
      // --- closest to seed
      int b=R2P2likelihood(ring.radius);
      if (b<0) continue;
      float r2p2=GetLr(b);
      float dx=ring.phi()-RingSearchTrack::phi;
      float dy=ring.lambda()-RingSearchTrack::lambda;
      float dr=sqrt(dx*dx+dy*dy);
      if (best_res>dr) { best_res=dr; besti=i; }
      if (verb) cout<<RingSearchTrack::rtra<<" "<<ring.radius<<" r2p2("<<b<<")"<<r2p2
		    <<" tx/ty="<<ring.phi()-RingSearchTrack::phi
		    <<"/"<<ring.lambda()-RingSearchTrack::lambda
		    <<"  sn="<<ring.refit_sn
		    <<"  p="<<1/RingSearchTrack::ip<<endl;
    }

  }
  if (besti<0) return;
  Ring &bestring=getRing(besti);
  bestring.accept=true;
}



extern "C"  void rkmain_(const int    & icovar, 
              const double & zout, 
              const double & fstep, 
              const double * parameters, 
              const double * prec, 
              const double * p_track, 
              const double * derivatives, 
              const int    & ierror);



int RingSearchARTE::swim(const RPNT &source,RPNT &dest,const double zout) {
  double pin[]={source.zf,source.xf,source.yf,source.txf,source.tyf,source.pf};
  double pou[6];
  int ierror=swim(pin,pou,zout);
  dest.zf =pou[0]; dest.xf =pou[1]; dest.yf =pou[2];
  dest.txf=pou[3]; dest.tyf=pou[4]; dest.pf =pou[5];
  return ierror;
}



int RingSearchARTE::swim(const double *source,double *dest,const double zout) {
  const int    icovar=2;
  double       fstep=50;
  double       prec[4];
  double       derivatives[5][5];
  int          ierror;

  prec[0] = 0.001/fabs(zout-source[0]);
  prec[1] = prec[0];
  prec[2] = 0.0001/fabs(zout-source[0]);
  prec[3] = prec[2];
  
  rkmain_(icovar,zout,fstep,source,prec,dest,&derivatives[0][0],ierror);

  return ierror;
}


////////////////////////// booking of ARTE tables ////////////////////////


void RingSearchARTE::FillRSEG() {
  int verb=0;

  RSEG r;
  ArtePointer<RSEG>rseg;
  //cout<<"RingSearchARTE::FillRSEG() numrings="<<numrings()<<endl;
  for (int i=0; i<numrings(); i++) {

    Ring &ring=getRing(i);

    // -- check distance seed-ring center
    if (RingSearchTrack::exist && 
	RingSearchAttrib::choose_best_ring && 
	(!ring.accept) ) continue;

    // -- ring info
    r.xf = ring.phi();
    r.yf = ring.lambda();

    r.time   = ring.thetaC();
    float r0 = ArtePointer<RITH>(1)->rad0;
    r.dtim   = cos(r0)/cos(r.time);

    r.txf = ring.tx();
    r.tyf = ring.ty();

    r.pf  = float(ring.numhit_assign);
    r.ze  = ring.numhit_refit;
    r.ye  = ring.numhit_bgrd;
    r.xe  = ring.numhit_3sigma;
    r.zf  = ring.f;
    r.txe = ring.Rb;
    r.tye = ring.refit_S;
    r.pe  = ring.sigma_signal;

    r.nhit= (int)ring.numhit_3sigma;

    r.chi = ring.chi2;

    r.cvf[ 0] = 0;
    r.cvf[ 1] = 0;
    r.cvf[ 2] = 0;
    r.cvf[ 5] = ring.covariance(0,0);
    r.cvf[ 8] = ring.covariance(0,1);
    r.cvf[12] = ring.covariance(0,2);
    r.cvf[ 9] = ring.covariance(1,1);
    r.cvf[13] = ring.covariance(1,2);
    r.cvf[14] = ring.covariance(2,2);
    r.cvf[ 3] = 0;
    r.cvf[ 6] = 0;
    r.cvf[10] = 0;
    r.cvf[ 4] = 0;
    r.cvf[ 7] = 0;
    r.cvf[11] = 0;
    
    // -- component
    r.cmp = Rsegc::rich;

    // -- fit type
    if (RingSearchTrack::exist) r.fit=12;       // seeded track
    else {
      float ratio=ring.thetaC()/r0;
      ratio*=ratio;
      if (ring.numhit_refit>50*ratio) r.fit=10; // possible gamma conversion
      else r.fit=8;                             // charged track found standalone
    }

    // -- book relations
    // -- RSEG
    rseg=ArteTable<RSEG>::push_back(r);

    // -- RSEG-MTRA
    unsigned int iMtra=ring.mctrack();
    if (ArteTable<MTRA>::size() && iMtra<=ArteTable<MTRA>::size() && iMtra>0)
      createRelation(rseg,ArtePointer<MTRA>(iMtra),&RSEG::mtra);

    // -- seed based search: book RTRA-RSEG
    if (RingSearchTrack::exist && 
	RingSearchAttrib::seed_type==RingSearchType::rtra_seed) {
      createRelation(rseg, RingSearchTrack::rtra, &RSEG::rtra);
    }

    if (verb)
      cout<<"rich="<<rseg<<" "<<RingSearchTrack::phi<<" "<<RingSearchTrack::lambda<<" || "<<ring.phi()<<" "<<ring.lambda()<<endl;

    // -- book RSEG-HITB
    if (verb) cout<<"hitb="<<ArteTable<HITB>::size()
		  <<"  numhits="<<h.numhit
		  <<"  ring-hits="<<ring.numhit_assign<<" nrings="<<numrings()<<endl;    
    if (RingSearchAttrib::bookhitb) {
      for (int i=0; i<ring.numhit; i++) {
	Hit &h=(*ring.hit[i]);
	if (!ring.assigned[i]) continue;
	createRelation(rseg, h.photon_hitb, &RSEG::hitb);
      }
    }

  }
}


void RingSearchARTE::FillRTRA(ArtePointer<RTRA> t, ArtePointer<RSEG> rich) {
  // -- ring radius and center
  t->cer   = rich->time;
  t->txr   = rich->txf;
  t->tyr   = rich->tyf;
  if (!RingSearchTrack::exist) t->crich = rich.index(); // save crich only for sa rings

  // -- get number of rich hits
  int hri=0;
  for (ArteRelationIterator<HITB> ihitb=rich->hitb.begin();
       ihitb!=rich->hitb.end(); ++ihitb) hri++;
  if (!hri) hri=rich->nhit;
  if (!hri) hri=3;
  t->hri=hri;

  // -- fill rtra likelihoods if requested
  FillRTRA(t);
}


void RingSearchARTE::FillRTRA(ArtePointer<RTRA> t) {
  if (!RingSearchAttrib::fillrtra) return;
  // -- fill likelihoods 
  t->lre =lr[0];
  t->lrmu=lr[1];
  t->lrpi=lr[2];
  t->lrk =lr[3];
  t->lrp =lr[4];
  // -- without rseg hri is lost, put 3 if not set already
  if (t->hri<=0) t->hri=3;
}


void RingSearchARTE::ResetID(ArtePointer<RTRA> t) {
  // -- preset lr to below threshold particles -
  // -- if ring not found the likelihood is spread
  // -- among these particles
  memset(&lr[0],0,sizeof(float)*5);
  /*float r0r0=ArtePointer<RITH>(1)->rad0*ArtePointer<RITH>(1)->rad0;
  float ipip=t->pf*t->pf;
  // -- find the lightest below-threshold particle
  int i=0;
  for (; i<5; i++) if (r0r0-mm[i]*ipip<1e-4) break;
  // -- set all heavy particles to equal share
  if (i<5) {
    float share=1./(5-i);
    for (int j=i; j<5; j++) lr[j]=share;
  }
  */

  // -- reset RTRA
  if (RingSearchAttrib::fillrtra) {
    if (RingSearchAttrib::use_simple_R2P2) {
      t->lre=t->lrmu=t->lrpi=t->lrk=t->lrp=10e4;
    }
    else {
      t->lre=lr[0]; t->lrmu=lr[1]; t->lrpi=lr[2]; t->lrk=lr[3]; t->lrp=lr[4];
    }
    // note, if riter was used we don't have this information any more (unless rseg exists)
    if (ArteTable<RSEG>::size()) t->hri=0; 
  }
}


/////////////////////// likelihood functions ///////////////////////////

void RingSearchARTE::makeR2P2lines() {
  int verb=0;
  if (verb) cout<<"RingSearchARTE::makeR2P2lines"<<endl;

  float  ipip = RingSearchTrack::ip*RingSearchTrack::ip;
  float  pp   = 1/ipip;
  float  sigma2_ipip = 4 * ipip*ipip * (RingSearchTrack::aa+RingSearchTrack::bb*pp);

  float r0   = ArtePointer<RITH>(1)->rad0;
  float r0r0 = r0*r0;
  for (int i=0; i<n_mass; i++) {
    iR2P2sigma2[i] = 1/(RingSearchTrack::sigma2_rr + mm[i]*mm[i] * sigma2_ipip);
    R2P2rad2[i]    = r0r0-mm[i]*ipip; 
    if (verb) cout<<"R2P2 lines "<<i<<": p="<<1/fabs(RingSearchTrack::ip)
		  <<"  r2="<<R2P2rad2[i]<<" sigma2="<<1/iR2P2sigma2[i]
		  <<" dp/p="<<sqrt(RingSearchTrack::aa+RingSearchTrack::bb*pp)
		  <<" sigma_ipip="<<sigma2_ipip<<" sigma2_rr="<<RingSearchTrack::sigma2_rr
		  <<" aa="<<RingSearchTrack::aa<<" bb="<<RingSearchTrack::bb<<endl;
  }
}



extern "C" float erfc_(float &x);

int RingSearchARTE::R2P2likelihood(ArtePointer<RTRA> t) {

  if (!RingSearchAttrib::id) return -1;

  // -- reset RTRA
  ResetID(t);

  // -- find rich segment
  // -- in case of mini's rseg is lost and cer has old value
  ArtePointer<RSEG> rich;
  ArteRelationIterator<RSEG> irseg=t->rseg.begin();
  for (;irseg!=t->rseg.end(); ++irseg) {
    if ((*irseg)->cmp&Rsegc::bitsign) continue;
    if ((*irseg)->cmp==Rsegc::rich) { 
      t->cer=(*irseg)->time; 
      t->txr=(*irseg)->txf;
      t->tyr=(*irseg)->tyf;
      rich=(*irseg);
      break; 
    }
  }

  // -- check if rich angle correct
  if (t->cer<.001 || t->cer>.07) { t->cer=0; return -1; }
  
  // -- calculate likelihoods
  int ret=R2P2likelihood(t->cer);

  // -- fill tables if rseg found
  if (irseg!=t->rseg.end())       FillRTRA(t,rich); // cer,txr,tyr,hri
  else                            FillRTRA(t);      // lr...,hri

  return ret;
}



int RingSearchARTE::R2P2likelihood(float radius) {

  int verb=0;

// -- calculate deviations from r2p2 lines
  static float chimax=50;
  float chi2=chimax;
  memset(&lr[0],0,sizeof(float)*n_mass);
  float rr=radius*radius;
  int bestM=-1;
  for (int i=0; i<n_mass; i++) {
    float d       = (rr - R2P2rad2[i]);
    lr[i]         = d*d*iR2P2sigma2[i];
    if (lr[i]<chi2) { chi2=lr[i]; bestM=i; }
  }

  if (verb) cout<<"lr="<<lr[0]<<" "<<lr[1]<<" "<<lr[2]<<" "<<lr[3]<<" "<<lr[4]<<endl;
  if (verb) cout<<"r="<<radius<<" p="<<1/RingSearchTrack::ip<<endl;

  if (RingSearchAttrib::use_simple_R2P2) return bestM;

  static float isqrt2=1/sqrt(2);
  lrsum=0;
  if (bestM>-1) {
    // -- calculate probabilities
    for (int i=0; i<n_mass; i++) {
      if (verb) cout<<"likelihoods "<<i<<":  r2p2="<<lr[i];
      if (lr[i]>chimax) lr[i]=0;
      else {
	lr[i]=sqrt(lr[i])*isqrt2;
	lr[i]=erfc_(lr[i]);
	lrsum+=lr[i];
      }
      if (verb) cout<<"  lr="<<lr[i]
		    <<" (rr-r0r0)^2="<<(rr - R2P2rad2[i])*(rr - R2P2rad2[i])
		    <<"  R2P2r2="<<R2P2rad2[i]<<endl;
    }
    if (verb) cout<<endl;

    // -- renormalize
    if (lrsum>1e-6 && RingSearchAttrib::renorm) for (int i=0; i<n_mass; i++) lr[i]/=lrsum;
  }
  else memset(&lr[0],0,sizeof(float)*n_mass);

  return bestM;
}



RingSearchARTE *RICH_Search=0;




//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////// FUNCTIONS CALLED BY USER ////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

extern "C" {

void rcrings_(float r1,float r2,float dr,int t) {
  RingSearchTrack::exist     = false;
  RingSearchAttrib::reuse    = true;
  RingSearchAttrib::bookrseg = true;
  if (!RICH_Search) RICH_Search=new RingSearchARTE;
  RICH_Search->getrings(r1,r2,dr,t);
}

void rcringsdb_() {
  RingSearchTrack::exist     = false;
  RingSearchAttrib::reuse    = true;
  RingSearchAttrib::bookrseg = true;
  RingSearchAttrib::bookhitb = true;
  if (!RICH_Search) RICH_Search=new RingSearchARTE;
  RICH_Search->getrings();
}

void rcringsid_(ArtePointer<RTRA> track) {
  if (!RICH_Search) RICH_Search=new RingSearchARTE;
  if ((track->fit%100)!=0) return;
  if (fabs(track->pf)>1 || fabs(track->pf)<.0025)   return;
  RingSearchTrack::exist     = true;
  RingSearchTrack::rtra      = track;
  RingSearchAttrib::reuse    = true;
  RICH_Search->getrings();
  RICH_Search->R2P2likelihood(track);
}

void rcringsclean_() {
  RingSearchTrack::exist     = false;
  RingSearchAttrib::reuse    = false;
  RingSearchAttrib::bookrseg = true;
  RingSearchAttrib::bookhitb = true;
  if (!RICH_Search) {
    fprintf(stderr,"$ RISE: Ring cleanup called w/o previous call to ringseach\n");
    return;
  }
  RICH_Search->getrings();
}

	   }



int ringid(ArtePointer<RTRA> t) {
  if (!RICH_Search) RICH_Search=new RingSearchARTE;
  bool old1=RingSearchAttrib::id;
  RingSearchAttrib::id       = true;
  RingSearchTrack::ip        = t->pf;
  RICH_Search->makeR2P2lines();
  int ret=RICH_Search->R2P2likelihood(t);
  RingSearchAttrib::id       = old1;
  return ret;
}


float* ringid(float pf, float radius) {
  if (!RICH_Search) RICH_Search=new RingSearchARTE;
  RingSearchTrack::ip        = pf;
  RingSearchAttrib::fillrtra = false;
  RICH_Search->makeR2P2lines();
  RICH_Search->R2P2likelihood(radius);
  return RICH_Search->GetLr();
}

void ringsearchRTRA() {
  int verb=0;
  if (!RICH_Search) RICH_Search=new RingSearchARTE;
  static int multi=0;
  static int nrej=0, nacc=0, nfound=0;
  for(ArteTable<RTRA>::iterator irtra=ArteTable<RTRA>::begin();
      irtra!=ArteTable<RTRA>::end(); irtra++) {

    if (verb) cout<<"Ring search RTRA="<<(*irtra)<<endl;;

    // -- charged track
    if (((*irtra)->fit%100)!=0) { if (verb) cout<<"=fit!=0 "<<endl; nrej++; continue; }
    
    // -- reset fields that might be filled by others
    RICH_Search->ResetID(*irtra);

    // -- only likelihoods from already found rings
    if (!RingSearchAttrib::rlevel && RingSearchAttrib::fillrtra) { 
      if (!multi) {
	cout<<"$ RISE: likelihoods computed from found rings"<<endl;
	multi=1;
      }
      ringid((*irtra));
      int rise_standalone=1; 
      (*irtra)->flr = rise_standalone << 8 + (*irtra)->flr;
      continue;
    }


    // -- check for already found rings (rlevel=1 or 3)
    if (RingSearchAttrib::rlevel==3 ) {
      if (!multi) {
	cout<<"$ RISE: RTRA seed search after stand-alone search"<<endl;
	multi=1;
      }
      ArteRelationIterator<RSEG> irseg=(*irtra)->rseg.begin();
      for (;irseg!=(*irtra)->rseg.end(); ++irseg) {
	if ((*irseg)->cmp&Rsegc::bitsign) continue;
	if ((*irseg)->cmp==Rsegc::rich) break;
      }
      if (irseg!=(*irtra)->rseg.end()) { 
	if (verb) cout<<" has ring"<<endl; nrej++;
	if (RingSearchAttrib::fillrtra) ringid((*irtra));
	continue;
      }
    }


    // ---  momentum rejection
    if (fabs((*irtra)->pf)>1 || fabs((*irtra)->pf)<.0025)   { 
      if (verb) cout<<"invalid p = "<<(*irtra)->pf<<endl; 
      nrej++; 
      continue; 
    }

    // -- check geometrical acceptance
    /////if ((*irtra)->flr<=0) continue;

    // -- make cuts on the minimum number of hits
    if ((*irtra)->hsi<RingSearchAttrib::minHsi) { if (verb) cout<<"low_hsi "<<endl; continue; }
    if ((*irtra)->hin<RingSearchAttrib::minHin) { if (verb) cout<<"low_hin "<<endl; continue; }
    if ((*irtra)->hou<RingSearchAttrib::minHou) { if (verb) cout<<"low_hou "<<endl; continue; }
    if ((*irtra)->hou+(*irtra)->hin<RingSearchAttrib::minHio) { if (verb) cout<<"low_hio "<<endl; continue; }
    if ((*irtra)->hca<RingSearchAttrib::minHca &&
	(*irtra)->hmu<RingSearchAttrib::minHmu)  { if (verb) cout<<"low_hca_hmu "<<endl; continue; }

    if (verb>100) cout<<"hits: "<<(*irtra)->hsi<<" "
		  <<(*irtra)->hin<<" "
		  <<(*irtra)->hou<<" "
		  <<(*irtra)->hca<<endl;


    if (RingSearchAttrib::rlevel==1&&!multi) {
      cout<<"$ RISE: RTRA seed search only"<<endl;
      multi=1;
    }
    if (RingSearchAttrib::rlevel==2&&!multi) {
      cout<<"$ RISE: RTRA search with cleanup"<<endl;
      multi=1;
    }

    // -- do seeded ringsearch
    RingSearchAttrib::seed_type     = RingSearchType::rtra_seed;
    RingSearchAttrib::id            = true;
    RingSearchAttrib::use_underdogs = false;
    RingSearchAttrib::reuse         = true;

    nacc++;
    rcringsid_((*irtra));
    if (RICH_Search->numrings()) nfound++;
  }
  // -- do cleanup
  if (RingSearchAttrib::rlevel==2) rcringsclean_();
  if (verb)  cout<<"rtra rejected="<<nrej<<"  accepted="<<nacc<<"  founf rings="<<nfound<<endl;
}



void ringsearchRSEG() {
  if (!RICH_Search) RICH_Search=new RingSearchARTE;
  for(ArteTable<RSEG>::iterator irseg=ArteTable<RSEG>::begin();
      irseg!=ArteTable<RSEG>::end(); irseg++) {
    // -- choose pure PC
    if ((*irseg)->fit!=0) continue; 
    if (((*irseg)->cmp&(Rsegc::wch|Rsegc::patt))!=(Rsegc::wch|Rsegc::patt)) continue;
    //cout<<"rseg="<<(*irseg)<<" "<<hex<<(*irseg)->cmp<<dec<<endl;

    // -- do general ringsearch
    RingSearchTrack::exist      = true;
    RingSearchAttrib::seed_type = RingSearchType::rseg_seed;
    RingSearchTrack::rseg       = (*irseg);
    RingSearchAttrib::reuse     = false;
    RingSearchAttrib::id        = true;

    RingSearchAttrib::finer_search_step = false;
    RICH_Search->getrings();
  }
  rcringsclean_();
}


void removerings() {
  for (ArteTable<RSEG>::iterator irseg=ArteTable<RSEG>::begin();
       irseg!=ArteTable<RSEG>::end(); ++irseg) {
    if ((*irseg)->cmp==Rsegc::rich) 
      (*irseg)->cmp=(Rsegc::bitsign|Rsegc::rich);
  }
}
