//-----------------------------------------------------------------------//
//
//  Likelihood ring fit
//  Author: Denis Dujmic
//  Created: Mar-00
//
//  Modifications:
//    29 May (DD) code polishing
//   Jun 09, 2000 (DD): new kuip commands
//
//-----------------------------------------------------------------------//
#include "rise/ring.h"
#include "CLHEP/Matrix/SymMatrix.h"
#include "CLHEP/Matrix/Vector.h"

  static int errf=0;

  static float n0T=1.3e4;
  static float a16=2.44e-6;
  static float a4 =4*a16;
  static float cutSigma=12;
  static float sigma16=.5e-3;
  static float sigma4=.5e-3;
  static float sigma=sigma16;
  static float isigma=1/sigma;
  static float maxSigma=cutSigma*sigma;
  static float maxSigma2=maxSigma*maxSigma;
  static float Sigma16=sqrt(a16/12. + sigma16*sigma16);
  static float Sigma4 =sqrt( a4/12. + sigma4*sigma4);
  static float pi_fact=6.34936e-2; // 1/(2*PI)^(3/2)
  static int   n_par=4;

  static int total_negative_alpha=0;
  static int total_negative_error=0;
  static int total_negative_alpha_and_error=0;
  static int total_good_matrix=0;
  static int total_inversion_error=0;
  static int total_upper_limit[4];
  static int total_lower_limit[4];
  static int total_skipping_update=0;
  static int total_update=0;
  static int max_iterations=15;
  static int max_errors=15;


  // -- parameters
  static HepVector       param(n_par,0);
  static HepVector       pinit(n_par,0);
  static HepVector       shift(n_par,0);
  static HepVector       step (n_par,0);
  static HepVector       lower(n_par,0);
  static HepVector       upper(n_par,0);
  // -- error matrix
  static HepSymMatrix    Error(n_par,0);
  static HepSymMatrix    alpha(n_par,0);
  // -- shifts
  static HepVector       beta  (n_par,0);
  static HepVector       gamma1(n_par,0);
  static HepVector       gamma2(n_par,0);



/////////////////////////////////// fit function //////////////////////////////////////
void Ring::fitfun(int &npar, double *grad, double &f, double *par, int iflag) {
  
  double  dlnF_dx=0,    dlnF_dy=0,    dlnF_dr=0,    dlnF_db=0;
  memset(&grad[0],0x0,npar*sizeof(double));

  // -- calculate function and derivatives
  double lnF=0;
  for (int i=0; i<numhit; i++) {
    corrHit &ch=hit_list[i];

    register float ddx = ch.x - par[0];
    register float ddy = ch.y - par[1];
    register float rri = sqrt(ddx*ddx+ddy*ddy);
    
    float sigmai=0;
    if      (ch.h->size==16) sigmai=Sigma16;
    else if (ch.h->size==4)  sigmai=Sigma4;
    else {
      fprintf(stderr,"Hit size not found %d\n",ch.h->size);
      continue;
    }
    
    float t       = (rri-par[2])/sigmai;
    t*=t;
    double factor = 1 / ( par[2] + par[3]*par[3]*sigmai*exp(0.5*t) );
    
    switch (iflag) {
    case  0:
    case  4:
      if (t<50) lnF += log( par[3]*par[3] + par[2]/sigmai*exp(-0.5*t) );
      else      lnF += log( par[3]*par[3] );
      break;
      
    case 10:
      dlnF_dx += par[2]*ddx*(rri-par[2])/(rri*sigmai*sigmai)*factor;
    case 11:
      dlnF_dy += par[2]*ddy*(rri-par[2])/(rri*sigmai*sigmai)*factor;
    case 12:
      dlnF_dr += (sigmai*sigmai+par[2]*(rri-par[2]))/(sigmai*sigmai)*factor;
    case 13:
      if (t<50) dlnF_db += 2*par[3]*sigmai*exp(0.5*t)*factor;
      else      dlnF_db += 2*par[3]*sigmai;
      break;
    }
  }

  double f_ring=1;
  
  switch (iflag) {
  case  0:
  case  4:
    f = n0T * ( par[3]*par[3]*search_area*pi_fact + par[2]*par[2]*f_ring ) - lnF; 
    break;
    
  case 10:
    grad[0] =  - dlnF_dx;
  case 11:
    grad[1] =  - dlnF_dy;
  case 12:
    grad[2] = n0T * 2*par[2] * f_ring - dlnF_dr;
  case 13:
    grad[3] = n0T * 2*par[3] * search_area * pi_fact - dlnF_db;
    break;
  } 
}



///////////////////////// get ROI ///////////////////////////////
int Ring::get_hits(Hits &hh) {
  int verb=0;

  if (verb) cout<<"Ring::get_hits"<<endl;

  if (RingSearchTrack::exist) {
    RICH_LambdaPhi->Correct()->newRing(ix0, iy0, radius,RingSearchTrack::bx, RingSearchTrack::by);
  }
  else
    RICH_LambdaPhi->Correct()->newRing(ix0, iy0, radius);

  corrHit ch;

  float drr=0;
  numhit=0;
  for (int i=0;i<hh.numhit;i++) {    
      Hit &h=hh.hit[i];

      //if (h.assigned) continue;/////////////////////////////////////////////////////

      register float ddx = RICH_LambdaPhi->Correct()->dPhi   ( h.x, h.y, h.plane);
      register float ddy = RICH_LambdaPhi->Correct()->dLambda( h.x, h.y, h.plane); 
      register float rri = sqrt(ddx*ddx+ddy*ddy);
      register float a   = (rri-radius);
      if (a*a>maxSigma2) continue;

      // -- corrected hits
      ch.x=ix0+ddx;
      ch.y=iy0+ddy;
      ch.h=&h;
      //hit_list.push_back(ch);

      hit_list[numhit]=ch;

      // -- uncorrected hits (historic reasons, used for ring assigment to mc track)
      hit[numhit++]=&h;

      drr+=sqrt(ddx*ddx+ddy*ddy);
      if (verb) cout<<"*";
  }
  if (!numhit) return 1;
  drr/=numhit;
  if (verb) cout<<" numhit="<<numhit<<endl;

  radius=drr;

  return 0;
}


int Ring::prepare() {
  

  if (RingSearchTrack::exist && RingSearchAttrib::use_seed_phi) 
    pinit(1)=param(1)=x0=RingSearchTrack::phi;
  else
    pinit(1)=param(1)=x0=ix0;

  if (RingSearchTrack::exist && RingSearchAttrib::use_seed_lambda)
    pinit(2)=param(2)=y0=RingSearchTrack::lambda;
  else 
    pinit(2)=param(2)=y0=iy0;

  pinit(3)=param(3)=radius;      // rad
  pinit(4)=param(4)=numhit-n0T*param(3)*param(3); // Rb
  if (param(4)<.01) param(4)=.01;
  else              param(4)=sqrt(param(4));


  // -- check if close to seed
  if (!IsCloseToSeed()) return 0;

  upper(1)=1;
  upper(2)=1;
  upper(3)=.06;
  upper(4)=10;

  lower(1)=-1;
  lower(2)=-1;
  lower(3)=.02;
  lower(4)=.01;

  step(1)=step(2)=1e-6;
  step(3)=1e-6;
  step(4)=1e-5;

  search_area=4*M_PI*radius*maxSigma;

  return 1;
}


int Ring::check_limits() {
  int ret=0;
  for (int i=1; i<=n_par; i++) {
    if (param(i)>upper(i)) {
      total_upper_limit[i-1]++;
      param(i)=upper(i);
      ret=1;
    }
    if (param(i)<lower(i)) {
      param(i)=lower(i);
      total_lower_limit[i-1]++;
      ret=2;
    }
  }
  return ret;
}



////////////////////////////////// ring fit /////////////////////////////////////////
int Ring::fitring(Hits &hh) {
  int verb=0;
  if (verb) cout<<"Ring::fitring"<<endl;

  // -- check if found ring correstponds to seed
  if (!prepare()) return 0;

  // -- get hits
  get_hits(hh);

  // -- reset fit
  double f;
  double f_cut=.01; // change for exiting
  int kicks=0, negerr=0;
  niter=nerro=0;
  double f_old=1e5;
  double f_first;
  float sign[3]={1};
  float scale[3]={.00025}; 
  bool  kick=false;
  bool  reinvert=false;

  // -- initial function
  int iflag=0;
  fitfun(n_par, &beta(1), f, &param(1), iflag);
  f_first=f;

  if (verb>2) cout<<"First function="<<f_first<<endl;
  if (verb>2) cout<<"Initial params: "<<param(1)<<" "<<param(2)<<" "<<param(3)<<" "<<param(4)*param(4)<<endl;

  // -- minimize
  double lambda=.01;//.001;
  while (1) {

    // -- check for loop exit
    if (kick==false && reinvert==false) {
      if (fabs((f-f_old)/f)<f_cut)   break;
      f_old=f;
    }
    if (niter>max_iterations)     break;
    if (nerro>max_errors)         break;

    kick=false;
    reinvert=false;

    // -- find first derivatives
    iflag=10;
    fitfun(n_par, &beta(1), f, &param(1), iflag);

    if (verb==4) cout<<"derivatives="<<beta<<endl;

    // -- find second derivatives
    for (int i=0; i<n_par; i++) {
      iflag=10+i;
      param(i+1) -= step(i+1);
      fitfun(n_par, &gamma1(1), f, &param(1), iflag);
      param(i+1) += step(i+1);
      
      param(i+1) += step(i+1);
      iflag=10+i;
      fitfun(n_par, &gamma2(1), f, &param(1), iflag);
      param(i+1) -= step(i+1);
      for (int j=n_par; j>i; j--) {
	alpha(j,i+1) = (gamma2(j)-gamma1(j))/(2*step(i+1)) * (1+lambda) ;
      }
    }

    if (verb==4) cout<<"Alpha="<<alpha<<endl;

    beta*=(-.5);

    // -- find error matrix
    iflag=0;
    for (int i=1; i<=n_par; i++) if (fabs(alpha(i,i))<1e-20) {
      iflag=1;
      break;
    }
    if (!iflag) Error=alpha.inverse(iflag);
    if (iflag) {
      //cout<<"Alpha that caused wrong inversion:"<<alpha<<endl;
      //cout<<gamma1<<gamma2<<step<<endl;
      nerro++;
      if (errf) cerr<<"Problems with matrix inversion"<<endl;
      total_inversion_error++;
      reinvert=true;
      continue;
    }
    for (int i=1; i<=n_par; i++) {
      if (alpha(i,i)>=0 && Error(i,i)>=0) total_good_matrix++;
      if (alpha(i,i)<0  && Error(i,i)<0 ) total_negative_alpha_and_error++;
      if (alpha(i,i)<0) {
	total_negative_alpha++;
	if (errf)   cerr<<"Alpha matrix: negative diagonal elements row "<<i<<" after "<<niter<<" iterations"<<endl;
	if (errf>1) cerr<<"Alpha="<<alpha<<endl;
      }
      if (Error(i,i)<0) {
	total_negative_error++;
	negerr++;
	if (errf) cerr<<"Error matrix: negative diagonal elements row "<<i<<" after "<<niter<<" iterations"<<endl;
	//for (int j=1; j<=n_par; j++) Error(i,j)=Error(j,i)=0;
	Error(i,i) = -Error(i,i);
      }
    }

    // -- find shifts
    shift=Error*beta;
    param+=shift;

    // -- make a kick if not in minimum
    for (int i=1; i<n_par; i++) {
      if (alpha(i,i)<0) {
	kick=true;
	param(i)+=sign[i-1]*scale[i-1];
	sign[i-1] *=(-1);
	if (sign[i-1]>0) scale[i-1]*=1.5;
      }
    }
    if (kick) {
      iflag=0; fitfun(n_par, &beta(1), f, &param(1), iflag);
      kicks++;
      nerro++;
      ////////////////////////////////////////////////////////////////////////////////////////continue;
    }
    if (kick>10) kick=false;
   
   check_limits();

    niter++;
    iflag=0;
    fitfun(n_par, &beta(1), f, &param(1), iflag);
    
    // -- check conversion
    if (f<f_old)  lambda *= .1;
    else {
      //param  -= shift;
      lambda *= 10.;
    }
    
  } // while loop

  
  // -- overlooped and problems with inversion
  if ( reinvert && (niter>max_iterations||nerro>max_errors) ) return 0;

  if (f_first<f) {
    if (errf) cerr<<"Skipping update , first="<<f_first<<"  fit="<<f
		  <<" iter="<<niter<<"  err="<<nerro
		  <<endl;
    total_skipping_update++;
    return 0;
  }
  else total_update++;

  chi2=refit_L=f;

  if (!finalize()) return 0;

  
  if (verb>2) cout<<"Final  funct:"<<f<<endl;
  if (verb>2) cout<<"Final params: "<<param(1)<<" "<<param(2)<<" "<<param(3)<<" "<<param(4)*param(4)<<endl;
  if (verb>2) cout<<"Final errors: "<<sqrt(Error(1,1))<<" "<<sqrt(Error(2,2))<<" "<<sqrt(Error(3,3))<<" "
		  <<2*param(4)*sqrt(Error(4,4))<<endl;
  if (verb>2) cout<<"In this fit: iterations="<<niter<<" errors="<<nerro<<" kicks="<<kicks<<" negerr="<<negerr<<endl;


  if (verb) cout<<"total good matrix="<<total_good_matrix
		<<"  total negative alpha="<<total_negative_alpha
		<<"  total_negative_error="<<total_negative_error
		<<"  total_negative_alpha&error="<<total_negative_alpha_and_error
		<<"  total_inversion_error="<<total_inversion_error
		<<"  total_Br_limit="<<total_upper_limit[3]
		<<"  total_radius_limit="<<total_lower_limit[2]
		<<" (iteration level)"<<endl;
  if (verb) cout<<"  total update="<<total_update
		<<"  total_skipping_update="<<total_skipping_update<<" (fit level)"<<endl;

  return 1;
}




//////////////////////// finalize fit /////////////////////////////////
int Ring::finalize() {

  // -- replace old parameters
  x0=param(1);
  y0=param(2);
  radius=param(3);
  Rb=param(4)*param(4);

  for (int i=1; i<n_par; i++) {
    if (Error(i,i)<0) {
      if (errf) cerr<<"negetive diagonal elements row "<<i<<endl;
      Error(i,i)= - Error(i,i);
    }
    for (int j=1; j<n_par; j++) {
      cov[i-1][j-1]=Error(i,j); 
    }
  }

  av_accept=0;
  numhit_peak=0;
  refit_rms=0;

  for (int i=0; i<numhit; i++) {
    corrHit &ch=hit_list[i];

    float sigmai=0;
    if      (ch.h->size==16) sigmai=Sigma16;
    else if (ch.h->size==4 ) sigmai=Sigma4;
    else {
      fprintf(stderr,"Hit not found\n");
      continue;
    }

    float ddd = (Rb>0) ? radius/(Rb*sigmai) : radius/sigmai;
    if (ddd<1) continue;
    float delta_r = sqrt(2*log(ddd));
    float ddx = ch.h->x - x0;
    float ddy = ch.h->y - y0; 
    float dri = fabs(sqrt(ddx*ddx+ddy*ddy)-radius);
    float rri = dri/sigmai;
    if (rri>delta_r) continue;
    
    ch.h->assigned=1; // global
    assigned[i]=true; // local;

    numhit_peak++;

    refit_rms+=dri*dri;
    av_accept+=delta_r*sigmai;
  }

  // --- quality cuts:
  if (!numhit_peak) {
    if (errf) cerr<<"No counts under the peak: x0="<<x0<<" y0="<<y0<<" r0="<<radius<<" Rb="<<Rb<<endl;
    return 0; // no counts under peak
  }
  av_accept/=float(numhit_peak);

  numhit_3sigma = numhit_peak;
  sigma_signal  = av_accept;

  float accept_area=4*M_PI*radius*av_accept;

  numhit_bgrd  = accept_area * n0T * pi_fact * Rb; // background photons
  numhit_refit = numhit_peak-numhit_bgrd; // signal photons

  refit_sn  = (numhit_bgrd>0) ? numhit_refit/sqrt(numhit_bgrd) : 100; // S/sqrt(B)
  refit_rms = sqrt(refit_rms/numhit_peak); // sigma=rms

  // significance
  float A     = n0T*search_area*pi_fact;
  refit_Lbg = (n0T*radius*radius+A*Rb) - numhit*log(Rb);
  //refit_Lbg   = numhit*(log(numhit/A)-1);
  float Qdiff = refit_Lbg-refit_L;
  refit_S   = (Qdiff>0) ? sqrt(2*(refit_Lbg-refit_L)) : 0;

  if (refit_S<RingSearchAttrib::Scut) {
    for (int i=0; i<numhit; i++) {
      corrHit &ch=hit_list[i];
      ch.h->assigned=0;   // global
      assigned[i]=false;  // local 
    }
    return 0;
  }

  // -- fit type
  //fittype = (RingSearchTrack::exist) ? RingSearchType::seed : RingSearchType::standalone;
  //fittype = (fittype|RingSearchType::lkbg);

  return 1;
}


bool Ring::IsCloseToSeed() {

  if (!RingSearchTrack::exist) return true;

  dphi_seed=RingSearchTrack::phi    - x0;
  dlam_seed=RingSearchTrack::lambda - y0;

  if (fabs(dphi_seed)>RingSearchTrack::max_phi_residual ||
      fabs(dlam_seed)>RingSearchTrack::max_lambda_residual) { 
    return false;
  }

  return true;
}
