// *****************************************************************************
//
// source:    L3_Impact.C
//
// type:      source code
//
// created:   06 Dec 1999
//
// author:    Stefan Scharein, Ullrich Schwanke
//            HERA-B Collaboration
//            DESY Zeuthen
//            Platanenallee 6
//            D-15738 Zeuthen
//            Germany
//            email: scharein@ifh.de, schwanke@ifh.de
//
// symbol:    
//
// changes:
//
// *****************************************************************************
#include "L3_Impact.hh"
#include <CLHEP/Matrix/Vector.h>
#include <CLHEP/Vector/ThreeVector.h>

void L3_Impact::calc(L3_Track &t, const L3_Vertex &vtx)
{
  Hep3Vector p( t.x()-t.refz()*t.mx(),t.y()-t.refz()*t.my(),0);
  Hep3Vector v = t.Vector();
  Hep3Vector q = vtx.vertex();
  Hep3Vector k=q-p;
  float C = (k*v)/v.mag2();
  //t_mag = sqrt( k.mag2() - C*(k*v) );  
  t_mag = ( C*v - k ).mag();
  //calculate derivatives for vertex
  Hep3Vector Dq = (k - C*v)*(1.0/t_mag);
  //calculate derivatives for track
  HepVector Dtrack(4);  
  Dtrack[0] = -Dq.x();  //x
  Dtrack[1] = -Dq.y();  //y
  float L = t.refz()-C;
  Dtrack[2] = L*Dq.x(); //mx
  Dtrack[3] = L*Dq.y(); //my
  L3_CovMatrix<4> trackCov = t.getCov();
  L3_CovMatrix<3> vertexCov = vtx.getCov();
  t_variance = dot(Dtrack,trackCov*Dtrack) + (Dq*(vertexCov*Dq));  
}

void L3_Impact::calc(L3_Track &t, const L3_Target &targ)
{
  t.kalTransport(targ.z());
  if( targ.dx() > targ.dy() ){ //horizontal wire
    t_mag = t.y() - targ.y();
    t_variance = t.Cy();
  } else {
    t_mag = t.x() -  targ.x();
    t_variance = t.Cx();
  }
  if( t_mag<0.0 ){
    t_sign = -1.0;
    t_mag = fabs(t_mag);
  }
}

void L3_Impact::calc(L3_Track &t1, L3_Track &t2)
{
  Hep3Vector p1(t1.x()-t1.refz()*t1.mx(),t1.y()-t1.refz()*t1.my(),0);
  Hep3Vector v1 = t1.Vector();
  Hep3Vector p2(t2.x()-t2.refz()*t2.mx(),t2.y()-t2.refz()*t2.my(),0);
  Hep3Vector v2 = t2.Vector();
  Hep3Vector p = p1-p2;
  Hep3Vector w = v1.cross(v2);
  double magw = w.mag();
  t_mag = (w*p)/magw;
  double sign = +1.0;
  if( t_mag < 0.0 ){
    t_mag = -t_mag;
    sign = -1.0;
  }
  double C1 = 1.0/magw;
  double C2 = t_mag/magw;
  HepVector Dtra1(4),Dtra2(4);
  //first track
  Dtra1[0] = C1 * sign * w.x(); //x1
  Dtra1[1] = C1 * sign * w.y(); //y1
  Dtra1[2] = C1 * ( sign*(-t1.refz()*w.x()-p.y())-C2*(-w.y()+v2.y()*w.z()));  //mx1
  Dtra1[3] = C1 * ( sign*(-t1.refz()*w.y()+p.x())-C2*( w.x()-v2.x()*w.z()));  //my1
  //second track
  Dtra2[0] = - Dtra1[0]; //x1
  Dtra2[1] = - Dtra1[1]; //y2
  Dtra2[2] = C1 * ( sign*(t2.refz()*w.x()+p.y())-C2*( w.y()-v1.y()*w.z()));   //mx2
  Dtra2[3] = C1 * ( sign*(t2.refz()*w.y()-p.x())-C2*(-w.x()+v1.x()*w.z()));   //my2
  L3_CovMatrix<4> trackCov = t1.getCov();
  t_variance = dot(Dtra1,trackCov*Dtra1);
  trackCov = t2.getCov();
  t_variance += dot(Dtra2,trackCov*Dtra2);
}

void L3_Impact::calc(L3_Track& t, const L3_Vertex& v,const L3_Target *p){
  if( p==NULL ) p = &L3_Geometry::ClosestTarget(v.vertex());
  t.kalTransport(p->z());

  t_mag = sqrt( pow(t.x()-v.x(),2) + pow(t.y()-v.y(),2) );
  if( p->dx() > p->dy() ){ //horizontal wire
    if( t.y() < v.y() ) t_sign = -1.0; //i.e. t.y()-v.y() < 0.0
  } else { //vertical wire
    if( t.x() < v.x() ) t_sign = -1.0; //i.e. t.x()-v.x() < 0.0
  }
  //----- track
  HepVector Dtrack(2);
  Dtrack[0] = (t.x()-v.x())/t_mag;
  Dtrack[1] = (t.y()-v.y())/t_mag;
  L3_CovMatrix<2> trackCov;
  trackCov(0,0) = t.Cov(0,0);
  trackCov(1,1) = t.Cov(1,1);
  trackCov(1,0) = t.Cov(1,0);
  trackCov(0,1) = t.Cov(0,1);
  //------ vertex
  HepVector Dvtx(2);
  Dvtx[0] = -Dtrack[0];
  Dvtx[1] = -Dtrack[1];
  L3_CovMatrix<2> vertexCov;
  vertexCov(0,0) = v.Cov(0,0);
  vertexCov(1,1) = v.Cov(1,1);
  vertexCov(1,0) = v.Cov(1,0);
  vertexCov(0,1) = v.Cov(0,1);  
  //------ calculate variance
  t_variance = dot(Dtrack,trackCov*Dtrack) + dot(Dvtx,vertexCov*Dvtx);  
}


